/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "animcontext.h"
#include "worldstate.h"

Sith::Gameplay::Animation* Sith::Gameplay::Animation::StaticInterface = 0;

int Sith::Gameplay::Animation::AnimContext::AnimCount = 0;

Sith::Gameplay::Animation::Animation() : Task("Animation Process")
{
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple instances of the animation process for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	// register cog bindings as soon as you remember
	Sith::COG::CogLib::InjectLibraryFunction("stopanim", 1, &Stop);
	Sith::COG::CogLib::InjectLibraryFunction("stopsurfaceanim", 1, &StopSurfaceAnim);
	Sith::COG::CogLib::InjectLibraryFunction("slideceilingsky", 2, &SlideCeilingSky);
	Sith::COG::CogLib::InjectLibraryFunction("slidehorizonsky", 2, &SlideHorizonSky);
	Sith::COG::CogLib::InjectLibraryFunction("slidesurface", 3, &SlideSurface);
	Sith::COG::CogLib::InjectLibraryFunction("slidewall", 3, &SlideSurface);
	Sith::COG::CogLib::InjectLibraryFunction("surfaceanim", 3, &SurfaceAnim);
	Sith::COG::CogLib::InjectLibraryFunction("setsurfacecel", 2, &SetSurfaceCel);
	Sith::COG::CogLib::InjectLibraryFunction("setwallcel", 2, &SetSurfaceCel);
	Sith::COG::CogLib::InjectLibraryFunction("getsurfacecel", 1, &GetSurfaceCel);
	Sith::COG::CogLib::InjectLibraryFunction("getwallcel", 1, &GetSurfaceCel);

	Sith::COG::CogLib::InjectLibraryFunction("adddynamictint", 4, &AddDynamicTint);
	Sith::COG::CogLib::InjectLibraryFunction("freecoloreffect", 1, &FreeColorEffect);
	Sith::COG::CogLib::InjectLibraryFunction("newcoloreffect", 10, &NewColorEffect);
};

Sith::Gameplay::Animation::~Animation()
{
	StopAll();
};

Sith::SITHERROR Sith::Gameplay::Animation::Process()
{
	double dT = theTimer.Query();

	if(Gameplay::WorldState::Get().CurrentLevel.isValid())
	{
		for(std::list<AnimContext*>::iterator it = AnimList.begin(); it != AnimList.end(); ++it)
		{
			(*it)->Update(dT);
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Gameplay::Animation::StopAll()
{
	AnimContext* curr;
	for(std::list<AnimContext*>::iterator it = AnimList.begin(); it != AnimList.end(); ++it)
	{
		curr = (*it);
		delete curr;
	};

	AnimList.clear();
	return SE_OK;
};

Sith::SITHERROR Sith::Gameplay::Animation::StopAnim(int Anim)
{
	AnimContext* curr;
	for(std::list<AnimContext*>::iterator it = AnimList.begin(); it != AnimList.end(); ++it)
	{
		if((*it)->GetIndex() == Anim)
		{
			curr = (*it);
			AnimList.erase(it);
			delete curr;
			break;
		};
	};

	return SE_OK;
};

Sith::Gameplay::Animation::AnimContext::AnimContext()
{
	Index = AnimCount++;
};

Sith::Gameplay::Animation::CeilingSkyAnim::CeilingSkyAnim(float u_speed, float v_speed)
{
	dU = u_speed;
	dV = v_speed;
};

Sith::SITHERROR Sith::Gameplay::Animation::CeilingSkyAnim::Update(double dT)
{
	VFS::WisePointer<Res::Level> CurrLev = Gameplay::WorldState::Get().CurrentLevel;

	if(CurrLev.isValid())
	{
		CurrLev->Header.CeilingSkyOffset[0] += (dU * (float)dT);
		CurrLev->Header.CeilingSkyOffset[1] += (dV * (float)dT);
	};

	return SE_OK;
};

Sith::Gameplay::Animation::HorizonSkyAnim::HorizonSkyAnim(float u_speed, float v_speed)
{
	dU = u_speed;
	dV = v_speed;
};

Sith::SITHERROR Sith::Gameplay::Animation::HorizonSkyAnim::Update(double dT)
{
	VFS::WisePointer<Res::Level> CurrLev = Gameplay::WorldState::Get().CurrentLevel;

	if(CurrLev.isValid())
	{
		CurrLev->Header.HorizonSkyOffset[0] += (dU * (float)dT);
		CurrLev->Header.HorizonSkyOffset[1] += (dV * (float)dT);
	};

	return SE_OK;
};

Sith::Gameplay::Animation::SurfaceSlideAnim::SurfaceSlideAnim(int _Surface, Sith::Math::Vector _Direction, float _Speed)
{
	Direction = _Direction;
	Surface = _Surface;
	Speed = _Speed;
	dU = 0;
	dV = 0;

	VFS::WisePointer<Res::Level> CurrLev = Gameplay::WorldState::Get().CurrentLevel;

	if((CurrLev.isValid()) && (Surface >= 0 && Surface < CurrLev->SurfaceList.size()))
	{
		Res::Level::SURFACE *s = &CurrLev->SurfaceList[Surface];
		if (s->TextureIndex != -1)
		{
			//Sith::Win32::HBMGeometry TexGeom = VFS::ResourceManager::Get().GetMaterial("dflt.mat", "dflt.cmp")->GetHBM(0).GetGeometry();

			//dU = s->TexDensityU;
			dU = s->Tangent.X * Direction.x + s->Tangent.Y * Direction.y + s->Tangent.Z * Direction.z;
			dV = s->Binormal.X * Direction.x + s->Binormal.Y * Direction.y + s->Binormal.Z * Direction.z;
			dU *= .1f * Speed * s->TexDensityU;
			dV *= .1f * Speed * s->TexDensityV;
		}
	};

	return;
};

Sith::SITHERROR Sith::Gameplay::Animation::SurfaceSlideAnim::Update(double dT)
{
	VFS::WisePointer<Res::Level> CurrLev = Gameplay::WorldState::Get().CurrentLevel;

	if(CurrLev.isValid())
	{
		CurrLev->SurfaceList[Surface].TexShiftU += (float)dT * dU;
		CurrLev->SurfaceList[Surface].TexShiftV += (float)dT * dV;
	};

	return SE_OK;
};

Sith::Gameplay::Animation::SurfaceCelAnim::SurfaceCelAnim(int _Surface, float _Framerate, int _Flags)
{
	Surface = _Surface;
	Frametime = 1 / _Framerate;
	Flags = _Flags;

	TimeSinceLastFrame = 0.0;

	if(Flags & Sith::Flags::SurfaceAnimation::StartOnThirdCelAndSkipFirst)
	{
		CurrentFrame = 2;
	}
	else if(Flags & Sith::Flags::SurfaceAnimation::StartOnFourthCelAndSkipFirst2)
	{
		CurrentFrame = 3;
	}
	else
	{
		CurrentFrame = 0;
	};

	if(Gameplay::WorldState::Get().CurrentLevel.isValid() && (Surface >= 0) && (Surface < Gameplay::WorldState::Get().CurrentLevel->SurfaceList.size()))
	{
        Framecount = Sith::VFS::ResourceManager::Get().GetMaterial(Gameplay::WorldState::Get().CurrentLevel->TextureList[Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Surface].TextureIndex].TextureName, "dflt.cmp")->GetHBMCount();
		Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Surface].CurrentCel = CurrentFrame;
	};

	return;
};

Sith::SITHERROR Sith::Gameplay::Animation::SurfaceCelAnim::Update(double dT)
{
	VFS::WisePointer<Res::Level> CurrLev = Gameplay::WorldState::Get().CurrentLevel;

	if(CurrLev.isValid())
	{
		if((Surface >= 0) && (Surface < CurrLev->SurfaceList.size()))
		{
			TimeSinceLastFrame += dT;
			if(TimeSinceLastFrame >= Frametime)
			{
				TimeSinceLastFrame = 0.0;

				++CurrentFrame;

				if(CurrentFrame == Framecount)
				{
					if(Flags & Sith::Flags::SurfaceAnimation::Looping)
					{
						if(Flags & Sith::Flags::SurfaceAnimation::StartOnThirdCelAndSkipFirst)
						{
							CurrentFrame = 1;
						}
						else if(Flags & Sith::Flags::SurfaceAnimation::StartOnFourthCelAndSkipFirst2)
						{
							CurrentFrame = 2;
						}
						else
						{
							CurrentFrame = 0;
						};
					}
					else
					{
						CurrentFrame = Framecount - 1;
					};
				};

				Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Surface].CurrentCel = CurrentFrame;
			};
		};
	};

	return SE_OK;
};

COGLIB Sith::Gameplay::Animation::SetSurfaceCel(Sith::COG::VMType* surface, Sith::COG::VMType* celnum)
{
	VFS::WisePointer<Res::Level> curLevel = Gameplay::WorldState::Get().CurrentLevel;

	if(curLevel.isValid())
	{
		if((surface->GetInt() >= 0) && (surface->GetInt() < curLevel->SurfaceList.size()))
		{
            Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].CurrentCel = celnum->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::Animation::GetSurfaceCel(Sith::COG::VMType* surface)
{
	VFS::WisePointer<Res::Level> curLevel = Gameplay::WorldState::Get().CurrentLevel;

	if(curLevel.isValid())
	{
		if((surface->GetInt() >= 0) && (surface->GetInt() < curLevel->SurfaceList.size()))
		{
			return new COG::VMType((int)Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].CurrentCel);
		};
	};
	
	return new COG::VMType(COG::VMTC_VOID);
};