/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* COG scanner */

%option noyywrap
%option prefix="AR"

%{
#include "stdafx.h"

#include "ares.h"

int ARerror(char *s);
int ARlineno = 1;

_aresval ARESval;

YY_BUFFER_STATE ARStringBuffer;

%}

/* Simple digit */
DIGIT		[0-9]
/* Alphanumeric characters used in a hex number*/
HEXALPHA	[a-fA-F]
/* Hextail part */
HEXTAIL		({DIGIT}|{HEXALPHA}){1,8}
/* Hexadecimal number */
HEX		0[xX]{HEXTAIL}
/* Identifier */
ID 	[a-zA-Z_][a-zA-Z0-9_]*

%%

"#"[^\n]*\n	/* eat one-line comments */ ++ARlineno;
"//"[^\n]*\n	/* eat one-line comments */ ++ARlineno;

{HEX}		{
			size_t Len = strlen(yytext);
			int *digit = new int[Len - 2];
			int RetVal = 0;
			unsigned int Count = 0;
			unsigned int i = 0, j = 0;
			for(i = 2; i < strlen(yytext); ++i)
			{
				if(yytext[i] > 0x29 && yytext[i] < 0x40)
					digit[i - 2] = yytext[i] & 0x0F;
				else if(yytext[i] >= 'a' && yytext[i] <= 'f')
					digit[i - 2] = (yytext[i] & 0x0F) + 9;
				else if(yytext[i] >= 'A' && yytext[i] <= 'F')
					digit[i - 2] = (yytext[i] & 0x0F) + 9;
				else break;
				
				++Count;
			}
			j = Count - 1;
			for(i = 0; i < Count; ++i)
			{
				RetVal = RetVal | (digit[i] << (j << 2));
				--j;
			};
			
			ARESval.Integer = RetVal;
			delete[] digit;
			return ARES_INTEGER;
		}

{DIGIT}+	{ ARESval.Integer = atoi(yytext); return ARES_INTEGER; }

{DIGIT}+"."{DIGIT}*[fF]		{ ARESval.Float = (float)atof(yytext); return ARES_FLOAT; }

{DIGIT}+"."{DIGIT}*	{ ARESval.Float = (float)atof(yytext); return ARES_FLOAT; }

\"(\\.|[^\\"])*\"	{ ARESval.String = strdup(ARtext); return ARES_STRING; }

{ID}	{ ARESval.String = strdup(ARtext); return ARES_IDENTIFIER; }

[\n]		/* eat up newlines, add to line counter */ ++ARlineno;
[ \t]+		/* eat up whitespace */

.	printf("Unrecognized character: %s\n", ARtext);
	
%%

bool ARSetInputString(const char* Input)
{
	ARlineno = 1;
	ARStringBuffer = yy_scan_string(Input);
	return true;
};

bool ARCloseInputString()
{
	yy_delete_buffer(ARStringBuffer);
	return true;
};