/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "cmpres.h"

Sith::Res::Colormap::Colormap(std::string Name) : Sith::VFS::Resource(Name)
{
	memset(Palette, 0x00, sizeof(ColormapElement) * 256);
	return;
};

Sith::Res::Colormap::~Colormap()
{
	return;
};

Sith::SITHERROR Sith::Res::Colormap::LoadFromFile(FILE *FileHandle)
{
	CMPHEADER Header;

	if(fread(&Header, sizeof(CMPHEADER), 1, FileHandle) != 1)
	{
		ConOut("ERROR: Could not load colormap header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	};

	if(strncmp(Header.Magic, "CMP ", 4) != 0)
	{
		ConOut("ERROR: Not a valid colormap file.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	};

	if(Header.Version != 30)
	{
		ConOut("ERROR: Unrecognized colormap version.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	};

	// Still here? It's a good header. Read in colormap data.
	for(int i = 0; i < 256; ++i)
	{
		fread(&Palette[i].Elements.R, sizeof(unsigned char), 1, FileHandle);
		fread(&Palette[i].Elements.G, sizeof(unsigned char), 1, FileHandle);
		fread(&Palette[i].Elements.B, sizeof(unsigned char), 1, FileHandle);
		Palette[i].Elements.A = 0xFF;
	};

	// Ignore the rest. Sith 2 doesn't use some crappy software renderer ;)
	return SE_OK;
};