/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* COG scanner */

%option noyywrap
%option prefix="COG"

%{
#include "stdafx.h"

#include "vmtype.h"

#include "cog.y.hpp"
int COGerror(char *s);
int COGlineno = 1;

YY_BUFFER_STATE StringBuffer;

%}

/* Simple digit */
DIGIT		[0-9]
/* Alphanumeric characters used in a hex number*/
HEXALPHA	[a-fA-F]
/* Hextail part */
HEXTAIL		({DIGIT}|{HEXALPHA}){1,8}
/* Hexadecimal number */
HEX		0[xX]{HEXTAIL}
/* Identifier */
ID 	[a-zA-Z_][a-zA-Z0-9_]*

%%

"#"[^\n]*\n	/* eat one-line comments */ ++COGlineno;
"//"[^\n]*\n	/* eat one-line comments */ ++COGlineno;

"break"		return VMTT_KEYWORD_BREAK;
"call"		return VMTT_KEYWORD_CALL;
"continue"	return VMTT_KEYWORD_CONTINUE;
"else"		return VMTT_KEYWORD_ELSE;
"for"		return VMTT_KEYWORD_FOR;
"if"		return VMTT_KEYWORD_IF;
"return"	return VMTT_KEYWORD_RETURN;
"stop"		return VMTT_KEYWORD_RETURN; /* stop does the same thing as return */
"while"		return VMTT_KEYWORD_WHILE;
"do"		return VMTT_KEYWORD_DO;

"symbols"	return VMTT_KEYWORD_SYMBOLS;
"code"		return VMTT_KEYWORD_CODE;
"end"		return VMTT_KEYWORD_END;

"true"		{ COGlval.Boolean = true; return VMTT_LITERAL_BOOLEAN; }
"false"		{ COGlval.Boolean = false; return VMTT_LITERAL_BOOLEAN; }

"getparam"			return VMTT_KEYWORD_GETPARAM;
"getsenderid"		return VMTT_KEYWORD_GETSENDERID;
"getsenderref"		return VMTT_KEYWORD_GETSENDERREF;
"getsendertype"		return VMTT_KEYWORD_GETSENDERTYPE;
"getsourceref"		return VMTT_KEYWORD_GETSOURCEREF;
"getsourcetype"		return VMTT_KEYWORD_GETSOURCETYPE;
"returnex"			return VMTT_KEYWORD_RETURNEX;
"sleep"				return VMTT_KEYWORD_SLEEP;

{HEX}		{
			size_t Len = strlen(yytext);
			int *digit = new int[Len - 2];
			int RetVal = 0;
			unsigned int Count = 0;
			unsigned int i = 0, j = 0;
			for(i = 2; i < strlen(yytext); ++i)
			{
				if(yytext[i] > 0x29 && yytext[i] < 0x40)
					digit[i - 2] = yytext[i] & 0x0F;
				else if(yytext[i] >= 'a' && yytext[i] <= 'f')
					digit[i - 2] = (yytext[i] & 0x0F) + 9;
				else if(yytext[i] >= 'A' && yytext[i] <= 'F')
					digit[i - 2] = (yytext[i] & 0x0F) + 9;
				else break;
				
				++Count;
			}
			j = Count - 1;
			for(i = 0; i < Count; ++i)
			{
				RetVal = RetVal | (digit[i] << (j << 2));
				--j;
			};
			
			COGlval.Integer = RetVal;
			delete[] digit;
			return VMTT_LITERAL_INTEGER_HEX;
		}

{DIGIT}+	{ COGlval.Integer = atoi(yytext); return VMTT_LITERAL_INTEGER; }

{DIGIT}*"."{DIGIT}+[fF]		{ COGlval.Float = (float)atof(yytext); return VMTT_LITERAL_FLOAT; }

{DIGIT}*"."{DIGIT}+	{ COGlval.Double = atof(yytext); return VMTT_LITERAL_DOUBLE; }

\"(\\.|[^\\"])*\"	{ COGlval.String = strdup(yytext); return VMTT_LITERAL_STRING; }

{ID}	{ COGlval.String = strdup(yytext); return VMTT_IDENTIFIER; }

">>="		return VMTT_PUNC_ASSIGN_SHL;
"<<="		return VMTT_PUNC_ASSIGN_SHR;
"+="		return VMTT_PUNC_ASSIGN_ADD;
"-="		return VMTT_PUNC_ASSIGN_SUB;
"*="		return VMTT_PUNC_ASSIGN_MUL;
"/="		return VMTT_PUNC_ASSIGN_DIV;
"%="		return VMTT_PUNC_ASSIGN_MOD;
"&="		return VMTT_PUNC_ASSIGN_AND;
"^="		return VMTT_PUNC_ASSIGN_XOR;
"|="		return VMTT_PUNC_ASSIGN_OR;
">>"		return VMTT_PUNC_SHL;
"<<"		return VMTT_PUNC_SHR;
"++"		return VMTT_PUNC_INC;
"--"		return VMTT_PUNC_DEC;
"&&"		return VMTT_PUNC_EQUALITY_AND;
"||"		return VMTT_PUNC_EQUALITY_OR;
"<="		return VMTT_PUNC_EQUALITY_LTE;
">="		return VMTT_PUNC_EQUALITY_GTE;
"=="		return VMTT_PUNC_EQUALITY_EQ;
"!="		return VMTT_PUNC_EQUALITY_NEQ;
";"		return VMTT_PUNC_ENDSTMT;
("{"|"<%")	return VMTT_PUNC_LBRACE;
("}"|"%>")	return VMTT_PUNC_RBRACE;
","		return VMTT_PUNC_COMMA;
"::"	return VMTT_PUNC_DBLCOLON;
":"		return VMTT_PUNC_COLON;
"="		return VMTT_PUNC_ASSIGN;
"("		return VMTT_PUNC_LPAREN;
")"		return VMTT_PUNC_RPAREN;
("["|"<:")	return VMTT_PUNC_LBRACKET;
("]"|":>")	return VMTT_PUNC_RBRACKET;
"&"		return VMTT_PUNC_OR;
"!"		return VMTT_PUNC_EQUALITY_NOT;
"-"		return VMTT_PUNC_SUB;
"+"		return VMTT_PUNC_ADD;
"*"		return VMTT_PUNC_MUL;
"/"		return VMTT_PUNC_DIV;
"%"		return VMTT_PUNC_MOD;
"<"		return VMTT_PUNC_EQUALITY_LT;
">"		return VMTT_PUNC_EQUALITY_GT;
"^"		return VMTT_PUNC_XOR;
"|"		return VMTT_PUNC_OR;
"'"		return VMTT_PUNC_APOS;

[\n]		/* eat up newlines, add to line counter */ ++COGlineno;
[ \t\r]+		/* eat up whitespace */

.	COGerror("Unrecognized character");
	
%%

bool SetInputString(const char* Input)
{
	StringBuffer = yy_scan_string(Input);
	return true;
};

bool CloseInputString()
{
	yy_delete_buffer(StringBuffer);
	return true;
};