/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "vmtype.h"
#include "opcodes.h"
#include "instruction.h"
#include "vmheap.h"
#include "stackmachine.h"
#include "coglib.h"
#include "cogparser.h"

#include "cog.y.hpp"

#include <fstream>

// Prototype for the parser
extern int COGparse(void);
extern bool SetInputString(const char* Input);
extern bool CloseInputString();

extern int COGlineno;

Sith::COG::StackMachine *GlobalStackMachine = 0;

/*
std::string dbgConvertToString(Sith::COG::Opcode OPCODE)
{
	switch(OPCODE)
	{
	case Sith::COG::Opcodes::NOP:
		return "NOP";

	case Sith::COG::Opcodes::SYSCALL:
		return "SYSCALL";
	case Sith::COG::Opcodes::LIBCALL:
		return "LIBCALL";
	case Sith::COG::Opcodes::MSGCALL:
		return "MSGCALL";
	case Sith::COG::Opcodes::RETURN:
		return "RETURN";

	case Sith::COG::Opcodes::LOAD:
		return "LOAD";
	case Sith::COG::Opcodes::STOR:
		return "STOR";

	case Sith::COG::Opcodes::ILOD:
		return "ILOD";
	case Sith::COG::Opcodes::ISTR:
		return "ISTR";

	case Sith::COG::Opcodes::PUSH:
		return "PUSH";
	case Sith::COG::Opcodes::POP:
		return "POP";
	case Sith::COG::Opcodes::FLSH:
		return "FLSH";
	case Sith::COG::Opcodes::RPSH:
		return "RPSH";

	case Sith::COG::Opcodes::ADD:
		return "ADD";
	case Sith::COG::Opcodes::SUB:
		return "SUB";
	case Sith::COG::Opcodes::MUL:
		return "MUL";
	case Sith::COG::Opcodes::DIV:
		return "DIV";
	case Sith::COG::Opcodes::MOD:
		return "MOD";

	case Sith::COG::Opcodes::AND:
		return "AND";
	case Sith::COG::Opcodes::OR:
		return "OR";
	case Sith::COG::Opcodes::XOR:
		return "XOR";
	case Sith::COG::Opcodes::NOT:
		return "NOT";

	case Sith::COG::Opcodes::INC:
		return "INC";
	case Sith::COG::Opcodes::DEC:
		return "DEC";
	case Sith::COG::Opcodes::SHL:
		return "SHL";
	case Sith::COG::Opcodes::SHR:
		return "SHR";

	case Sith::COG::Opcodes::CZ:
		return "CZ";
	case Sith::COG::Opcodes::CNZ:
		return "CNZ";

	case Sith::COG::Opcodes::CE:
		return "CE";
	case Sith::COG::Opcodes::CNE:
		return "CNE";

	case Sith::COG::Opcodes::CGT:
		return "CGT";
	case Sith::COG::Opcodes::CLT:
		return "CLT";
	case Sith::COG::Opcodes::CGTE:
		return "CGTE";
	case Sith::COG::Opcodes::CLTE:
		return "CLTE";

	case Sith::COG::Opcodes::CT:
		return "CT";
	case Sith::COG::Opcodes::CNT:
		return "CNT";

	case Sith::COG::Opcodes::CAND:
		return "CAND";
	case Sith::COG::Opcodes::COR:
		return "COR";

	case Sith::COG::Opcodes::JMP:
		return "JMP";

	case Sith::COG::Opcodes::JZ:
		return "JZ";
	case Sith::COG::Opcodes::JNZ:
		return "JNZ";

	case Sith::COG::Opcodes::JE:
		return "JE";
	case Sith::COG::Opcodes::JNE:
		return "JNE";

	case Sith::COG::Opcodes::JGT:
		return "JGT";
	case Sith::COG::Opcodes::JLT:
		return "JLT";
	case Sith::COG::Opcodes::JGTE:
		return "JGTE";
	case Sith::COG::Opcodes::JLTE:
		return "JLTE";

	case Sith::COG::Opcodes::JT:
		return "JT";
	case Sith::COG::Opcodes::JNT:
		return "JNT";

	case Sith::COG::Opcodes::JAND:
		return "JAND";
	case Sith::COG::Opcodes::JOR:
		return "JOR";

	case Sith::COG::Opcodes::ZEQ:
		return "ZEQ";

	default:
		return "INVALID_OPCODE";
	};
};*/

// LoadFromCOG();
// - Loads the GOB file at "Path" into the current stack machine
Sith::SITHERROR Sith::COG::StackMachine::Parse(std::string Script)
{
	COGlineno = 1;
	Parser::Buffer = Script.c_str();
	Parser::Offset = 0;

	const char* CodeSection = Script.c_str();

	std::string Tok;

	while(Parser::Buffer[Parser::Offset] != '\0')
	{
		Tok = Parser::GetIdentifier();

		if(Tok == "flags")
		{
			if(Parser::GetPunc() == "=")
			{
				Parser::GetInteger();
			}
			else
			{
				std::ostringstream Stream;
				Stream << "ERROR: Unexpected punctuator on line " << COGlineno << ".\n";
				ConOut(Stream.str(), ConOut::Standard);

				return SE_COG_SYNTAX_ERROR;
			};
		}
		else if(Tok == "symbols")
		{
			while(Parser::Buffer[Parser::Offset] != '\0')
			{
				// Type code:
				std::string Type = Parser::GetIdentifier();
				if(Type == "end") break;

				int currlineno = COGlineno;

				COG::VMTypeCode tc;

				if(Type == "ai") tc = VMTC_AI;
				else if(Type == "bool") tc = VMTC_BOOL;
				else if(Type == "cog") tc = VMTC_COG;
				else if(Type == "double") tc = VMTC_DOUBLE;
				else if(Type == "flex") tc = VMTC_FLEX;
				else if(Type == "float") tc = VMTC_FLOAT;
				else if(Type == "int") tc = VMTC_INT;
				else if(Type == "keyframe") tc = VMTC_KEYFRAME;
				else if(Type == "material") tc = VMTC_MATERIAL;
				else if(Type == "message") tc = VMTC_INSTRUCTION_ADDRESS;
				else if(Type == "model") tc = VMTC_MODEL;
				else if(Type == "sector") tc = VMTC_SECTOR;
				else if(Type == "sound") tc = VMTC_SOUND;
				else if(Type == "string") tc = VMTC_STRING;
				else if(Type == "surface") tc = VMTC_SURFACE;
				else if(Type == "template") tc = VMTC_TEMPLATE;
				else if(Type == "thing") tc = VMTC_THING;
				else if(Type == "vector") tc = VMTC_VECTOR;
				else if(Type == "void") tc = VMTC_VOID;
				else
				{
					std::ostringstream Stream;
					Stream << "WARNING: Unexpected type \'" << Type << "\' on line " << COGlineno << "\n";
					ConOut(Stream.str(), ConOut::Standard);
					tc = VMTC_VOID;
					return SE_COG_SYNTAX_ERROR;
				};

				COG::VMType AssignVal(tc);

				std::string Name = Parser::GetIdentifier();

				if(Parser::GetPunc() == "=")
				{
					// Handle assignment:
					switch(tc)
					{
					case VMTC_AI:
					case VMTC_COG:
					case VMTC_KEYFRAME:
					case VMTC_MATERIAL:
					case VMTC_MODEL:
					case VMTC_SOUND:
						Parser::GetFilename();
						break;

					case VMTC_BOOL:
						{
                            std::string Booleo;
							Booleo = Parser::GetIdentifier();
							if(Booleo.length() > 0)
							{
								if(Booleo == "true")
								{
									AssignVal = VMType(true);
								}
								else if(Booleo == "false")
								{
									AssignVal = VMType(false);
								}
								else
								{
									std::ostringstream Stream;
									Stream << "WARNING: Attempted to assign \'" << Booleo << "\' to boolean symbol \'" << Name << "\'.\n";
									ConOut(Stream.str(), ConOut::Warning);
								};
							}
							else
							{
								if(Parser::GetFloat() != 0)
								{
									AssignVal = VMType(true);
								}
								else
								{
									AssignVal = VMType(false);
								};
							};
						};
						break;

					case VMTC_TEMPLATE:
						AssignVal = VMType(tc, Parser::GetTemplate());
						break;
					case VMTC_STRING:
						AssignVal = VMType(tc, Parser::GetString());
						break;

					case VMTC_DOUBLE:
					case VMTC_FLOAT:
					case VMTC_FLEX:
						AssignVal = VMType(Parser::GetFloat());
						break;

					case VMTC_INT:
						{
							int OldPos = Parser::Offset;

							int Val1 = Parser::GetInteger();
							
							Parser::Offset = OldPos;

							int Val2 = (int)Parser::GetFloat();

							if(Val1 > Val2)
							{
								Parser::Offset = OldPos;
								Parser::GetInteger();
								AssignVal = VMType(Val1);
							}
							else
							{
								AssignVal = VMType(Val2);
							};
						};
						break;

					default:
						{
							std::ostringstream Stream;
							Stream << "WARNING: Unexpected default definition for symbol \'" << Name << "\'.\n";
							ConOut(Stream.str(), ConOut::Standard);
						};
						break;
					};
				};

				VMHeap::SymbolExtensionTable Ext;

				while(Parser::Buffer[Parser::Offset] != '\0')
				{
					// Nuke potential punctuator (LEC SUCKS!)
					Parser::GetPunc();

					if(currlineno != COGlineno) break;

					std::string ExtName = Parser::GetIdentifier();

					if(ExtName == "local")
					{
						Ext.isLocal = true;
					}
					else if(ExtName == "nolink")
					{
						Ext.NoLink = true;
					}
					else if(ExtName == "desc")
					{
						if(Parser::GetPunc() != "=")
						{
							std::ostringstream Stream;
							Stream << "WARNING: Expected \'=\' for \'Desc\' on line " << COGlineno << ".\n";
							ConOut(Stream.str(), ConOut::Warning);
							Parser::SkipToNewline();
							break;
						};

						Ext.Desc = Parser::GetIdentifier();
					}
					else if(ExtName == "linkid")
					{
						if(Parser::GetPunc() != "=")
						{
							std::ostringstream Stream;
							Stream << "WARNING: Expected \'=\' for \'LinkID\' on line " << COGlineno << ".\n";
							ConOut(Stream.str(), ConOut::Warning);
							Parser::SkipToNewline();
							break;
						};

						Ext.LinkID = Parser::GetInteger();
					}
					else if(ExtName == "mask")
					{
						if(Parser::GetPunc() != "=")
						{
							std::ostringstream Stream;
							Stream << "WARNING: Expected \'=\' for \'Mask\' on line " << COGlineno << ".\n";
							ConOut(Stream.str(), ConOut::Warning);
							Parser::SkipToNewline();
							break;
						};

						Ext.Mask = Parser::GetInteger();
					}
					else
					{
						std::ostringstream Stream;
						Stream << "WARNING: Unrecognized symbol extension \'" << ExtName << "\' on line " << COGlineno << ".\n";
						ConOut(Stream.str(), ConOut::Warning);
						break;
					};
				};

				Heap.AddLocal(Name, AssignVal, Ext);
			};

			while(Parser::Buffer[Parser::Offset] != '\0')
			{
				Parser::SkipSpace();

				if(_strnicmp(&Parser::Buffer[Parser::Offset], "code", 4) == 0)
				{
					Parser::Offset +=4;

					CodeSection = &Parser::Buffer[Parser::Offset];
					break;
				};

				if(Parser::Buffer[Parser::Offset] == '\n') ++COGlineno;

				++Parser::Offset;
			};

			if(Parser::Buffer[Parser::Offset] == '\0')
			{
				ConOut("ERROR: Failed to find code section in strict COG.\n", ConOut::Error);
				return SE_COG_SYNTAX_ERROR;
			};

			break;
		}
		else break;
	};

	// Assign the global stack machine 'this', so the parser can access it
	GlobalStackMachine = this;
	SetInputString(CodeSection);
	if(COGparse() == 1)
	{
        CloseInputString();
		return SE_COG_SYNTAX_ERROR;
	}
	else
	{
		CloseInputString();

/*		//	GlobalStackMachine = 0;
		//	freopen("CON", "r", stdin);

		// DEBUG OUTPUT
		// Dumps the CODE SEGMENT into a file of the following format:
		// [HEX INSTRUCTION] [DATA AS INT] [DATA AS ADDR]
		char DbgBuf[128];
		std::ofstream DbgLog("dbgout.txt");
		int DbgData;

		for(std::vector<Instruction>::iterator it = Code.begin(); it != Code.end(); ++it)
		{
			DbgData = it->DATA.GetAddr();
			if(DbgData == -1)
			{
				DbgData = it->DATA.GetInt();
			}

			sprintf(DbgBuf, "%s %d\n", dbgConvertToString(it->OPCODE).c_str(), DbgData);
			DbgLog << DbgBuf;
		};

		DbgLog.close();*/

		return SE_OK;
	};
};
