/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _COGPARSER_H_INCLUDED_
#define _COGPARSER_H_INCLUDED_

extern int COGlineno;

namespace Sith
{
	namespace COG
	{
		namespace Parser
		{
			const char* Buffer;
			int Offset;

			inline void SkipToNewline()
			{
				while(Buffer[Offset] != '\0')
				{
					if(Buffer[Offset] == '\n')
					{
						return;
					};

					++Offset;
				};

				return;
			};

			inline void SkipSpace()
			{
				while(Buffer[Offset] != '\0')
				{
					if(Buffer[Offset] == '\n')
					{
						++COGlineno;
						++Offset;
					}
					else if(isspace((int)Buffer[Offset]))
					{
						++Offset;
						continue;
					}
					else if(Buffer[Offset] == '#')
					{
						SkipToNewline();
					}
					else if((Buffer[Offset] == '/') && (Buffer[Offset + 1] == '/'))
					{
						SkipToNewline();
					}
					else return;
				};

				return;
			};

			inline std::string GetPunc()
			{
				SkipSpace();
				std::string RetVal;

				while(Buffer[Offset] != '\0')
				{
					if(ispunct((int)Buffer[Offset]))
					{
						RetVal.push_back(Buffer[Offset++]);
						return RetVal;
					}
					else break;
				};

				return RetVal;
			};

			inline std::string GetIdentifier()
			{
				SkipSpace();
				std::string RetVal;

				while(Buffer[Offset] != '\0')
				{
					if(isalnum((int)Buffer[Offset]) || (Buffer[Offset] == '_'))
					{
						RetVal.push_back(Buffer[Offset++]);
					}
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetFilename()
			{
				SkipSpace();
				std::string RetVal;

				while(Buffer[Offset] != '\0')
				{
					if(isalnum((int)Buffer[Offset]) || (Buffer[Offset] == '_') || (Buffer[Offset] == '-') || (Buffer[Offset] == '.') || (Buffer[Offset] == '~'))
					{
						RetVal.push_back(Buffer[Offset++]);
					}
					else break;
				};
				return RetVal;
			};

			inline std::string GetTemplate()
			{
				SkipSpace();
				std::string RetVal;

				while(Buffer[Offset] != '\0')
				{
					if(isalnum((int)Buffer[Offset]) || (Buffer[Offset] == '_') || (Buffer[Offset] == '-') || (Buffer[Offset] == '+'))
					{
						RetVal.push_back(Buffer[Offset++]);
					}
					else break;
				};

				return RetVal;
			};

			inline std::string GetString()
			{
				SkipSpace();
				std::string RetVal;

				if(Buffer[Offset] == '\"')
				{
					while((Buffer[Offset] != '\"') && (Buffer[Offset] != '\0'))
					{
						if(Buffer[Offset] == '\\')
						{
							++Offset;
							switch(Buffer[Offset])
							{
							case '\\':
								RetVal.push_back('\\');
								break;
								
							case 'n':
								RetVal.push_back('\n');
								break;

							case 't':
								RetVal.push_back('\t');
								break;

							case '\"':
								RetVal.push_back('\"');
								break;

							case '\'':
								RetVal.push_back('\'');
								break;

							default:
								{
									std::ostringstream Stream;
									Stream << "WARNING: Unidentified escape sequence \'\\" << Buffer[Offset] << "\' on line " << COGlineno << "\n";
									ConOut(Stream.str(), ConOut::Warning);
								};
								break;
							};
						}
						else RetVal.push_back(Buffer[Offset]);

						++Offset;
					};
				};

				return RetVal;
			};

			inline int GetInteger()
			{
				SkipSpace();
				if(Buffer[Offset] == '\0') return 0;
				std::string Retval;
				// Hex integer
				if(((Buffer[Offset] == '0') && ((Buffer[Offset + 1] == 'x') || (Buffer[Offset + 1] == 'X'))) || ((Buffer[Offset] == 'x') || Buffer[Offset] == 'X'))
				{
					Offset += 2;
					while(Buffer[Offset] != '\0')
					{
						if(isdigit((int)Buffer[Offset]))
						{
							Retval.push_back(Buffer[Offset++]);
							continue;
						}
						else if((Buffer[Offset] >= 'A') && (Buffer[Offset] <= 'F'))
						{
							Retval.push_back(Buffer[Offset++]);
							continue;
						}
						else if((Buffer[Offset] >= 'a') && (Buffer[Offset] <= 'f'))
						{
							Retval.push_back(Buffer[Offset++]);
							continue;
						}
						else break;
					};
					size_t Len = Retval.length();
					int *digit = new int[Len];
					int RetVal = 0;
					unsigned int Count = 0;
					unsigned int i = 0, j = 0;
					for(i = 0; i < Retval.length(); ++i)
					{
						if(Retval[i] > 0x29 && Retval[i] < 0x40)
							digit[i] = Retval[i] & 0x0F;
						else if(Retval[i] >= 'a' && Retval[i] <= 'f')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else if(Retval[i] >= 'A' && Retval[i] <= 'F')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else break;

						++Count;
					};
					j = Count - 1;
					for(i = 0; i < Count; ++i)
					{
						RetVal = RetVal | (digit[i] << (j << 2));
						--j;
					};
					return RetVal;
				}
				else // Normal integer
				{
					if(Buffer[Offset] == '-')
					{
						Retval.push_back(Buffer[Offset++]);
					};

					while(Buffer[Offset] != '\0')
					{
						if(isdigit((int)Buffer[Offset]))
						{
							Retval.push_back(Buffer[Offset++]);
							continue;
						}
						else
						{
							break;
						};
					};

					return atoi(Retval.c_str());
				};
			};

			float GetFloat()
			{
				if(Buffer[Offset] == '\0') return 0.0f;

				SkipSpace();

				std::string Retval;

				if(Buffer[Offset] == '-')
				{
					Retval.push_back(Buffer[Offset++]);
				};

				if(Buffer[Offset] == '.')
				{
					// inject 0 before decimal place..
					Retval.push_back('0');
				};

				while(Buffer[Offset] != '\0')
				{
					if(isdigit((int)Buffer[Offset]))
					{
						Retval.push_back(Buffer[Offset++]);
						continue;
					}
					else if(Buffer[Offset] == '.')
					{
						Retval.push_back(Buffer[Offset++]);
						continue;
					}
					else
					{
						break;
					};
				};

				return (float)atof(Retval.c_str());
			};
		};
	};
};

#endif // _COGPARSER_H_INCLUDED_
