/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// CONOUT.H
// - Defines the standard output model for error information and the console

#ifndef _CONOUT_H_INCLUDED_
#define _CONOUT_H_INCLUDED_

namespace Sith
{
	namespace System
	{
		class ConOut
		{
		public:
			enum Filter
			{
				Standard = 0x01,	// emits standard (stdout)
				Echo = 0x02,		// echoes commands
				Network = 0x04,		// network messages (chat, updates)

				Warning = 0x08,		// displays warnings (moderate problems)
				Error = 0x10,		// displays errors (recoverable problems)
				Critical = 0x20,	// displays critical errors (system failures)

				Pedantic = 0x40,	// generic status updates that only Jon`C cares about
			};

		private:
			struct OutputFilter
			{
				std::ostringstream& Stream;
				unsigned int FilterClass;

				OutputFilter(std::ostringstream &_Stream, unsigned int _FilterClass) : Stream(_Stream), FilterClass(_FilterClass)
				{
					return;
				};
			};

			static std::list<OutputFilter> StreamList;

		public:
			ConOut(std::string Output, Filter MessageType)
			{
				for(std::list<OutputFilter>::iterator it = StreamList.begin(); it != StreamList.end(); ++it)
				{
					if(MessageType & it->FilterClass)
					{
						(it->Stream) << Output;
					};
				};

				return;
			};

			static inline SITHERROR RegisterOutput(std::ostringstream &Output, unsigned int FilterClass)
			{
				StreamList.push_back(OutputFilter(Output, FilterClass));
				return SE_OK;
			};
		};

		class ConIn
		{
		private:
			struct InputFilter
			{
				std::ostringstream& Stream;

				InputFilter(std::ostringstream &_Stream) : Stream(_Stream)
				{
					return;
				};
			};

			static std::list<InputFilter> StreamList;

		public:
			ConIn(std::string Input)
			{
				for(std::list<InputFilter>::iterator it = StreamList.begin(); it != StreamList.end(); ++it)
				{
					(it->Stream) << Input;
				};

				return;
			};

			static inline SITHERROR RegisterInput(std::ostringstream &Input)
			{
				StreamList.push_back(InputFilter(Input));
				return SE_OK;
			};
		};
	};
};
#endif // _CONOUT_H_INCLUDED_