/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _CONSOLE_H_INCLUDED_
#define _CONSOLE_H_INCLUDED_

namespace Sith
{
	namespace System
	{
		class Console : public Sith::System::Task
		{
		private:
			std::ostringstream Stream;

		public:
			static inline SITHERROR RunScriptFromPhysicalLocation(std::string Path)
			{
				FILE *COGFile = 0;
				COGFile = fopen(Path.c_str(), "r");
				if(COGFile == 0)
				{
					ConOut("ERROR: Failed to open COG file, \'" + Path + "\'.\n", ConOut::Error);
					return SE_FILE_NOT_FOUND;
				};

				std::stringstream Buffer;

				char Temp;
				while(true)
				{
					fread(&Temp, sizeof(char), 1, COGFile);
					if(!feof(COGFile))
					{
						Buffer << Temp;
					}
					else
					{
						break;
					};
				};

				fclose(COGFile);

				COG::StackMachine theMachine;
				theMachine.Parse(Buffer.str());

				if(theMachine.Heap.TranslateLocalHeapName("startup") != -1)
				{
					int Addr = theMachine.Heap.Get(theMachine.Heap.TranslateLocalHeapName("startup"), -1).GetAddr();
					theMachine.Execute(Addr, COG::VMType(), -1, COG::VMType(), COG::VMType(), COG::VMType(), COG::VMType(), COG::VMType());
				}
				else
				{
					theMachine.Execute(0, COG::VMType(), -1, COG::VMType(), COG::VMType(), COG::VMType(), COG::VMType(), COG::VMType());
				};

				return SE_OK;
			};

			// COG VM BINDINGS----
			static std::string StringRegister;

			COGLIBAPI Print(COG::VMType *InString)
			{
				ConOut(InString->GetString(), ConOut::Standard);
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI PrintFlex(COG::VMType *InFloat)
			{
				std::ostringstream Stream;
				Stream << InFloat->GetFloat() << "\n";
				ConOut(Stream.str(), ConOut::Standard);
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI PrintInt(COG::VMType *InInt)
			{
				std::ostringstream Stream;
				Stream << InInt->GetInt() << "\n";
				ConOut(Stream.str(), ConOut::Standard);
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI PrintVector(COG::VMType* InVector)
			{
				std::ostringstream Stream;
				Stream << "<x." << InVector->GetVector().x << " y." << InVector->GetVector().y << " z." << InVector->GetVector().z << ">\n";
				ConOut(Stream.str(), ConOut::Standard);
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI run(COG::VMType *InString)
			{
				RunScriptFromPhysicalLocation(InString->GetString());
				return new COG::VMType(COG::VMTC_VOID);
			};

			Console();
			~Console();

			SITHERROR Process();
		};
	};
};

#endif // _CONSOLE_H_INCLUDED_