/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "dx_render.h"
#include "dx_hwbm.h"

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Create(Sith::Win32::HBMGeometry *GeoData)
{
	unsigned int Levels;

	DWORD Usage = 0;
	D3DFORMAT Format;
    
	if(GeoData->Pow2 == false)
	{
		if(Sith::Win32::Renderer::Get().IsBMSupported())
		{
			Levels = 1; // BMs cannot be mip-mapped
		}
		else
		{
			// Device does not support BMs.
			// Renderer won't even try to use this resource; just display a warning and be done with it.
			Sith::ConOut("WARNING: Device does not support BMs, but loading one was attempted.\n", ConOut::Warning);
			return SE_GENERAL_WARNING;
		};
	}
	else if(GeoData->Pow2 == true)
	{
		Levels = 0; // Use hardware mipmapping
		Usage |= D3DUSAGE_AUTOGENMIPMAP;
	};

	if(GeoData->bitdepth == HBMF_A8R8G8B8)
	{
		Format = D3DFMT_A8R8G8B8;
	}
	else if(GeoData->bitdepth == HBMF_X8R8G8B8)
	{
		Format = D3DFMT_X8R8G8B8;
	}
	else if(GeoData->bitdepth == HBMF_R8G8B8)
	{
		Format = D3DFMT_R8G8B8;
	}
	else if(GeoData->bitdepth == HBMF_A1R5G5B5)
	{
		Format = D3DFMT_A1R5G5B5;
	}
	else if(GeoData->bitdepth = HBMF_R5G6B5)
	{
		Format = D3DFMT_R5G6B5;
	}
	else
	{
		Sith::ConOut("ERROR: Attempted to instantiate a hardware bitmap with an unsupported bitdepth.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};

	HRESULT ErrorCode = Sith::Win32::Renderer::Get().GetDevice()->CreateTexture(GeoData->xdim, GeoData->ydim, Levels, Usage, Format, D3DPOOL_MANAGED, &TextureData, NULL);

	if(FAILED(ErrorCode))
	{
		Sith::ConOut("ERROR: Failed to create hardware bitmap.\n", ConOut::Error);

		GeoData->bitdepth = HBMF_BADFORMAT;
		GeoData->Pow2 = false;
		GeoData->xdim = 0;
		GeoData->ydim = 0;

		return SE_GENERAL_ERROR;
	};

	x_res = GeoData->xdim;
	y_res = GeoData->ydim;
	bitdepth = GeoData->bitdepth;
	Pow2 = GeoData->Pow2;

	return SE_OK;
};

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Lock(Sith::Win32::HBMGeometry *GeoDataBuffer)
{
	if(Locked)
	{
		ConOut("WARNING: Attempted to lock a hardware bitmap, but it's already locked.\n", ConOut::Warning);
		return SE_GENERAL_WARNING;
	};

	if(TextureData)
	{
		D3DLOCKED_RECT LRect;
		HRESULT ErrorCode = TextureData->LockRect(0, &LRect, NULL, D3DLOCK_DISCARD);
		if(FAILED(ErrorCode))
		{
			ConOut("ERROR: Failed to lock hardware bitmap resource.\n", ConOut::Error);
			ZeroMemory(GeoDataBuffer, sizeof(HBMGeometry));
			return SE_GENERAL_ERROR;
		};

		GeoDataBuffer->bitdata = LRect.pBits;
		GeoDataBuffer->bitdepth = bitdepth;
		GeoDataBuffer->xdim = x_res;
		GeoDataBuffer->ydim = y_res;
		GeoDataBuffer->Pow2 = Pow2;

		Locked = true;
		
		return SE_OK;
	}
	else
	{
		ConOut("ERROR: Attempted to lock a hardware bitmap for writing. Most likely a Create call failed but it wasn't checked.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};
};

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Unlock()
{
	if(!Locked)
	{
		ConOut("WARNING: Attempted to unlock a hardware bitmap, but it isn't locked.\n", ConOut::Warning);
		return SE_GENERAL_WARNING;
	};

	if(!TextureData)
	{
		ConOut("ERROR: Attempted to unlock a hardware bitmap but it doesn't exist. Most likely a Create call failed but was never checked.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};

	HRESULT ErrorCode = TextureData->UnlockRect(0);
	if(FAILED(ErrorCode))
	{
		ConOut("ERROR: Attempt to unlock a hardware bitmap failed. Perhaps it doesn't exist, or it was never locked?\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};
	return SE_OK;
};