/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _DX_RENDER_H_INCLUDED_
#define _DX_RENDER_H_INCLUDED_

#include "dx_hwbm.h"

namespace Sith
{
	namespace Win32
	{
		struct D3DUIVertex
		{
			float x, y, z, rhw;
			DWORD color;
			float tu, tv;
		};

		struct D3DJKLVertex
		{
			float x, y, z;
			DWORD color;
			float tu, tv;
		};

		extern const DWORD D3DUIVertex_FVF;
		extern const DWORD D3DJKLVertex_FVF;

		class Renderer : public Task
		{
			friend class Sith::KernelProc;
		private:
			WNDCLASSEX wc;
			HWND hMainWnd;
			MSG Msg;

			static Renderer* StaticInterface;

			Renderer();
			~Renderer();

			static LRESULT CALLBACK RenderWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

			SITHERROR Startup();
			SITHERROR Process();

			SITHERROR DrawUIWidget(UI::BaseWidget& Widget);

			// ------ DIRECTX GRAPHICS ------
			IDirect3D9* lpD3D;
			IDirect3DDevice9* Device;
			HRESULT LastError;

			D3DPRESENT_PARAMETERS present_parameters;
			D3DCAPS9 device_capabilities;

			bool ClearEveryFrame;
			DWORD ClearColor;

			bool DeviceSupportsBMs;	// Does device support hardware acceleration and scaling for non-pow2 GUI BMs? If not, this causes horrible problems.
			// Virtually all modern consumer hardware supports conditional nonpow2s so I'm not too worried about it. Well... sorta worried. Thus the bool.

		public:
			static inline Renderer& Get()
			{
				return *StaticInterface;
			};

			inline IDirect3DDevice9* GetDevice()
			{
				return Device;
			};

			inline bool IsBMSupported()
			{
				return DeviceSupportsBMs;
			};
		};
	};
};

#endif // _DX_RENDER_H_INCLUDED_