/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _GAMESTATE_H_INCLUDED_
#define _GAMESTATE_H_INCLUDED_

#include "invfile.h"

namespace Sith
{
	namespace Gameplay
	{
		class GameState : public System::Task
		{
			friend class System::KernelProc;
		private:
			GameState();
			~GameState();

			static GameState* StaticInterface;

			SITHERROR Process();

			std::vector<Gameplay::Thing> Walkplayers;

			bool GameIsActive;

			bool DoPlayerPhysics;

		public:
			static inline GameState& Get()
			{
				return *StaticInterface;
			};

			int CreateThing(Gameplay::Template& tpl, Math::Vector Position, int SectorNum, bool DoScripts=true);
			void DestroyThing(int ThingNum);

			Math::Vector VecSurfaceCenter(Res::Level::SURFACE* Surface);
			void GenSurfVis(int Sector, Math::Vector Pos, Math::Vector LookVector, std::vector<Res::Level::SURFACE*> &SurfVis);
			void Activated(int ActivateThingNum);

			int GetCameraSector(Sith::Math::Vector Pos, int CurSector);

			std::vector<Res::Inventory::Bin> BinList;

			// Temporary:
			COGLIBAPI isMulti()
			{
				return new COG::VMType(false);
			};

			COGLIBAPI isServer()
			{
				return new COG::VMType(false);
			};

			COGLIBAPI map(COG::VMType* MapName);

			COGLIBAPI noclip()
			{
				if(Get().DoPlayerPhysics) Get().DoPlayerPhysics = false;
				else Get().DoPlayerPhysics = true;
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI rand()
			{
				return new COG::VMType(Sith::Math::Random::genrand());
			};

			COGLIBAPI SurfaceCenter(COG::VMType* surface);

			COGLIBAPI COGCreateThing(COG::VMType* tpl, COG::VMType* thingpos);
			COGLIBAPI COGCreateThingAtPos(COG::VMType* tpl, COG::VMType* sector, COG::VMType* pos, COG::VMType* ori);
			COGLIBAPI COGDestroyThing(COG::VMType* thingnum);
			COGLIBAPI findsectoratpos(COG::VMType* Pos);

			COGLIBAPI Warp(COG::VMType* wplyr);

			COGLIBAPI Tele(COG::VMType* surface);

			COGLIBAPI GetDifficulty()
			{
				return new COG::VMType(COG::CogLib::GetSetting("game::difficulty"));
			};

			COGLIBAPI GetInv(COG::VMType* Player, COG::VMType* Bin)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					return new COG::VMType(Get().BinList[Bin->GetInt()].Value);
				}
				return new COG::VMType(-1);
			};

			COGLIBAPI SetInv(COG::VMType* Player, COG::VMType* Bin, COG::VMType* Value)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					float NewValue;
					if (Value->GetFloat() > Get().BinList[Bin->GetInt()].Max)
					{
						NewValue = Get().BinList[Bin->GetInt()].Max;
					}
					else if (Value->GetFloat() < Get().BinList[Bin->GetInt()].Min)
					{
						NewValue = Get().BinList[Bin->GetInt()].Min;
					}
					else
					{
						NewValue = Value->GetFloat();
					}
					Get().BinList[Bin->GetInt()].Value = NewValue;
				}
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI ChangeInv(COG::VMType* Player, COG::VMType* Bin, COG::VMType* Delta)
			{
				// for now ignore player
				if((Bin->GetInt() >= 0) && (Bin->GetInt() < Get().BinList.size()))
				{
					float NewValue = Get().BinList[Bin->GetInt()].Value + Delta->GetInt();
					if(NewValue > Get().BinList[Bin->GetInt()].Max)
					{
						NewValue = Get().BinList[Bin->GetInt()].Max;
					}
					else if(NewValue < Get().BinList[Bin->GetInt()].Min)
					{
						NewValue = Get().BinList[Bin->GetInt()].Min;
					};

					Get().BinList[Bin->GetInt()].Value = NewValue;
				};

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI GetInvMax(COG::VMType* Player, COG::VMType* Bin)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					return new COG::VMType(Get().BinList[Bin->GetInt()].Max);
				}
				return new COG::VMType(-1);
			};

			COGLIBAPI GetInvMin(COG::VMType* Player, COG::VMType* Bin)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					return new COG::VMType(Get().BinList[Bin->GetInt()].Min);
				}
				return new COG::VMType(-1);
			};

			COGLIBAPI GetInvAvailable(COG::VMType* Player, COG::VMType* Bin)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					return new COG::VMType(Get().BinList[Bin->GetInt()].Available);
				}
				return new COG::VMType(-1);
			};

			COGLIBAPI GetInvActivated(COG::VMType* Player, COG::VMType* Bin)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					return new COG::VMType(Get().BinList[Bin->GetInt()].Activated);
				}
				return new COG::VMType(-1);
			};

			COGLIBAPI SetInvAvailable(COG::VMType* Player, COG::VMType* Bin, COG::VMType* Value)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					Get().BinList[Bin->GetInt()].Available = Value->GetBool();
				}
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI SetInvActivated(COG::VMType* Player, COG::VMType* Bin, COG::VMType* Value)
			{
				// for now ignore player
				if (Bin->GetInt() < Get().BinList.size())
				{
					Get().BinList[Bin->GetInt()].Activated = Value->GetBool();
				}
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI jkprintunistring(COG::VMType* Dest, COG::VMType* Str)
			{
				std::ostringstream Stream;
				Stream << "COG_";

				int Num = Str->GetInt();
				
				if(Num < 10000) Stream << 0;
				if(Num < 1000) Stream << 0;
				if(Num < 100) Stream << 0;
				if(Num < 10) Stream << 0;
				Stream << Num;

				ConOut(COG::CogLib::GetUniString(Stream.str()), ConOut::Standard);

				return new COG::VMType();
			};

			COGLIBAPI cogcreatebackpack(COG::VMType* PlayerThing)
			{
				// TODO: Something
				return new COG::VMType();
			};

			COGLIBAPI cogstopkey(COG::VMType* Thing, COG::VMType* TrackID, COG::VMType* Unk)
			{
				// TODO: Something
				return new COG::VMType();
			};

			COGLIBAPI cogplaykey(COG::VMType* Thing, COG::VMType* Key, COG::VMType* Unk, COG::VMType* Flags)
			{
				// TODO: What
				return new COG::VMType();
			};

			COGLIBAPI cogstoppovkey(COG::VMType* Thing, COG::VMType* TrackID, COG::VMType* Unk)
			{
				// TODO: Implement
				return new COG::VMType();
			};

			COGLIBAPI cogplaypovkey(COG::VMType* Thing, COG::VMType* Key, COG::VMType* Unk, COG::VMType* Flags)
			{
				// TODO: Write this
				return new COG::VMType();
			};

			COGLIBAPI coggetcurweapon(COG::VMType* Thing)
			{
				// TODO: Meh
				return new COG::VMType(1);
			};

			COGLIBAPI cogapplyforce(COG::VMType* Thing, COG::VMType* Force)
			{
				if(WorldState::Get().CurrentLevel.isValid())
				{
					if((Thing->GetInt() >= 0) && (Thing->GetInt() < WorldState::Get().ThingList.size()))
					{
						Gameplay::Thing* curThing = &WorldState::Get().ThingList[Thing->GetInt()];

						curThing->Velocity = Math::VectorScale(Force->GetVector(), 1 / curThing->Mass);
					};
				};

				return new COG::VMType();
			};

			COGLIBAPI dumb(COG::VMType* null, COG::VMType* pointer)
			{
				// HACK
				return new COG::VMType(1);
			};

			COGLIBAPI coggetcurrentcamera()
			{
				return new COG::VMType(0);
			};
		};
	};
};
#endif // _GAMESTATE_H_INCLUDED_