/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "invfile.h"

Sith::Res::Inventory::Inventory (std::string FileName) : Sith::VFS::Resource(FileName) {
	return;
}

Sith::SITHERROR Sith::Res::Inventory::LoadFromFile(const char *FileBuffer) {
	CurrentOffset = 0;
	LineNumber = 0;
	EndOfStream = false;
	Buffer = FileBuffer;

	BufferLen = strlen(FileBuffer);

	std::string Name;
	while (1) {
		while(!isalpha(Buffer[CurrentOffset])) { SkipComment(); SkipWhitespace(); }
		Bin NewBin;
		Name = GetContiguousToken();
		if (Name == "end" || EndOfStream) {
			break;
		}
		// we have an entry
		NewBin.Name = Name;
		NewBin.ID = GetIntegerToken();
		NewBin.Value = NewBin.Min = GetFloatToken();
		NewBin.Max = GetFloatToken();
		NewBin.Flags = GetIntegerToken();
		// is there an attached cog?
		int OldOffset = CurrentOffset;
		if (GetWordToken() != "cog") {
			CurrentOffset = OldOffset;
			NewBin.COG.clear();
		} else {
			if (GetPuncTokenChar() != '=') {
				return SE_FILE_CORRUPT;
			}
			NewBin.COG = GetFilenameToken();
		}

		NewBin.Used = true;

		// if the ids aren't contiguous, add some empty ones; this is so indexing is much faster
		if (NewBin.ID >= BinList.size()) {
			BinList.resize(NewBin.ID + 1);
		}

		BinList[NewBin.ID] = NewBin;
	}

	return SE_OK;
}
