/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "vmproc.h"

Sith::Res::Level::Level(std::string SystemName) : Sith::VFS::Resource(SystemName)
{
	return;
};

Sith::SITHERROR Sith::Res::Level::ParseHeader()
{
	std::string TokenBuffer;
	std::string Var;

	while(CurrentOffset < BufferLen)
	{
		Var.clear();
		while(true)
		{
			TokenBuffer = GetWordToken();

			if(TokenBuffer == "section")
			{
				CurrentOffset -= 7;
				return SE_OK;
			};

			if(TokenBuffer.length() > 0)
			{
				Var += " ";
				Var += TokenBuffer;
			}
			else break;
		};

		// Now I have a machine gun. >:)

		if(Var == " version")
		{
			Header.Version = GetIntegerToken();
		}
		else if(Var == " world gravity")
		{
			Header.Gravity = GetFloatToken();
		}
		else if(Var == " ceiling sky z")
		{
			Header.CeilingSkyZ = GetFloatToken();
		}
		else if(Var == " horizon distance")
		{
			Header.HorizonDistance = GetFloatToken();
		}
		else if(Var == " horizon pixels per rev")
		{
			Header.HorizonPixelsPerRev = GetFloatToken();
		}
		else if(Var == " horizon sky offset")
		{
			Header.HorizonSkyOffset[0] = GetFloatToken();
			Header.HorizonSkyOffset[1] = GetFloatToken();
		}
		else if(Var == " ceiling sky offset")
		{
			Header.CeilingSkyOffset[0] = GetFloatToken();
			Header.CeilingSkyOffset[1] = GetFloatToken();
		}
		else if(Var == " mipmap distances")
		{
			Header.MipMapDistances[0] = GetFloatToken();
			Header.MipMapDistances[1] = GetFloatToken();
			Header.MipMapDistances[2] = GetFloatToken();
			Header.MipMapDistances[3] = GetFloatToken();
		}
		else if(Var == " lod distances")
		{
			Header.LODDistances[0] = GetFloatToken();
			Header.LODDistances[1] = GetFloatToken();
			Header.LODDistances[2] = GetFloatToken();
			Header.LODDistances[3] = GetFloatToken();
		}
		else if(Var == " perspective distance")
		{
			Header.PerspectiveDistance = GetFloatToken();
		}
		else if(Var == " gouraud distance")
		{
			Header.GouraudDistance = GetFloatToken();
		}
		else
		{
			ConOut("WARNING: Level header contains unrecognized elements.\n", ConOut::Warning);
			return SE_OK;
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseMaterials()
{
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "materials") return SE_FILE_CORRUPT;

	GetIntegerToken(); // num of mats, only it isn't correct so it's 
	// pretty damn useless, dammit

	std::string LALR;
	LALR.clear();

	while(CurrentOffset < BufferLen)
	{
		// lookahead
		LALR = GetWordToken();
		if(LALR == "end") break;
		else
		{
			CurrentOffset -= LALR.length();
			LALR.clear();
		};

		TextureList.push_back(MATDATA());

		GetIntegerToken(); // Number
		GetPuncToken(); // Colon

		TextureList.back().TextureName = GetFilenameToken();
		TextureList.back().XScale = GetFloatToken();
		TextureList.back().YScale = GetFloatToken();
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseGeoresource()
{
	int i;

	// Parse colormaps subsection

	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "colormaps") return SE_FILE_CORRUPT;

	int Num = GetIntegerToken();

	for(i = 0; i < Num; ++i)
	{
		GetIntegerToken(); // Number
		GetPuncToken(); // Colon

		ColormapList.push_back(GetFilenameToken());
	};

	// Parse vertices subsection

	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "vertices") return SE_FILE_CORRUPT;

	Num = GetIntegerToken();

	for(i = 0; i < Num; ++i)
	{
		GeovertexList.push_back(GEOVERTEX());

		GetIntegerToken(); // num
		GetPuncToken(); // colon

		GeovertexList.back().X = GetFloatToken();
		GeovertexList.back().Y = GetFloatToken();
		GeovertexList.back().Z = GetFloatToken();
	};

	// Parse texturevertex subsection

	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "texture") return SE_FILE_CORRUPT;
	if(GetWordToken() != "vertices") return SE_FILE_CORRUPT;

	Num = GetIntegerToken();

	for(i = 0; i < Num; ++i)
	{
		TexvertexList.push_back(TEXVERTEX());

		GetIntegerToken(); // num
		GetPuncToken(); // colon

		TexvertexList.back().U = GetFloatToken();
		TexvertexList.back().V = GetFloatToken();
	};

	// Parse adjoins subsection
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "adjoins") return SE_FILE_CORRUPT;

	Num = GetIntegerToken();

	for(i = 0; i < Num; ++i)
	{
		AdjoinList.push_back(ADJOIN());

		GetIntegerToken(); // num
		GetPuncToken(); // colon

		AdjoinList.back().Flags = GetIntegerToken();
		AdjoinList.back().Mirror = GetIntegerToken();
		AdjoinList.back().Dist = GetFloatToken();
	};

	// Parse surfaces subsection
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "surfaces") return SE_FILE_CORRUPT;

	Num = GetIntegerToken();

	for(i = 0; i < Num; ++i)
	{
		SurfaceList.push_back(SURFACE());

		SurfaceList.back().SurfaceNumber = GetIntegerToken(); // num
		GetPuncToken(); // colon

		SurfaceList.back().TextureIndex = GetIntegerToken();

		SurfaceList.back().SurfaceFlags = GetIntegerToken();
		SurfaceList.back().FaceFlags = GetIntegerToken();
		SurfaceList.back().GeoMode = GetIntegerToken();
		SurfaceList.back().LightMode = GetIntegerToken();
		SurfaceList.back().TextureMode = GetIntegerToken();

		SurfaceList.back().AdjoinIndex = GetIntegerToken();

		SurfaceList.back().ExtraLight = GetFloatToken();

		SurfaceList.back().VertexCount = GetIntegerToken();

		for(int j = 0; j < SurfaceList.back().VertexCount; ++j)
		{
			SurfaceList.back().VertexList.push_back(SURFACEVERTEX());

			SurfaceList.back().VertexList.back().GeovertexIndex = GetIntegerToken();

			GetPuncToken(); // comma

			SurfaceList.back().VertexList.back().TexvertexIndex = GetIntegerToken();

		};

		std::vector<float> Intensities;
		unsigned int NewLinePos = FindNewLine() - 2;
		SkipWhitespace();
		Intensities.clear();
		while (CurrentOffset < NewLinePos) {
			Intensities.push_back (GetFloatToken());
		}
		if (Intensities.size() == SurfaceList.back().VertexCount) {				// JK
			for (j = 0; j < SurfaceList.back().VertexCount; ++j) {
				SurfaceList.back().VertexList[j].LightChroma = SurfaceList.back().VertexList[j].LightR = SurfaceList.back().VertexList[j].LightG = SurfaceList.back().VertexList[j].LightB = Intensities[j];
			}
		} else if (Intensities.size() == (SurfaceList.back().VertexCount * 4)) {	// MOTS
			for (j = 0; j < SurfaceList.back().VertexCount; ++j) {
				SurfaceList.back().VertexList[j].LightChroma = Intensities[j * 4];
				SurfaceList.back().VertexList[j].LightR = Intensities[j * 4 + 1];
				SurfaceList.back().VertexList[j].LightG = Intensities[j * 4 + 2];
				SurfaceList.back().VertexList[j].LightB = Intensities[j * 4 + 3];
			}
		} else {
			std::ostringstream Stream;
			Stream << "ERROR: Incorrect amount of intensities: " << Intensities.size() << " for " << SurfaceList.back().VertexCount << " vertices, " << Intensities.back() << ".\n";
			ConOut (Stream.str(), ConOut::Error);
			return SE_FILE_CORRUPT;
		}

		/*for(j = 0; j < SurfaceList.back().VertexCount; ++j)
		{
			SurfaceList.back().VertexList[j].LightChroma = GetFloatToken();

			// TODO: MOTS code even though MOTS sucks

			SurfaceList.back().VertexList[j].LightChroma = SurfaceList.back().VertexList[j].LightChroma;
			SurfaceList.back().VertexList[j].LightChroma = SurfaceList.back().VertexList[j].LightChroma;
			SurfaceList.back().VertexList[j].LightChroma = SurfaceList.back().VertexList[j].LightChroma;
		};*/
	};

	// Surface normals subsection
	for(i = 0; i < SurfaceList.size(); ++i)
	{
		GetIntegerToken(); // num
		GetPuncToken(); // colon

		SurfaceList[i].Normal.X = GetFloatToken();
		SurfaceList[i].Normal.Y = GetFloatToken();
		SurfaceList[i].Normal.Z = GetFloatToken();
	};

	// end of section of horrors
	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseSector()
{
	std::string TokenBuffer;
	std::string TokenBuffer2;
	std::string Var;
	bool isValid = false;

	while(CurrentOffset < BufferLen)
	{
		Var.clear();
		while(true)
		{
			TokenBuffer.clear();
			TokenBuffer = GetWordToken();

			if((TokenBuffer == "sector") && (isValid == true))
			{
				CurrentOffset -= 6;
				return SE_OK;
			}
			else if(TokenBuffer == "section")
			{
				// end of section, aborticus
				CurrentOffset -= 7;
				return SE_OK;
			};

			if(TokenBuffer.length() > 0)
			{
				// throw back for a second
				CurrentOffset -= TokenBuffer.length();
				TokenBuffer2 = GetFilenameToken();
				if(TokenBuffer2.length() > TokenBuffer.length())
				{
					// it's a filename, throw back all the way
					CurrentOffset -= TokenBuffer2.length();
					break;
				}
				else
				{
					// no, get it back
					Var += " ";
					Var += TokenBuffer;
				};
			}
			else break;
		};

		// Now I have a machine gun. >:)

		if(Var == " sector")
		{
			SectorList.back().SectorNumber = GetIntegerToken();
			isValid = true;
		}
		else if(Var == " flags")
		{
			SectorList.back().SectorFlags = GetIntegerToken();
		}
		else if(Var == " ambient light")
		{
			SectorList.back().AmbientLight = GetFloatToken();
		}
		else if(Var == " extra light")
		{
			SectorList.back().ExtraLight = GetFloatToken();
		}
		else if(Var == " colormap")
		{
			SectorList.back().ColormapIndex = GetIntegerToken();
		}
		else if(Var == " tint")
		{
			SectorList.back().Tint[0] = GetFloatToken();
			SectorList.back().Tint[1] = GetFloatToken();
			SectorList.back().Tint[2] = GetFloatToken();
		}
		else if(Var == " boundbox")
		{
			SectorList.back().BoundBox[0] = GetFloatToken();
			SectorList.back().BoundBox[1] = GetFloatToken();
			SectorList.back().BoundBox[2] = GetFloatToken();
			SectorList.back().BoundBox[3] = GetFloatToken();
			SectorList.back().BoundBox[4] = GetFloatToken();
			SectorList.back().BoundBox[5] = GetFloatToken();
		}
		else if(Var == " collidebox")
		{
			SectorList.back().CollideBox[0] = GetFloatToken();
			SectorList.back().CollideBox[1] = GetFloatToken();
			SectorList.back().CollideBox[2] = GetFloatToken();
			SectorList.back().CollideBox[3] = GetFloatToken();
			SectorList.back().CollideBox[4] = GetFloatToken();
			SectorList.back().CollideBox[5] = GetFloatToken();
		}
		else if(Var == " center")
		{
			SectorList.back().Center.x = GetFloatToken();
			SectorList.back().Center.y = GetFloatToken();
			SectorList.back().Center.z = GetFloatToken();
		}
		else if(Var == " radius")
		{
			SectorList.back().Radius = GetFloatToken();
		}
		else if(Var == " vertices")
		{
			SectorList.back().VertexCount = GetIntegerToken();
			for(int i = 0; i < SectorList.back().VertexCount; ++i)
			{
				GetIntegerToken(); // num
				GetPuncToken(); // colon
				SectorList.back().VertexIndexList.push_back(GetIntegerToken());
			};
		}
		else if(Var == " surfaces")
		{
			SectorList.back().FirstSurfaceIndex = GetIntegerToken();
			SectorList.back().SurfaceCount = GetIntegerToken();
		}
		else if(Var == " sound")
		{
			SectorList.back().SoundName = GetFilenameToken();
			SectorList.back().SoundVolume = GetFloatToken();
		}
		else
		{
			std::ostringstream Stream;
			Stream << "WARNING: Sector " << (SectorList.size() - 1) << " contains unrecognized element \'" + Var + "\'.\n";
			ConOut(Stream.str(), ConOut::Warning);
			return SE_OK;
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseSectors()
{
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "sectors") return SE_FILE_CORRUPT;

	int NumSectors = GetIntegerToken();

	for(int i = 0; i < NumSectors; ++i)
	{
		SectorList.push_back(SECTOR());
		if(Failed(ParseSector()))
		{
			std::ostringstream Stream;
			Stream << "ERROR: Failed to parse sector " << SectorList.size() << ".\n";
			ConOut(Stream.str(), ConOut::Error);
			return SE_FILE_CORRUPT;
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseLights()
{
	if(GetWordToken() != "editor") return SE_FILE_CORRUPT;
	if(GetWordToken() != "lights") return SE_FILE_CORRUPT;

	int NumLights = GetIntegerToken();

	for(int i = 0; i < NumLights; ++i)
	{
		LightList.push_back(LIGHT());

		GetIntegerToken(); // Num
		GetPuncToken(); // Colon

		LightList.back().Type = GetIntegerToken();

		LightList.back().X = GetFloatToken();
		LightList.back().Y = GetFloatToken();
		LightList.back().Z = GetFloatToken();
		LightList.back().Intensity = GetFloatToken();
		LightList.back().Range = GetFloatToken();
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseCogs()
{
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "cogs") return SE_FILE_CORRUPT;

	int NumCogs = GetIntegerToken();

	std::string COGFileName;

	while(CurrentOffset < BufferLen)
	{
		if(GetWordToken() == "end")
		{
			return SE_OK;
		};

		COG::StackMachine NewScript;
		VFS::WisePointer<Res::COGScript> NewPtr;

		GetIntegerToken(); // num
		GetPuncToken(); // colon

		COGFileName = GetFilenameToken();

		NewPtr = VFS::ResourceManager::Get().GetScript(COGFileName);
		if(NewPtr.isValid())
		{
			NewScript = NewPtr->GetScript();

			int NumArgs = NewScript.Heap.GetConfigurableValueCount();
			for(int j = 0; j < NumArgs; ++j)
			{
				int Arg = NewScript.Heap.GetConfigurableValue(j);
				COG::VMTypeCode TypeC = NewScript.Heap.Get(Arg, -1).GetType();
				switch(TypeC)
				{
				case COG::VMTC_KEYFRAME:
				case COG::VMTC_COG:
				case COG::VMTC_AI:
				case COG::VMTC_MATERIAL:
				case COG::VMTC_MODEL:
				case COG::VMTC_SOUND:
					NewScript.Heap.Set(Arg, -1, COG::VMType(GetFilenameToken()));
					break;
				case COG::VMTC_TEMPLATE:
					// TODO: Replace with GetTemplateToken();
					NewScript.Heap.Set(Arg, -1, COG::VMType(GetTemplateToken()));
					break;
				case COG::VMTC_STRING:
					break;

				case COG::VMTC_VOID:
					break;

				case COG::VMTC_BOOL:
					if(GetWordToken() == "true")
					{
						NewScript.Heap.Set(Arg, -1, COG::VMType(true));
					}
					else
					{
						NewScript.Heap.Set(Arg, -1, COG::VMType(false));
					};
					break;

				case COG::VMTC_DOUBLE:
				case COG::VMTC_FLEX:
				case COG::VMTC_FLOAT:
					NewScript.Heap.Set(Arg, -1, COG::VMType(GetFloatToken()));
					break;

				case COG::VMTC_VECTOR:
					{
						float X, Y, Z;
						GetPuncToken(); // (
						X = GetFloatToken();
						GetPuncToken(); // /
						Y = GetFloatToken();
						GetPuncToken(); // /
						Z = GetFloatToken();
						GetPuncToken(); // )
						NewScript.Heap.Set(Arg, -1, COG::VMType(Math::Vector(X, Y, Z)));
					};
					break;

				default:	// int-types
					NewScript.Heap.Set(Arg, -1, COG::VMType(TypeC, GetIntegerToken(), -1));
					break;
				};
			};

			// Done loading, skip to start of next script
			SkipToNewLine();

			// Done filling it, I think.
			// Add it to the cog table
			Cogs.push_back(NewScript);
		}
		else
		{
			ConOut("WARNING: Failed to load script \'" + COGFileName + "\'.\n", ConOut::Warning);
			// Skip to next cog script, I guess....

			SkipToNewLine();
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::ParseTemplates()
{
	unsigned int WorldTemplates, OldLineNumber;
	std::string Name, ParentName, ParamName, EnumValue;
	Sith::Gameplay::Template *TemplatePtr;

	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "templates") return SE_FILE_CORRUPT;

	WorldTemplates = GetIntegerToken();
	Templates.resize (WorldTemplates);

	std::vector<Sith::Gameplay::Template>::iterator it;
	for (it = Templates.begin(); it != Templates.end(); ++it)
	{
		Name = GetTemplateToken();
		ParentName = GetTemplateToken();
		// standard jk requires dependencies to be placed before, so this should be fine
		// might be nice to do all the checking at the end but that would take longer
		// Locate the parent if necessary and copy its data over
		if (ParentName != "none")
		{
			TemplatePtr = FindTemplate(ParentName);
			if(TemplatePtr != NULL)
			{
				it->CopyTemplate (TemplatePtr);
			}
			else
			{
				ConOut ("ERROR: Could not load parent template " + ParentName + " for " + Name + ".\n", ConOut::Error);
				return SE_FILE_CORRUPT;
			}
		}
		it->TemplateName = Name;
		//ConOut ("DEBUG: " + Name + " - " + ParentName + "\n", ConOut::Echo);

		OldLineNumber = LineNumber;
		// Now loop through all the parameters until we hit a newline
		while (OldLineNumber == LineNumber)
		{
			ParamName = GetTemplateParamToken();

			if (Buffer[CurrentOffset++] != '=')
			{
				//ConOut("ERROR: Token after parameter name \"" + ParamName + "\" was not '='.\n", ConOut::Error);
				//return SE_FILE_CORRUPT;
				CurrentOffset -= ParamName.length() + 1;
				break;	// this is a hack, the line code doesn't work otherwise wtf
			}
			if (ParamName == "thingflags")
			{
				it->ThingFlags = GetIntegerToken();
			}
			else if (ParamName == "typeflags")
			{
				it->TypeFlags = GetIntegerToken();
			}
			else if (ParamName == "type")
			{
				std::string newtype = GetWordToken();
				
				if(newtype == "actor") it->Type = Gameplay::Template::TT_ACTOR;
				else if(newtype == "camera") it->Type = Gameplay::Template::TT_CAMERA;
				else if(newtype == "cog") it->Type = Gameplay::Template::TT_COG;
				else if(newtype == "corpse") it->Type = Gameplay::Template::TT_CORPSE;
				else if(newtype == "debris") it->Type = Gameplay::Template::TT_DEBRIS;
				else if(newtype == "explosion") it->Type = Gameplay::Template::TT_EXPLOSION;
				else if(newtype == "ghost") it->Type = Gameplay::Template::TT_GHOST;
				else if(newtype == "item") it->Type = Gameplay::Template::TT_ITEM;
				else if(newtype == "particle") it->Type = Gameplay::Template::TT_PARTICLE;
				else if(newtype == "player") it->Type = Gameplay::Template::TT_PLAYER;
				else if(newtype == "weapon") it->Type = Gameplay::Template::TT_WEAPON;
				else it->Type = Gameplay::Template::TT_FREE;
			}
			else if (ParamName == "collide")
			{
				unsigned int IntEnumValue = GetIntegerToken();
				if (IntEnumValue == 0)
				{
					it->Collide = Sith::Gameplay::Template::CT_GHOST;
				}
				else if (IntEnumValue == 1 || IntEnumValue == 2)
				{
					it->Collide = Sith::Gameplay::Template::CT_SPHERE;
				}
				else if (IntEnumValue == 3) 
				{
					it->Collide = Sith::Gameplay::Template::CT_POLY;
				}
				else 
				{
					ConOut ("ERROR: Invalid collide.", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else if (ParamName == "move")
			{
				EnumValue = GetWordToken();
				if (EnumValue == "path")
					it->Move = Sith::Gameplay::Template::MT_PATH;
				else if (EnumValue == "physics")
					it->Move = Sith::Gameplay::Template::MT_PHYSICS;
				else if (EnumValue == "none")
					it->Move = Sith::Gameplay::Template::MT_NONE;
				else
				{
					ConOut ("ERROR: Invalid move.", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else if (ParamName == "movesize") 
			{
				it->MoveSize = GetFloatToken();
			} 
			else if (ParamName == "size")
			{
				it->Size = GetFloatToken();
			} 
			else if (ParamName == "airdrag") 
			{
				it->AirDrag = GetFloatToken();
			}
			else if (ParamName == "staticdrag")
			{
				it->StaticDrag = GetFloatToken();
			}
			else if (ParamName == "surfdrag")
			{
				it->SurfDrag = GetFloatToken();
			}
			else if (ParamName == "buoyancy") 
			{
				it->Buoyancy = GetFloatToken();
			} 
			else if (ParamName == "mass") 
			{
				it->Mass = GetFloatToken();
			} 
			else if (ParamName == "angvel")
			{
				if (GetPuncTokenChar() != '(') 
				{
					ConOut("ERROR: Expected ( after \"angvel=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->AngVel.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in angvel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->AngVel.y = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in angvel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->AngVel.z = GetFloatToken();
				if (GetPuncTokenChar() != ')')
				{
					ConOut("ERROR: Expected ) after angvel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "maxrotthrust")
			{
				it->MaxRotThrust = GetFloatToken();
			}
			else if (ParamName == "maxrotvel")
			{
				it->MaxRotVel = GetFloatToken();
			}
			else if (ParamName == "vel")
			{
				if(GetPuncTokenChar() != '(')
				{
					ConOut("ERROR: Expected ( after \"vel=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Velocity.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in vel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Velocity.y = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in vel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Velocity.z = GetFloatToken();
				if (GetPuncTokenChar() != ')')
				{
					ConOut("ERROR: Expected ) after vel value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else if(ParamName == "maxthrust")
			{
				it->MaxThrust = GetFloatToken();
			}
			else if (ParamName == "maxvel")
			{
				it->MaxVel = GetFloatToken();
			}
			else if (ParamName == "orient") 
			{
				if (GetPuncTokenChar() != '(') 
				{
					ConOut("ERROR: Expected ( after \"orient=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Orient.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in orient value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Orient.y = GetFloatToken();
				if (GetPuncTokenChar() != '/')
				{
					ConOut("ERROR: Expected / in orient value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->Orient.z = GetFloatToken();
				if (GetPuncTokenChar() != ')') 
				{
					ConOut("ERROR: Expected ) after orient value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "physflags") 
			{
				it->PhysFlags = GetIntegerToken();
			}
			else if (ParamName == "pitchrange")
			{
				it->PitchRange = GetFloatToken();
			} 
			else if (ParamName == "yawrange")
			{
				it->YawRange = GetFloatToken();
			}
			else if (ParamName == "cog") 
			{
				it->COG = Sith::VFS::ResourceManager::Get().GetScript(GetFilenameToken());
				if(!it->COG.isValid()) 
				{
					ConOut("WARNING: Could not load cog script referenced by template.\n", ConOut::Error);
				}
			}
			else if (ParamName == "blasttime") 
			{
				it->BlastTime = GetFloatToken();
			} 
			else if (ParamName == "explode") 
			{
				it->Explode = FindTemplate (GetTemplateToken());
				if (it->Explode == NULL)
				{
					ConOut ("ERROR: Could not find template for explode.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "debris")
			{
				it->Debris = FindTemplate (GetTemplateToken());
				if (it->Debris == NULL)
				{
					ConOut ("ERROR: Could not find template for debris.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "flashrgb") 
			{
				if (GetPuncTokenChar() != '(') 
				{
					ConOut("ERROR: Expected ( after \"flashrgb=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FlashRGB.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in flashrgb value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FlashRGB.y = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in flashrgb value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FlashRGB.z = GetFloatToken();
				if (GetPuncTokenChar() != ')') 
				{
					ConOut("ERROR: Expected ) after flashrgb value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else if (ParamName == "fleshhit")
			{
				it->FleshHit = FindTemplate (GetTemplateToken());
				if (it->FleshHit == NULL) 
				{
					ConOut ("ERROR: Could not find template for fleshhit.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "force") 
			{
				it->Force = GetFloatToken();
			} 
			else if (ParamName == "model3d")
			{
				if(ColormapList.size() > 0)
				{
                    it->Model3D = VFS::ResourceManager::Get().GetModel(GetFilenameToken(), ColormapList[0]);
				}
				else
				{
					it->Model3D = VFS::ResourceManager::Get().GetModel(GetFilenameToken());
				};
			} 
			else if (ParamName == "particle")
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::Particle> Particle
			}
			else if (ParamName == "material") 
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::Material> Material
			} 
			else if (ParamName == "sprite") 
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::Sprite> Sprites
			}
			else if (ParamName == "count") 
			{
				it->Count = GetIntegerToken();
			} 
			else if (ParamName == "elementsize")
			{
				it->ElementSize = GetFloatToken();
			} 
			else if (ParamName == "minsize")
			{
				it->MinSize = GetFloatToken();
			} 
			else if (ParamName == "range")
			{
				it->Range = GetFloatToken();
			} 
			else if (ParamName == "rate")
			{
				it->Rate = GetFloatToken();
			} 
			else if (ParamName == "creatething")
			{
				std::string tplname = GetTemplateToken();
				if(tplname == "none")
				{
					it->CreateThing = 0;
				}
				else
				{
                    it->CreateThing = FindTemplate(tplname);
					if (it->CreateThing == NULL)
					{
						ConOut ("ERROR: Could not find template for creatething.\n", ConOut::Error);
						return SE_FILE_CORRUPT;
					}
				};
			}
			else if (ParamName == "damage") 
			{
				it->Damage = GetFloatToken();
			}
			else if (ParamName == "mindamage") 
			{
				it->MinDamage = GetFloatToken();
			} 
			else if (ParamName == "damageclass") 
			{
				it->DamageClass = GetIntegerToken();
			} 
			else if (ParamName == "aiclass")
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::AIClass> Aiclass;
			} 
			else if (ParamName == "puppet") 
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::Puppet> Puppet;
			} 
			else if (ParamName == "soundclass") 
			{
				GetFilenameToken();
				// Sith::VFS::WisePointer<Res::Soundclass> Soundclass;
			}
			else if (ParamName == "chance") 
			{
				it->Chance = GetFloatToken();
			}
			else if (ParamName == "error")
			{
				it->Error = GetFloatToken();
			} 
			else if (ParamName == "eyeoffset")
			{
				if (GetPuncTokenChar() != '(') 
				{
					ConOut("ERROR: Expected ( after \"eyeoffset=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->EyeOffset.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in eyeoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->EyeOffset.y = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in eyeoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->EyeOffset.z = GetFloatToken();
				if (GetPuncTokenChar() != ')') 
				{
					ConOut("ERROR: Expected ) after eyeoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} else if (ParamName == "fireoffset")
			{
				if (GetPuncTokenChar() != '(')
				{
					ConOut("ERROR: Expected ( after \"fireoffset=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FireOffset.x = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in fireoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FireOffset.y = GetFloatToken();
				if (GetPuncTokenChar() != '/') 
				{
					ConOut("ERROR: Expected / in fireoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->FireOffset.z = GetFloatToken();
				if (GetPuncTokenChar() != ')') 
				{
					ConOut("ERROR: Expected ) after fireoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else if (ParamName == "fov") 
			{
				it->FOV = GetFloatToken();
			} 
			else if (ParamName == "health")
			{
				it->Health = GetFloatToken();
			}
			else if (ParamName == "maxhealth")
			{
				it->MaxHealth = GetFloatToken();
			}
			else if (ParamName == "height") 
			{
				it->Height = GetFloatToken();
			} 
			else if (ParamName == "jumpspeed")
			{
				it->JumpSpeed = GetFloatToken();
			}
			else if (ParamName == "maxheadpitch")
			{
				it->MaxHeadPitch = GetFloatToken();
			}
			else if (ParamName == "minheadpitch")
			{
				it->MinHeadPitch = GetFloatToken();
			} 
			else if (ParamName == "weapon") 
			{
				std::string Temp = GetTemplateToken();
				if(Temp == "none")
				{
					it->Weapon = 0;
				}
				else
				{
					it->Weapon = FindTemplate(Temp);
					if (it->Weapon == NULL) 
					{
						ConOut ("ERROR: Could not find template for weapon.\n", ConOut::Error);
						return SE_FILE_CORRUPT;
					}
				};
			} 
			else if(ParamName == "weapon2")
			{
				std::string Temp = GetTemplateToken();
				if(Temp == "none")
				{
					it->Weapon = 0;
				}
				else
				{
					it->Weapon = FindTemplate(Temp);
					if(it->Weapon == NULL)
					{
						ConOut("ERROR: Could not find template for weapon2.\n", ConOut::Error);
						return SE_FILE_CORRUPT;
					};
				};
			}
			else if (ParamName == "light") 
			{
				it->Light = GetFloatToken();
			} 
			else if (ParamName == "maxlight") 
			{
				it->MaxLight = GetFloatToken();
			} 
			else if (ParamName == "lightintensity")
			{
				it->LightIntensity = GetFloatToken();
			} 
			else if (ParamName == "lightoffset")
			{
				if (GetPuncTokenChar() != '(') 
				{
					ConOut("ERROR: Expected ( after \"lightoffset=\".\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->LightOffset.x = GetFloatToken();
				if (GetPuncTokenChar() != '/')
				{
					ConOut("ERROR: Expected / in lightoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->LightOffset.y = GetFloatToken();
				if (GetPuncTokenChar() != '/')
				{
					ConOut("ERROR: Expected / in lightoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
				it->LightOffset.z = GetFloatToken();
				if (GetPuncTokenChar() != ')')
				{
					ConOut("ERROR: Expected ) after lightoffset value.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			} 
			else if (ParamName == "respawn") 
			{
				it->Respawn = GetFloatToken();
			}
			else if (ParamName == "timer")
			{
				it->Timer = GetFloatToken();
			}
			else if (ParamName == "trailrandangle") 
			{
				it->TrailRandAngle = GetFloatToken();
			}
			else if(ParamName == "trailcylradius")
			{
				it->TrailCylRadius = GetFloatToken();
			}
			else if (ParamName == "trailthing") 
			{
				it->TrailThing = FindTemplate (GetTemplateToken());
				if (it->TrailThing == NULL) 
				{
					ConOut ("ERROR: Could not find template for trailthing.\n", ConOut::Error);
					return SE_FILE_CORRUPT;
				}
			}
			else 
			{
				ConOut ("WARNING: Unknown template parameter " + ParamName + ".\n", ConOut::Warning);
			}
		}
	}
	return SE_OK;
}

Sith::SITHERROR Sith::Res::Level::ParseThings()
{
	if(GetWordToken() != "world") return SE_FILE_CORRUPT;
	if(GetWordToken() != "things") return SE_FILE_CORRUPT;

	GetIntegerToken(); // totally useless thing count...

	while(CurrentOffset < BufferLen)
	{
		if(GetWordToken() == "end") return SE_OK;
		else
		{
			Gameplay::Thing NewThing;

			GetIntegerToken(); // num
			GetPuncToken(); // colon

			NewThing = *FindTemplate(GetTemplateToken());

			NewThing.Name = GetTemplateToken();

			NewThing.Pos.x = GetFloatToken();
			NewThing.Pos.y = GetFloatToken();
			NewThing.Pos.z = GetFloatToken();
			NewThing.Orient.x = GetFloatToken();
			NewThing.Orient.y = GetFloatToken();
			NewThing.Orient.z = GetFloatToken();

			NewThing.Sector = GetIntegerToken();

			NewThing.HeadPitch = 0;
			NewThing.Instanced = true;

			NewThing.AttachIndex = -1;
			NewThing.AttachState = Gameplay::Thing::FREE;
			
			NewThing.Velocity = Math::VectorSet(0.0f, 0.0f, 0.0f);
			NewThing.Force = 0.0f;
			NewThing.Thrust = Math::VectorSet(0.0f, 0.0f, 0.0f);

			int NewlineOffset = FindNewLine();
			std::string ParamName;
			while (CurrentOffset < NewlineOffset) {
				ParamName = GetWordToken();
				if (ParamName == "numframes") {	// frame definitions
					if (GetPuncTokenChar() != '=') {
						return SE_FILE_CORRUPT;
					}
					NewThing.Frames.resize(GetIntegerToken());
					for (int i = 0; i < NewThing.Frames.size(); i++) {
						// parse the frame
						if (GetWordToken() != "frame") {
							return SE_FILE_CORRUPT;
						}
						if (GetPuncTokenChar() != '=') {
							return SE_FILE_CORRUPT;
						}
						if (GetPuncTokenChar() != '(') {
							return SE_FILE_CORRUPT;
						}
						NewThing.Frames[i].X = GetFloatToken();
						if (GetPuncTokenChar() != '/') {
							return SE_FILE_CORRUPT;
						}
						NewThing.Frames[i].Y = GetFloatToken();
						if (GetPuncTokenChar() != '/') {
							return SE_FILE_CORRUPT;
						}
						NewThing.Frames[i].Z = GetFloatToken();
						char PuncChar = GetPuncTokenChar();
						if (PuncChar == ':') {	// affect orientation
							NewThing.Frames[i].AffectOrient = true;
							NewThing.Frames[i].Pitch = GetFloatToken();
							if (GetPuncTokenChar() != '/') {
								return SE_FILE_CORRUPT;
							}
							NewThing.Frames[i].Yaw = GetFloatToken();
							if (GetPuncTokenChar() != '/') {
								return SE_FILE_CORRUPT;
							}
							NewThing.Frames[i].Roll = GetFloatToken();
							if (GetPuncTokenChar() != ')') {
								return SE_FILE_CORRUPT;
							}
						} else if (PuncChar == ')') {
							NewThing.Frames[i].AffectOrient = false;
						} else {
							return SE_FILE_CORRUPT;
						}
						SkipWhitespace();
					}
				} else if (ParamName == "thingflags") {	// some other param we aren't parsing yet
					if (GetPuncTokenChar() != '=') {
						return SE_FILE_CORRUPT;
					}
					NewThing.ThingFlags = GetIntegerToken();
				} else if (ParamName == "aiclass") {
					if (GetPuncTokenChar() != '=') {
						return SE_FILE_CORRUPT;
					}
					NewThing.AIClass = GetFilenameToken();
				} else if (ParamName == "vel") {
					if (GetPuncTokenChar() != '=') {
						return SE_FILE_CORRUPT;
					}
					if (GetPuncTokenChar() != '(') {
						return SE_FILE_CORRUPT;
					}
					NewThing.Velocity.x = GetFloatToken();
					if (GetPuncTokenChar() != '/') {
						return SE_FILE_CORRUPT;
					}
					NewThing.Velocity.y = GetFloatToken();
					if (GetPuncTokenChar() != '/') {
						return SE_FILE_CORRUPT;
					}
					NewThing.Velocity.z = GetFloatToken();
					if (GetPuncTokenChar() != ')') {
						return SE_FILE_CORRUPT;
					}
				} else {
					// skip to next bit
					while (!isspace(Buffer[CurrentOffset]) && CurrentOffset < NewlineOffset) CurrentOffset++;
				}
				if (CurrentOffset > NewlineOffset) {
					CurrentOffset = NewlineOffset;
				}
			}

			NewThing.CurFrame = NewThing.DestFrame = 0;

            Things.push_back(NewThing);
			SkipToNewLine();
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Res::Level::Parse()
{
	std::string LALR;
	while(CurrentOffset < BufferLen)
	{
		// Parse a list of sections...
		if(GetWordToken() != "section")
		{
			SkipToNextSection();
			continue;
		};
		if(GetPuncToken() != ":") return SE_FILE_CORRUPT;

		std::string SectionName = GetWordToken();

		if(SectionName == "jk")
		{
			SkipToNextSection();
		}
		else if(SectionName == "header")
		{
			if(Failed(ParseHeader())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "materials")
		{
			if(Failed(ParseMaterials())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "georesource")
		{
			if(Failed(ParseGeoresource())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "sectors")
		{
			if(Failed(ParseSectors())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "lights")
		{
			if(Failed(ParseLights())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "cogs")
		{
			if(Failed(ParseCogs())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "templates")
		{
			if(Failed(ParseTemplates())) return SE_FILE_CORRUPT;
		}
		else if(SectionName == "things")
		{
			if(Failed(ParseThings())) return SE_FILE_CORRUPT;
		}
		else
		{
			ConOut("WARNING: Unrecognized section \'" + SectionName + "\', bypassing.\n", ConOut::Warning);
			SkipToNextSection();
		};
	};

//	return Compile();
	return SE_OK;
};

/*Sith::SITHERROR Sith::Res::Level::Compile()
{
	CompiledHeader.Version = Header.Version;
	CompiledHeader.Gravity = Header.Gravity;
	
	CompiledHeader.CeilingSkyZ = Header.CeilingSkyZ;
	CompiledHeader.CeilingSkyOffset = Sith::Math::Vector(Header.CeilingSkyOffset[0], Header.CeilingSkyOffset[1]);

	CompiledHeader.HorizonDistance = Header.HorizonDistance;
	CompiledHeader.HorizonPixelsPerRev = Header.HorizonPixelsPerRev;
	CompiledHeader.HorizonSkyOffset = Math::Vector(Header.HorizonSkyOffset[0], Header.HorizonSkyOffset[1]);

	// Skip rest, whatever.

	// Copy over sector geometry
	for(std::vector<SECTOR>::iterator it = SectorList.begin(); it != SectorList.end(); ++it)
	{
		CompiledGeometry.push_back(Sith::World::LevelSector());

		CompiledGeometry.back().SectorNumber = it->SectorNumber;
		CompiledGeometry.back().SectorFlags = it->SectorFlags;

		CompiledGeometry.back().Center = Math::Vector(it->Center[0], it->Center[1], it->Center[2]);

		CompiledGeometry.back().Radius = it->Radius;

		// Skip bounding box and collidebox for now, in a hurry...

		CompiledGeometry.back().AmbientLight = it->AmbientLight;
		CompiledGeometry.back().ExtraLight = it->ExtraLight;
	};

	return SE_OK;
};*/

Sith::SITHERROR Sith::Res::Level::LoadFromFile(const char* FileBuffer)
{
	CurrentOffset = 0;
	LineNumber = 0;
	EndOfStream = false;
	Buffer = FileBuffer;

	BufferLen = strlen(FileBuffer);

	if(Failed(Parse()))
	{
		std::ostringstream Stream;
		Stream << "ERROR: JKL syntax error on line " << LineNumber << ".\n";
		ConOut(Stream.str(), ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	else
	{
		// Load textures
		for(int i = 0; i < TextureList.size(); ++i)
		{
			if(ColormapList.size() > 0)
			{
                MatList.push_back(VFS::ResourceManager::Get().GetMaterial(TextureList[i].TextureName, ColormapList[0]));
			}
			else
			{
				MatList.push_back(VFS::ResourceManager::Get().GetMaterial(TextureList[i].TextureName, "dflt.cmp"));
			};
		};

		// Process sector membership information into extensions.
		for(std::vector<SECTOR>::iterator it = SectorList.begin(); it != SectorList.end(); ++it)
		{
			int SurfaceIndex = it->FirstSurfaceIndex;

			for(int i = 0; i < it->SurfaceCount; ++i)
			{
				SurfaceList[SurfaceIndex].ParentSector = &(*it);
				it->SurfaceList.push_back(&SurfaceList[SurfaceIndex]);
				++SurfaceIndex;
			};
		};

		// Process adjoin information into extensions:
		for(std::vector<SURFACE>::iterator it = SurfaceList.begin(); it != SurfaceList.end(); ++it)
		{
            if((it->AdjoinIndex >= 0) && (it->AdjoinIndex < AdjoinList.size()))
			{
				for(std::vector<SURFACE>::iterator jt = SurfaceList.begin(); jt != SurfaceList.end(); ++jt)
				{
					if(jt->AdjoinIndex == AdjoinList[it->AdjoinIndex].Mirror)
					{
						it->AdjoinMirrorSurface = &(*jt);
						it->AdjoinedSector = jt->ParentSector;
						break;
					}
				};
			}
			else
			{
				it->AdjoinMirrorSurface = 0;
				it->AdjoinedSector = 0;
			};
		};

		// calculate tangent, binormal vectors for all surfaces
		for(std::vector<SURFACE>::iterator it = SurfaceList.begin(); it != SurfaceList.end(); ++it)
		{
			// based on tangent binormal approach from NVMeshMender
			// assumes that texture coordinates are uniform over the surface (which doesn't
			// have to be the case, but in every level I have seen it is

			if (it->TextureIndex == -1) // need tex coords
			{
				continue;
			}

			int v2 = 2;

			float s1, t1, s2, t2;
			s1 = TexvertexList[it->VertexList[1].TexvertexIndex].U - TexvertexList[it->VertexList[0].TexvertexIndex].U;
			t1 = TexvertexList[it->VertexList[1].TexvertexIndex].V - TexvertexList[it->VertexList[0].TexvertexIndex].V;
			s2 = TexvertexList[it->VertexList[v2].TexvertexIndex].U - TexvertexList[it->VertexList[0].TexvertexIndex].U;
			t2 = TexvertexList[it->VertexList[v2].TexvertexIndex].V - TexvertexList[it->VertexList[0].TexvertexIndex].V;

			float tmp;
			while (fabs(s1 * t2 - s2 * t1) < .0001f)
			{
				v2++;
				if (v2 >= it->VertexCount)
				{
					break;
				}
				s2 = TexvertexList[it->VertexList[v2].TexvertexIndex].U - TexvertexList[it->VertexList[0].TexvertexIndex].U;
				t2 = TexvertexList[it->VertexList[v2].TexvertexIndex].V - TexvertexList[it->VertexList[0].TexvertexIndex].V;
			}
			if (v2 >= it->VertexCount)
				continue;
			tmp = 1.0f / (s1 * t2 - s2 * t1);

			float p[3], q[3];
			p[0] = GeovertexList[it->VertexList[1].GeovertexIndex].X - GeovertexList[it->VertexList[0].GeovertexIndex].X;
			p[1] = GeovertexList[it->VertexList[1].GeovertexIndex].Y - GeovertexList[it->VertexList[0].GeovertexIndex].Y;
			p[2] = GeovertexList[it->VertexList[1].GeovertexIndex].Z - GeovertexList[it->VertexList[0].GeovertexIndex].Z;
			q[0] = GeovertexList[it->VertexList[v2].GeovertexIndex].X - GeovertexList[it->VertexList[0].GeovertexIndex].X;
			q[1] = GeovertexList[it->VertexList[v2].GeovertexIndex].Y - GeovertexList[it->VertexList[0].GeovertexIndex].Y;
			q[2] = GeovertexList[it->VertexList[v2].GeovertexIndex].Z - GeovertexList[it->VertexList[0].GeovertexIndex].Z;

			// tangent = V tex coord
			it->Tangent.X = tmp * (t2 * p[0] - t1 * q[0]);
			it->Tangent.Y = tmp * (t2 * p[1] - t1 * q[1]);
			it->Tangent.Z = tmp * (t2 * p[2] - t1 * q[2]);
			float invlen = 1 / sqrt(it->Tangent.X * it->Tangent.X + it->Tangent.Y * it->Tangent.Y + it->Tangent.Z * it->Tangent.Z);
			it->Tangent.X *= invlen;
			it->Tangent.Y *= invlen;
			it->Tangent.Z *= invlen;
			it->TexDensityU = invlen;

			// binormal = U tex coord
			it->Binormal.X = tmp * (s1 * q[0] - s2 * p[0]);
			it->Binormal.Y = tmp * (s1 * q[1] - s2 * p[1]);
			it->Binormal.Z = tmp * (s1 * q[2] - s2 * p[2]);
			invlen = 1 / sqrt(it->Binormal.X * it->Binormal.X + it->Binormal.Y * it->Binormal.Y + it->Binormal.Z * it->Binormal.Z);
			it->Binormal.X *= invlen;
			it->Binormal.Y *= invlen;
			it->Binormal.Z *= invlen;
			it->TexDensityV = invlen;
		};

		return SE_OK;
	};
};