/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _JKLFILE_H_INCLUDED_
#define _JKLFILE_H_INCLUDED_

namespace Sith
{
	namespace Res
	{
		class Level : public VFS::Resource
		{
		public:
			// Parsing data:
			unsigned int CurrentOffset;
			unsigned int LineNumber;
			size_t BufferLen;
			const char* Buffer;
			bool EndOfStream;

			SITHERROR ParseHeader();
			SITHERROR ParseSounds();
			SITHERROR ParseMaterials();
			SITHERROR ParseGeoresource();
			SITHERROR ParseSectors();
			SITHERROR ParseSector();
			SITHERROR ParseAiclass();
			SITHERROR ParseModels();
			SITHERROR ParseSprites();
			SITHERROR ParseKeyframes();
			SITHERROR ParseAnimclass();
			SITHERROR ParseSoundclass();
			SITHERROR ParseCogscripts();
			SITHERROR ParseCogs();
			SITHERROR ParseTemplates();
			SITHERROR ParseThings();
			SITHERROR ParseLights();

			std::vector<COG::StackMachine> Cogs;

			__forceinline void SkipComment()
			{
				if(Buffer[CurrentOffset] == '#')
				{
					while(CurrentOffset < BufferLen)
					{
						if(Buffer[CurrentOffset] == '\n')
						{
							++LineNumber;
							++CurrentOffset;
							return;
						}
						else if(Buffer[CurrentOffset] == '\0')
						{
							return;
							EndOfStream = true;
						};
						++CurrentOffset;
					};
				};
				return;
			};

			__forceinline void SkipWhitespace()
			{
				while(CurrentOffset < BufferLen)
				{
					SkipComment();
					if(Buffer[CurrentOffset] == '\r') ++LineNumber;
					else if(Buffer[CurrentOffset] == '\n' && CurrentOffset > 0 && Buffer[CurrentOffset - 1] != '\r') ++LineNumber;
					if(isspace((int)Buffer[CurrentOffset]))
					{
						++CurrentOffset;
						continue;
					}
					else break;
				};
				return;
			};

			// stolen from parpar by Jon`C :D
			__forceinline void SkipToNewLine() {
				while (CurrentOffset < BufferLen) {
					if (Buffer[CurrentOffset] == '\n') {
						++CurrentOffset;
						++LineNumber;
						return;
					}
					++CurrentOffset;
				}
				return;
			};

			__forceinline unsigned int FindNewLine() {
				unsigned int NewOffset = CurrentOffset;
				while (NewOffset < BufferLen) {
					if (Buffer[NewOffset] == '\n') {
						return NewOffset;
					}
					++NewOffset;
				}
				return 0;
			};

			__forceinline bool TestForNewLine()
			{
				while(CurrentOffset < BufferLen)
				{
					if(isspace((int)Buffer[CurrentOffset]))
					{
						++CurrentOffset;
						continue;
					}
					else if(Buffer[CurrentOffset] == '#')
					{
						SkipComment();
						return true;
					}
					else if(Buffer[CurrentOffset] == '\n')
					{
						++CurrentOffset;
						return true;
					}
					else
					{
						return false;
					};

					++CurrentOffset;
				};
			};

			inline std::string GetWordToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(isalpha((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetTemplateParamToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(isalnum((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetTemplateToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen && !isspace((int)Buffer[CurrentOffset]))
				{
					RetVal.push_back(Buffer[CurrentOffset++]);
					/*if(isalnum((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					if(Buffer[CurrentOffset] == '+')
						RetVal.push_back(Buffer[CurrentOffset++]);
					if(Buffer[CurrentOffset] == '_')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;*/
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetFilenameToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(isalnum((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '_')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '~')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '.')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '-')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetPuncToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(ispunct((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				return RetVal;
			};

			inline unsigned char GetPuncTokenChar()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return '\0';
				unsigned char RetVal = '\0';
				while(CurrentOffset < BufferLen)
				{
					if(ispunct((int)Buffer[CurrentOffset]))
						return Buffer[CurrentOffset++];
					else break;					
				};
				return RetVal;
			};

			inline int GetIntegerToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return 0;
				std::string Retval;
				// Hex integer
				if(((Buffer[CurrentOffset] == '0') && ((Buffer[CurrentOffset + 1] == 'x') || (Buffer[CurrentOffset + 1] == 'X'))) || ((Buffer[CurrentOffset] == 'x') || Buffer[CurrentOffset] == 'X'))
				{
					CurrentOffset += 2;
					while(CurrentOffset < BufferLen)
					{
						if(isdigit((int)Buffer[CurrentOffset]))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else if((Buffer[CurrentOffset] >= 'A') && (Buffer[CurrentOffset] <= 'F'))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else if((Buffer[CurrentOffset] >= 'a') && (Buffer[CurrentOffset] <= 'f'))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else break;
					};
					size_t Len = Retval.length();
					int *digit = new int[Len];
					int RetVal = 0;
					unsigned int Count = 0;
					unsigned int i = 0, j = 0;
					for(i = 0; i < Retval.length(); ++i)
					{
						if(Retval[i] > 0x29 && Retval[i] < 0x40)
							digit[i] = Retval[i] & 0x0F;
						else if(Retval[i] >= 'a' && Retval[i] <= 'f')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else if(Retval[i] >= 'A' && Retval[i] <= 'F')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else break;

						++Count;
					};
					j = Count - 1;
					for(i = 0; i < Count; ++i)
					{
						RetVal = RetVal | (digit[i] << (j << 2));
						--j;
					};
					return RetVal;
				}
				else // Normal integer
				{
					if(Buffer[CurrentOffset] == '-')
					{
						Retval.push_back(Buffer[CurrentOffset++]);
					};

					while(CurrentOffset < BufferLen)
					{
						if(isdigit((int)Buffer[CurrentOffset]))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else
						{
							break;
						};
					};

					return atoi(Retval.c_str());
				};
			};

			float GetFloatToken()
			{
				if(Buffer[CurrentOffset] == '\0') return 0.0f;

				SkipWhitespace();

				std::string Retval;

				if(Buffer[CurrentOffset] == '-')
				{
					Retval.push_back(Buffer[CurrentOffset++]);
				};

				if(Buffer[CurrentOffset] == '.')
				{
					// inject 0 before decimal place..
					Retval.push_back('0');
				};

				while(CurrentOffset < BufferLen)
				{
					if(isdigit((int)Buffer[CurrentOffset]))
					{
						Retval.push_back(Buffer[CurrentOffset++]);
						continue;
					}
					else if(Buffer[CurrentOffset] == '.')
					{
						Retval.push_back(Buffer[CurrentOffset++]);
						continue;
					}
					else
					{
						break;
					};
				};

				return (float)atof(Retval.c_str());
			};

			inline void SkipToNextSection()
			{
				for( ; CurrentOffset < BufferLen; ++CurrentOffset)
				{
					if((Buffer[CurrentOffset] == 's') || (Buffer[CurrentOffset] == 'S'))
					{
						++CurrentOffset;
						if((Buffer[CurrentOffset] == 'e') || (Buffer[CurrentOffset] == 'E'))
						{
							++CurrentOffset;
							if((Buffer[CurrentOffset] == 'c') || (Buffer[CurrentOffset] == 'C'))
							{
								++CurrentOffset;
								if((Buffer[CurrentOffset] == 't') || (Buffer[CurrentOffset] == 'T'))
								{
									++CurrentOffset;
									if((Buffer[CurrentOffset] == 'i') || (Buffer[CurrentOffset] == 'I'))
									{
										++CurrentOffset;
										if((Buffer[CurrentOffset] == 'o') || (Buffer[CurrentOffset] == 'O'))
										{
											++CurrentOffset;
											if((Buffer[CurrentOffset] == 'n') || (Buffer[CurrentOffset] == 'N'))
											{
												CurrentOffset -= 6;
												return;
											};
										};
									};
								};
							};
						};
					};

					if(Buffer[CurrentOffset] == '\n') ++LineNumber;

					if(Buffer[CurrentOffset] == '#')
					{
						SkipWhitespace();
						// dial it back one, yo. This one is self-driving :(
						--CurrentOffset;
					};
				};

				return;
			};

			SITHERROR Compile();
			SITHERROR Parse();

			struct JKLHEADER
			{
				int Version;

				float Gravity;

				float CeilingSkyZ;
				float HorizonDistance;
				float HorizonPixelsPerRev;

				float HorizonSkyOffset[2];

				float CeilingSkyOffset[2];

				float MipMapDistances[4];

				float LODDistances[4];

				float PerspectiveDistance;
				float GouraudDistance;

				int LevelTint[3];
			} Header;

			// std::vector<VFS::WisePointer<Res::Sound>> SoundList;
			struct MATDATA
			{
				std::string TextureName;
				float XScale, YScale;

				MATDATA()
				{
					return;
				};

				MATDATA(std::string _Texture, float _XScale, float _YScale) : TextureName(_Texture), XScale(_XScale), YScale(_YScale)
				{
					return;
				};
			};
			std::vector<MATDATA> TextureList;

			std::vector<std::string> ColormapList;

			struct GEOVERTEX
			{
				float X, Y, Z;

				GEOVERTEX()
				{
					return;
				};

				GEOVERTEX(float _X, float _Y, float _Z) : X(_X), Y(_Y), Z(_Z)
				{
					return;
				};
			};

			std::vector<GEOVERTEX> GeovertexList;

			struct TEXVERTEX
			{
				float U, V;

				TEXVERTEX()
				{
					return;
				};

				TEXVERTEX(float _U, float _V) : U(_U), V(_V)
				{
					return;
				};
			};

			std::vector<TEXVERTEX> TexvertexList;

			struct ADJOIN
			{
				int Flags;
				int Mirror;

				float Dist;

				ADJOIN()
				{
					return;
				};

				ADJOIN(int _Flags, int _Mirror, float _Dist) : Flags(_Flags), Mirror(_Mirror), Dist(_Dist)
				{
					return;
				};
			};

			std::vector<ADJOIN> AdjoinList;

			struct SURFACEVERTEX
			{
				int GeovertexIndex;
				int TexvertexIndex;
				float LightChroma;
				float LightR;
				float LightG;
				float LightB;

				SURFACEVERTEX()
				{
					GeovertexIndex = 0;
					TexvertexIndex = 0;
					LightChroma = 0;
					LightR = 0;
					LightG = 0;
					LightB = 0;
					return;
				};

				SURFACEVERTEX(int Vertex, int Texvert, float LChroma, float LR, float LG, float LB)
				{
					GeovertexIndex = Vertex;
					TexvertexIndex = Texvert;
					LightChroma = LChroma;
					LightR = LR;
					LightG = LG;
					LightB = LB;

					return;
				};
			};

			struct SURFACENORMAL
			{
				float X, Y, Z;

				SURFACENORMAL(float _X, float _Y, float _Z) : X(_X), Y(_Y), Z(_Z)
				{
					return;
				};

				SURFACENORMAL() : X(0), Y(0), Z(0)
				{
					return;
				};
			};

			struct SURFACE;

			struct SECTOR
			{
				int SectorNumber;

				int SectorFlags;

				float AmbientLight;
				float ExtraLight;
				int ColormapIndex;

				float Tint[3];

				float BoundBox[6];
				float CollideBox[6];

				Math::Vector Center;

				float Radius;

				int VertexCount;
				std::vector<int> VertexIndexList;

				int FirstSurfaceIndex;
				int SurfaceCount;

				std::vector<SURFACE*> SurfaceList;

				std::string SoundName;
				float SoundVolume;

				std::vector<COG::StackMachine*> LinkTable;
			};

			struct SURFACE
			{
				int SurfaceNumber;

				int TextureIndex;
				int SurfaceFlags;
				int FaceFlags;
				int GeoMode;
				int LightMode;
				int TextureMode;

				int AdjoinIndex;

				float ExtraLight;

				// Animation values, not in JKL:
				int CurrentCel;
				float TexShiftU, TexShiftV;

				int VertexCount;
				std::vector<SURFACEVERTEX> VertexList;

				// special effects/extensions
				SURFACE* AdjoinMirrorSurface;
				SECTOR* AdjoinedSector;

				SECTOR* ParentSector;

				// tangent and binormal not in JKL, they are the vectors representing U and V
				SURFACENORMAL Normal, Tangent, Binormal;

				// not in JKL, and not used by the renderer; these are the assumed U,V scaling
				float TexDensityU, TexDensityV;

				SURFACE()
				{
					CurrentCel = 0;
					TexShiftU = 0;
					TexShiftV = 0;

					AdjoinMirrorSurface = 0;
					AdjoinedSector = 0;
					ParentSector = 0;

					TexDensityU = 0;
					TexDensityV = 0;

					return;
				};

				SURFACE(int SurfNum, int Texture, int Flags, int Face, int Geo, int Light, int Tex, int Adjoin, float ExLight, int VertCount)
				{
					SurfaceNumber = SurfNum;
					TextureIndex = Texture;
					SurfaceFlags = Flags;
					FaceFlags = Face;
					GeoMode = Geo;
					LightMode = Light;
					TextureMode = Tex;
					AdjoinIndex = Adjoin;
					ExtraLight = ExLight;
					VertexCount = VertCount;

					CurrentCel = 0;
					TexShiftU = 0;
					TexShiftV = 0;

					AdjoinMirrorSurface = 0;
					AdjoinedSector = 0;
					ParentSector = 0;

					TexDensityU = 0;
					TexDensityV = 0;

					return;
				};

				std::vector<COG::StackMachine*> LinkTable;
			};

			std::vector<SURFACE> SurfaceList;

			std::vector<SECTOR> SectorList;

			struct LIGHT
			{
				int Type;

				float X, Y, Z, Intensity, Range;
			};

			std::vector<LIGHT> LightList;

			std::vector<Sith::Gameplay::Template> Templates;
			std::vector<Sith::Gameplay::Thing> Things;

			std::vector<VFS::WisePointer<Res::Material> > MatList;

		public:
//			World::LevelData CompiledHeader;
//			std::vector<World::LevelSector> CompiledGeometry;
//			std::vector<World::LevelSurface> CompiledSurfaces;

			Level(std::string SystemName);
			SITHERROR LoadFromFile(const char* FileBuffer);
			Sith::Gameplay::Template* FindTemplate (std::string SearchName) {
				std::vector<Sith::Gameplay::Template>::iterator it;
				for (it = Templates.begin(); it != Templates.end(); ++it) {
					if (it->TemplateName == SearchName) {
						return &(*it);
						break;
					}
				}
				return NULL;
			};
		};
	};
};

#endif // _JKLFILE_H_INCLUDED_