/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ktask.h"
#include "kernel.h"

#include "logger.h"

#include "console.h"
#include "wincon.h"
#include "win_chooser.hpp"

#include "resman.h"

#include "ogl_render.h"
#include "di_input.h"
#include "oal_sound.h"

#include "uiwidget.h"
#include "uimain.h"

#include "fpsproc.h"
#include "animcontext.h"
#include "vmproc.h"

#include "worldstate.h"
#include "gamestate.h"

#include "uihud.h"

Sith::System::KernelProc::KernelProc() : Sith::System::Task("Kernel Messaging System")
{
	// Common/arithmetic/COGLIB verbs:
	Sith::COG::CogLib::InjectLibraryFunction("vectoradd", 2, &COG::CogLib::VectorAdd);
	Sith::COG::CogLib::InjectLibraryFunction("vectorcross", 2, &COG::CogLib::VectorCross);
	Sith::COG::CogLib::InjectLibraryFunction("vectordist", 2, &COG::CogLib::VectorDist);
	Sith::COG::CogLib::InjectLibraryFunction("vectordistsq", 2, &COG::CogLib::VectorDistSq);
	Sith::COG::CogLib::InjectLibraryFunction("vectordot", 2, &COG::CogLib::VectorDot);
	Sith::COG::CogLib::InjectLibraryFunction("vectorlen", 1, &COG::CogLib::VectorLen);
	Sith::COG::CogLib::InjectLibraryFunction("vectorlensq", 1, &COG::CogLib::VectorLenSq);
	Sith::COG::CogLib::InjectLibraryFunction("vectornorm", 1, &COG::CogLib::VectorNorm);
	Sith::COG::CogLib::InjectLibraryFunction("vectorscale", 2, &COG::CogLib::VectorScale);
	Sith::COG::CogLib::InjectLibraryFunction("vectorset", 3, &COG::CogLib::VectorSet);
	Sith::COG::CogLib::InjectLibraryFunction("vectorsub", 2, &COG::CogLib::VectorSub);
	Sith::COG::CogLib::InjectLibraryFunction("vectorx", 1, &COG::CogLib::VectorX);
	Sith::COG::CogLib::InjectLibraryFunction("vectory", 1, &COG::CogLib::VectorY);
	Sith::COG::CogLib::InjectLibraryFunction("vectorz", 1, &COG::CogLib::VectorZ);

	// COGsole commands
	Sith::COG::CogLib::InjectLibraryFunction("lscmd", 0, &COG::CogLib::lscmd);
	Sith::COG::CogLib::InjectLibraryFunction("lsproc", 0, &Kernel::lsproc);
	Sith::COG::CogLib::InjectLibraryFunction("lsns", 0, &COG::VMHeap::lsns);
	Sith::COG::CogLib::InjectLibraryFunction("lsvar", 1, &COG::VMHeap::lsvar);
	Sith::COG::CogLib::InjectLibraryFunction("quit", 0, &Kernel::quit);
	Sith::COG::CogLib::InjectLibraryFunction("exit", 0, &Kernel::quit);

	Sith::COG::CogLib::InjectNamespace("system");
	Sith::COG::CogLib::AddSetting("system::uiconsole", COG::VMType(true));

	ExecutionState = RUNNING;
	QuitMessageRecieved = false;
};

Sith::SITHERROR Sith::System::KernelProc::Startup()
{
	new Sith::Logger();		// Instantiate logger
	new Sith::System::Console();	// Instantiate main console subsystem

	if(COG::CogLib::GetSetting("system::uiconsole").GetBool() == true)
	{
        new Sith::Win32::WinCon();			// Instantiate Win32 UI console
	};
	new Sith::Win32::WinChooser();

	ConOut("Welcome to Sith 2.\n\n", ConOut::Standard);

	new Sith::Win32::Sound();
	new Sith::Win32::Renderer();	// Instantiate DX renderer
	new Sith::Win32::Input();		// Instantiate input manager

	VFS::ResourceManager::Get(); // Create resource manager singleton -> vfs singleton

	new UI::UIMain();	// Instantiate user interface

	new Sith::Gameplay::WorldState();	// instantiate world state
	new Sith::Gameplay::GameState(); // instantiate game state

	new Sith::Gameplay::Animation();
	new Sith::FPSProc();
	new Sith::COG::VMProcess();

	new Sith::UI::HUD();

	ConOut("Kernel bootup complete.\n\n", ConOut::Standard);
	ConOut("Running autoexec.cog...\n", ConOut::Standard);
	if(Failed(Sith::System::Console::RunScriptFromPhysicalLocation("autoexec.cog")))
	{
		ConOut("WARNING: Failed to run autoexec.cog; using defaults.\n\n", ConOut::Warning);
	}
	else
	{
		ConOut("Done\n\n", ConOut::Standard);
	};

	VFS::ResourceManager::Get().Startup();

	return SE_OK;
};

Sith::System::KernelProc::~KernelProc()
{
	return;
};

Sith::SITHERROR Sith::System::KernelProc::MessageHandler(KSIGNAL Signal)
{
	switch(Signal)
	{
	case KS_QUIT:
		QuitMessageRecieved = true;
		System::Kernel::Get().SendSignal(KS_TERM);

		// Serialize global values to disk
		Sith::COG::VMHeap::SerializeGlobal("autoexec.cog");
		break;

	case KS_TERM:
		break;

	default:
		return DefaultMessageHandler(Signal);
	};

	return SE_OK;
};

Sith::SITHERROR Sith::System::KernelProc::Process()
{
	if((QuitMessageRecieved == true) && System::Kernel::Get().TaskList.size() <= 1)
	{
		ExecutionState = TERMINATED;
	};

	return SE_OK;
};

Sith::SITHERROR Sith::System::Kernel::Preempt()
{
	SITHERROR ReturnValue = SE_OK;

	if(TaskList.size() == 0)
	{
		return SE_KERNEL_QUIT;
	};

	for(std::list<Task*>::iterator it = TaskList.begin(); it != TaskList.end();)
	{
		if((*it)->SignalStack.size() > 0)
		{
			ReturnValue = (*it)->MessageLoop();
			++it;
		}
		else if((*it)->ExecutionState == Task::RUNNING)
		{
			(*it)->theProfiler.StartSample();
			ReturnValue = (*it)->Process();
			(*it)->theProfiler.EndSample();
			++it;
			continue;
		}
		else if((*it)->ExecutionState == Task::PAUSED)
		{
			++it;
			continue;
		}
		else if((*it)->ExecutionState == Task::TERMINATED)
		{
			delete (*it);
			it = TaskList.erase(it); // Assign (it) to the next task after the one being erased.
			continue;
		};
	};

	return SE_OK;
};

Sith::System::Kernel::Kernel() : TaskCount(0)
{
	new KernelProc();
	return;
};

Sith::System::Kernel::~Kernel()
{
	TaskCount = 0;
	return;
};
