/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "resman.h"

Sith::Res::Keyframe::Keyframe (std::string FileName) : Sith::VFS::Resource(FileName) {
	return;
}

Sith::SITHERROR Sith::Res::Keyframe::ParseHeader () {
	if (GetWordToken() != "flags") {
		ConOut ("ERROR: Expected flags in keyframe header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	KeyframeHeader.Flags = GetIntegerToken();

	if (GetWordToken() != "type") {
		ConOut ("ERROR: Expected type in keyframe header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	KeyframeHeader.Type = GetIntegerToken();

	if (GetWordToken() != "frames") {
		ConOut ("ERROR: Expected frames in keyframe header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	KeyframeHeader.Frames = GetIntegerToken();

	if (GetWordToken() != "fps") {
		ConOut ("ERROR: Expected fps in keyframe header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	KeyframeHeader.FPS = GetFloatToken();

	if (GetWordToken() != "joints") {
		ConOut ("ERROR: Expected joints in keyframe header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	KeyframeHeader.Joints = GetIntegerToken();

	return SE_OK;
}

Sith::SITHERROR Sith::Res::Keyframe::ParseKeyframeNodes () {
	unsigned int i, j;

	if (GetWordToken() != "nodes") {
		ConOut ("ERROR: Expected nodes in keyframe nodes.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}
	NumNodes = GetIntegerToken();

	Nodes.resize (NumNodes);

	for (i = 0; i < NumNodes; i++) {
		if (GetWordToken() != "node") {
			ConOut ("ERROR: Expected node in keyframe nodes entry.\n", ConOut::Error);
			return SE_FILE_CORRUPT;
		}
		Nodes[i].Index = GetIntegerToken();

		if (GetWordToken() != "node") {
			ConOut ("ERROR: Expected mesh name in keyframe nodes entry.\n", ConOut::Error);
			return SE_FILE_CORRUPT;
		}
		Nodes[i].MeshName = GetContiguousToken();

		if (GetWordToken() != "entries") {
			ConOut ("ERROR: Expected entries in keyframe nodes entry.\n", ConOut::Error);
			return SE_FILE_CORRUPT;
		}
		Nodes[i].NumEntries = GetIntegerToken();

		Nodes[i].Entries.resize (Nodes[i].NumEntries);
		
		for (j = 0; j < Nodes[i].NumEntries; j++) {
			Nodes[i].Entries[j].Num = GetIntegerToken();
			if (GetPuncToken() != ":") {
				ConOut ("ERROR: Expected colon after entry index.\n", ConOut::Error);
				return SE_FILE_CORRUPT;
			}

			Nodes[i].Entries[j].Frame = GetIntegerToken();
			Nodes[i].Entries[j].Flags = GetIntegerToken();

			Nodes[i].Entries[j].PositionOffset.x = GetFloatToken();
			Nodes[i].Entries[j].PositionOffset.y = GetFloatToken();
			Nodes[i].Entries[j].PositionOffset.z = GetFloatToken();

			Nodes[i].Entries[j].OrientationOffset.x = GetFloatToken();
			Nodes[i].Entries[j].OrientationOffset.y = GetFloatToken();
			Nodes[i].Entries[j].OrientationOffset.z = GetFloatToken();

			Nodes[i].Entries[j].DeltaPosition.x = GetFloatToken();
			Nodes[i].Entries[j].DeltaPosition.y = GetFloatToken();
			Nodes[i].Entries[j].DeltaPosition.z = GetFloatToken();

			Nodes[i].Entries[j].DeltaOrientation.x = GetFloatToken();
			Nodes[i].Entries[j].DeltaOrientation.y = GetFloatToken();
			Nodes[i].Entries[j].DeltaOrientation.z = GetFloatToken();
		}
	}

	return SE_OK;
}




Sith::SITHERROR Sith::Res::Keyframe::LoadFromFile(const char* FileBuffer)
{
	CurrentOffset = 0;
	LineNumber = 0;
	EndOfStream = false;
	Buffer = FileBuffer;
	std::string TokenBuffer;

	BufferLen = strlen(FileBuffer);

	while (CurrentOffset < BufferLen) {
		SkipToNextSection();
		CurrentOffset += 9;
		TokenBuffer = GetWordToken();
		if (TokenBuffer == "header") {
			if(Failed(ParseHeader())) {
				ConOut ("ERROR: Failed to parse keyframe header.\n", ConOut::Error);
				return SE_FILE_CORRUPT;
			}
		} else if (TokenBuffer == "markers") {
			// jkspecs has no idea what this section does, let's ignore it for now
		} else if (TokenBuffer == "keyframe" && GetWordToken() == "nodes") {
			if(Failed(ParseKeyframeNodes())) {
				ConOut ("ERROR: Failed to parse keyframe nodes.\n", ConOut::Error);
				return SE_FILE_CORRUPT;
			}
		} else {
			ConOut ("ERROR: Unknown section in keyframe file.\n", ConOut::Error);
			return SE_FILE_CORRUPT;
		}
	}

	return SE_OK;
}


