/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ktask.h"
#include "kernel.h"

// MessageLoop();
// - Loops through list of incoming messages
Sith::SITHERROR Sith::System::Task::MessageLoop()
{
	SITHERROR ReturnValue = SE_OK;
	while(!SignalStack.empty())
	{
		ReturnValue = MessageHandler(SignalStack.front());
		if(Failed(ReturnValue))
		{
			return ReturnValue;
		};

		SignalStack.pop();
	};
	return SE_OK;
};

// MessageHandler();
// - Overridable message handling.
Sith::SITHERROR Sith::System::Task::MessageHandler(Sith::KSIGNAL Signal)
{
	return DefaultMessageHandler(Signal);
};

// DefaultMessageHandler();
// - Implements the default behavior for most operations.
Sith::SITHERROR Sith::System::Task::DefaultMessageHandler(Sith::KSIGNAL Signal)
{
	switch(Signal)
	{
	case KS_TERM:
		ExecutionState = TERMINATED;
		return SE_OK;

	case KS_KILL:
		ExecutionState = TERMINATED;
		return SE_OK;

	case KS_PAUSE:
		ExecutionState = PAUSED;
		return SE_OK;

	case KS_START:
		ExecutionState = RUNNING;
		Startup();
		return SE_OK;

	case KS_NULL:
	case KS_QUIT:
	default:
		return SE_OK;
	};

	return SE_OK;
};

Sith::System::Task::Task(std::string Name) : TaskName(Name), ProcessID(0), ExecutionState(RUNNING), theProfiler(Name)
{
	ProcessID = System::Kernel::Get().RegisterTask(this);
	return;
};

Sith::SITHERROR Sith::System::Task::Startup()
{
	return SE_OK;
};

Sith::System::Task::~Task()
{
	return;
};