/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _KTASK_H_INCLUDED_
#define _KTASK_H_INCLUDED_

namespace Sith
{
	namespace System
	{
		// Task class
		// - Defines a task, blocking state(s) and lifespan
		class Task
		{
			friend class Kernel;
		private:
			Profiler theProfiler;

		protected:
			typedef enum KEXECUTIONSTATE
			{
				RUNNING,	// task is actively running
				PAUSED,		// task is paused and may be resumed
				TERMINATED	// task has been terminated and is awaiting destruction
			};

			const std::string TaskName;	// System name of the task (should be something recognizable if you do lspid() or somesuch
			std::list<Task*>::iterator ProcessID; // Process ID - should be assigned with the magic of the kernel

			KEXECUTIONSTATE ExecutionState;

			std::queue<KSIGNAL> SignalStack;

			SITHERROR MessageLoop();

			virtual SITHERROR Startup();
			virtual SITHERROR Process() = 0;

			SITHERROR DefaultMessageHandler(KSIGNAL Signal);
			virtual SITHERROR MessageHandler(KSIGNAL Signal);

			inline void SendSignal(KSIGNAL Signal)
			{
				SignalStack.push(Signal);
			};

		protected:
			Task(std::string Name);
			virtual ~Task();
		};
	};
};

#endif // _KTASK_H_INCLUDED_
