/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "logger.h"

Sith::Logger::Logger() : Task("System Logger")
{
	Sith::COG::CogLib::InjectNamespace("log");
	Sith::COG::CogLib::AddSetting("log::logfile", Sith::COG::VMType(std::string("debug.log")));
	Sith::COG::CogLib::AddSetting("log::profile", Sith::COG::VMType(std::string("profile.log")));

	ConOut::RegisterOutput(Stream, ConOut::Warning | ConOut::Error | ConOut::Critical | ConOut::Standard | ConOut::Network | ConOut::Echo | ConOut::Pedantic);
};

Sith::Logger::~Logger()
{
	if(LogFile)
	{
		fclose(LogFile);
	};

	return;
};

Sith::SITHERROR Sith::Logger::Startup()
{
	LogFile = 0;

	if(Sith::COG::CogLib::GetSetting("log::logfile").GetType() == COG::VMTC_STRING)
	{
		LogFile = fopen(Sith::COG::CogLib::GetSetting(std::string("log::logfile")).GetString().c_str(), "w");
		if(!LogFile)
		{
			ConOut("ERROR: Failed to create log file.\n", ConOut::Error);
			return SE_GENERAL_ERROR;
		}
		else
		{
			ConOut("Logging started.\n\n", ConOut::Pedantic);
			return SE_OK;
		};
	}
	else
	{
		ConOut("ERROR: Could not open log file; valid path not specified\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Logger::Process()
{
	if((LogFile) && (Stream.str().length() > 0))
	{
		fwrite(Stream.str().c_str(), Stream.str().length() * sizeof(char), 1, LogFile);
		Stream.str("");
	};

	return SE_OK;
};