/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "resman.h"

Sith::Res::Material::Material(std::string Name) : Sith::VFS::Resource(Name)
{
	xscale = 1.0f;
	yscale = 1.0f;
	return;
};

Sith::Res::Material::~Material()
{
	return;
};

Sith::SITHERROR Sith::Res::Material::LoadFromFileMat3(FILE* FileHandle)
{
	fread(&xscale, sizeof(float), 1, FileHandle);
	fread(&yscale, sizeof(float), 1, FileHandle);
	int seqcount = 0;
	fread(&seqcount, sizeof(__int32), 1, FileHandle);
	BitmapData.resize(seqcount);
	for (int n = 0; n < seqcount; n++)
	{
		int metacount = 0;
		fread(&metacount, sizeof(__int32), 1, FileHandle);
		for (int i = 0; i < metacount; i++)
		{
			char metaname[32];
			fread(metaname, 32, 1, FileHandle);
			int chunksize = 0;
			fread(&chunksize, sizeof(__int32), 1, FileHandle);			

			if(strncmp(metaname, "script", 6) == 0)
			{
				char* cogbuffer = new char[chunksize + 1];
				fread(cogbuffer, chunksize, 1, FileHandle);
				cogbuffer[chunksize] = '\0';
				if(Success(Script.Parse(cogbuffer)))
				{
					Scripted = true;
				}
				else
				{
					Scripted = false;
				};

				delete[] cogbuffer;
			}
			else if(BitmapData[n].CreateDDS(FileHandle, metaname) != SE_OK)
				return SE_GENERAL_ERROR;
		}
	}
	return SE_OK;
}

Sith::SITHERROR Sith::Res::Material::LoadFromFile(FILE* FileHandle, std::string CMPName)
{
	ColormapSystemName = CMPName;
	Scripted = false;

	// check for mat3
	char mat3header[8] = {0x4d, 0x41, 0x54, 0x20, 0x30, 0x00, 0x00, 0x00};
	char headerstart[8];
	fread(headerstart, 8, 1, FileHandle);
	if (memcmp(headerstart, mat3header, 8) == 0)
		return LoadFromFileMat3(FileHandle);

	// load a normal mat
	MATHEADER Header;
	memcpy(&Header, headerstart, 8);
	fread(((unsigned char *)&Header) + 8, sizeof(MATHEADER) - 8, 1, FileHandle);

	BitmapData.resize(Header.MatRecordCount, Win32::Hardware_Bitmap());

	std::list<MATCOLORRECORDHEADER> ColorRecord;
	std::list<MATRECORDHEADER> Record;
	std::list<TEXTUREDATAHEADER> TexDat;
	if (Header.Type == 0) // Color Mat
	{
		MATCOLORRECORDHEADER temp;
		TEXTUREDATAHEADER temp2;
		temp2.SizeX = 1;
		temp2.SizeY = 1;
		temp2.TransparentBool = 0;
		temp2.MipmapCount = 0;

		for(int i = 0; i < Header.MatRecordCount; ++i)
		{
			fread(&temp, sizeof(MATCOLORRECORDHEADER), 1, FileHandle);
			ColorRecord.push_back(temp);
			temp2.SizeX = 1;
			temp2.SizeY = 1;
			temp2.TransparentBool = 0;
			temp2.MipmapCount = 0;
			TexDat.push_back(temp2);
		};
	}
	else // Texture Mat
	{
		MATRECORDHEADER temp;

		// read in the cell headers
		for (int n = 0; n < Header.MatRecordCount; n++)
		{
			fread(&temp, sizeof(MATRECORDHEADER), 1, FileHandle);
			Record.push_back(temp);
		}
	}

	for(int currcel = 0; currcel < Header.MatRecordCount; ++currcel)
	{
		if(Header.Type != 0)
		{
			TEXTUREDATAHEADER temp;
			fread(&temp, sizeof(TEXTUREDATAHEADER), 1, FileHandle);
			TexDat.push_front(temp);
		};

		Win32::HBMGeometry Geo(TexDat.front().SizeX, TexDat.front().SizeY, Sith::Win32::HBMF_BGRA8, true);
		BitmapData[currcel].Create(&Geo);

		BitmapData[currcel].Lock(&Geo);

		VFS::WisePointer<Res::Colormap> Palette = VFS::ResourceManager::Get().GetColormap(CMPName);

		// index buffer, for color and 8-bit
		unsigned char* Buffer = 0;
		if (Header.Type == 0 || Header.Bitdepth == 8) // not supporting 16-bit color only mats?
			Buffer = new unsigned char[Geo.xdim * Geo.ydim];
		unsigned __int32* Tex = new unsigned __int32[Geo.xdim * Geo.ydim];

		if (Header.Type == 0) // Color Mat
		{
			Buffer[0] = ColorRecord.front().ColorNum;

			Tex[0] = Palette->Get(Buffer[0]);
			// 8-bit mat, check if transparency is enabled and if the index is the transparent index
			if (Header.Type == 2 && TexDat.front().TransparentBool && (Buffer[0] == Record.front().TransparentColor || Tex[0] == 0xFF000000))
			{
				*((unsigned char *)&Tex[0] + 3) = 0;
			}

			ColorRecord.pop_front();
		}
		else if (Header.Bitdepth == 8)
		{
			fread(Buffer, sizeof(unsigned char), Geo.xdim * Geo.ydim, FileHandle);

			for(unsigned int i = 0; i < (Geo.xdim * Geo.ydim); ++i)
			{
				Tex[i] = Palette->Get(Buffer[i]);
				// 8-bit mat, check if transparency is enabled and if the index is the transparent index
				if (Header.Type == 2 && TexDat.front().TransparentBool && (Buffer[i] == Record.front().TransparentColor || Tex[i] == 0xFF000000))
				{
					*((unsigned char *)&Tex[i] + 3) = 0;
				}
			};

			// skip past mipmaps
			for(int np = 1; np < TexDat.front().MipmapCount; ++np)
			{
                TexDat.front().SizeX /= 2;
				TexDat.front().SizeY /= 2;

				fseek(FileHandle, TexDat.front().SizeX * TexDat.front().SizeY, SEEK_CUR);
			};

			TexDat.pop_front();
		}
		else if (Header.Bitdepth == 16)
		{
			unsigned short *temp = new unsigned short [Geo.xdim * Geo.ydim];
			fread(temp, sizeof(unsigned short), Geo.xdim * Geo.ydim, FileHandle);
			// convert from 16 to 32 bit BGRA
			// not sure if transparent 16-bit actually works at the moment
			if (TexDat.front().TransparentBool) // 1555
			{
				for (int n = 0; n < Geo.xdim * Geo.ydim; n++)
				{
/*					Tex[n] = (int)((float)((temp[n] & 0x003E) >> 1) * (255.0f / 31.0f)) & 0x000000FF
						| (int)((float)((temp[n] & 0x07C0) >> 6) * (255.0f / 31.0f)) << 8 & 0x0000FF00
						| (int)((float)((temp[n] & 0xF800) >> 11) * (255.0f / 31.0f)) << 16 & 0x00FF0000
						| (temp[n] & 0x0001) * 0xFF000000;*/
					Tex[n] = (int)((float)((temp[n] & 0x001F)) * (255.0f / 31.0f)) & 0x000000FF
						| (int)((float)((temp[n] & 0x03E0) >> 5) * (255.0f / 31.0f)) << 8 & 0x0000FF00
						| (int)((float)((temp[n] & 0x7C00) >> 10) * (255.0f / 31.0f)) << 16 & 0x00FF0000
						| ((temp[n] & 0x8000) != 0) * 0xFF000000;
					if (temp[n] == Record.front().TransparentColor)
						Tex[n] &= 0x00FFFFFF;
				}
			}
			else // 565
			{
				for (int n = 0; n < Geo.xdim * Geo.ydim; n++)
				{
					Tex[n] = (int)((float)(temp[n] & 0x001F) * (255.0f / 31.0f)) & 0x000000FF
						| (int)((float)((temp[n] & 0x07E0) >> 5) * (255.0f / 63.0f)) << 8 & 0x0000FF00
						| (int)((float)((temp[n] & 0xF800) >> 11) * (255.0f / 31.0f)) << 16 & 0x00FF0000
						| 0xFF000000;
				}
			}
			delete [] temp;

			// skip past mipmaps
			for(int np = 1; np < TexDat.front().MipmapCount; ++np)
			{
                TexDat.front().SizeX /= 2;
				TexDat.front().SizeY /= 2;

				fseek(FileHandle, TexDat.front().SizeX * TexDat.front().SizeY * 2, SEEK_CUR);
			};

			TexDat.pop_front();
			Record.pop_front();
		}

		xscale = 1.0f;
		yscale = 1.0f;
		memcpy(Geo.bitdata, Tex, sizeof(__int32) * Geo.xdim * Geo.ydim);

		if (Buffer)
			delete[] Buffer;
		delete[] Tex;

		BitmapData[currcel].Unlock();
	};

	return SE_OK;
};