/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _MATRES_H_INCLUDED_
#define _MATRES_H_INCLUDED_

namespace Sith
{
	namespace Res
	{
		class Material : public VFS::Resource
		{
		private:
			struct MATHEADER
			{
				char Magic[4];				// Must/should be 'MAT '
				__int32 Version;			// Should be 0x32

				__int32 Type;				// 0 = color, 1 = ?, 2 = texture
				__int32 MatRecordCount;		// Number of MAT records
				__int32 TextureCount;		// Number of textures

				// Surface format should be built from the below values:
				__int32 Transparency;		// Unknown (perhaps transparency?)

				__int32 Bitdepth;			// Number of bits per pixel, valid formats: 8 (converted), 16, 32

				__int32 BlueBits;			// 0, 5, 8
				__int32 GreenBits;			// 0, 6 (16-bit 565), 5 (16-bit 1555), 8
				__int32 RedBits;			// 0, 5, 8

				// If Bluebits + greenbits + redbits don't add up to bitdepth, drop format down:
				// i.e., transparency + 32-bit + 8 8 8 = A8R8G8B8
				// no transparency + 32-bit + 8 8 8 = X8R8G8B8
				// no transparency + 24-bit 8 8 8 = R8G8B8 (no control, 24-bit defaults)
				// transparency + 16-bit 5 5 5 = A1R5G5B5
				// no transparency + 16-bit 5 5 5 = X1R5G5B5
				// any transparency + 16-bit 5 6 5 = R5G6B5
				// transparency + 8-bit any = upconverted to 32-bit A8R8G8B8
				// no transparency + 8-bit any = upconverted to 24-bit R8G8B8

				__int32 RedShl;
				__int32 GreenShl;
				__int32 BlueShl;

				__int32 RedShr;
				__int32 GreenShr;
				__int32 BlueShr;

				__int32 Unknown1;
				__int32 Unknown2;
				__int32 Unknown3;
			};

			struct MATRECORDHEADER
			{
				__int32 RecordType;			// 0 = color, 8 = texture
				__int32 TransparentColor;	// if 16-bit or 32-bit, RGB. If 8-bit, palette offset
				__int32 Unknown1;			// Padding, most likely
				__int32 Unknown2;			// More padding
				__int32 Unknown3;			// Yet more padding

				__int32 Unknown4;			// Further padding

				__int32 Unknown5;
				__int32 Unknown6;
				__int32 Unknown7;
				__int32 Unknown8;
			};

			struct TEXTUREDATAHEADER
			{
				__int32 SizeX;
				__int32 SizeY;
				__int32 TransparentBool;	// Use transparency? std bool, 0 = false, !0 = true
				
				__int32 Unknown1;
				__int32 Unknown2;

				__int32 MipmapCount;		// Ignored; using hardware mipmapping
			};

			struct MATCOLORRECORDHEADER
			{
				__int32 RecordType; // 0
				__int32 ColorNum; // index into palette
				float Unknown1;
				float Unknown2;
				float Unknown3;
				float Unknown4;
			};

			// Followed by texture data

			std::vector<Win32::Hardware_Bitmap> BitmapData;
			std::string ColormapSystemName;

			SITHERROR LoadFromFileMat3(FILE* FileHandle);

		public:
			float xscale, yscale;

			COG::StackMachine Script;
			bool Scripted;

			Material(std::string Name);
			~Material();

			inline std::string& GetColormapName()
			{
				return ColormapSystemName;
			};

			inline Win32::Hardware_Bitmap& GetHBM(int CellNumber)
			{
				if((CellNumber >= BitmapData.size()) || (CellNumber < 0)) CellNumber = 0;
				return BitmapData[CellNumber];
			};

			inline int GetHBMCount()
			{
				return BitmapData.size();
			};

			SITHERROR LoadFromFile(FILE* FileHandle, std::string ColormapSystemName);
		};
	};
};

#endif // _MATRES_H_INCLUDED_