/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "worldstate.h"
#include "oal_sound.h"

Sith::Win32::Sound* Sith::Win32::Sound::StaticInterface = 0;

Sith::Win32::Sound::Sound() : Task("OpenAL Audio Handler")
{
	// Set static interface
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple audio handler objects for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	Sith::COG::CogLib::InjectLibraryFunction("playsoundpos", 6, &PlaySoundPos);
	Sith::COG::CogLib::InjectLibraryFunction("playsoundlocal", 4, &PlaySoundLocal);
	Sith::COG::CogLib::InjectLibraryFunction("playsoundthing", 6, &PlaySoundThing);

	pDevice = NULL;
	pContext = NULL;

	strcpy((char *)devicename, "DirectSound3D");
//	strcpy((char *)devicename, "DirectSound");
//	strcpy((char *)devicename, "MMSYSTEM");
	pDevice = alcOpenDevice((char *)devicename); // NULL chooses the default device

	pContext = alcCreateContext(pDevice, NULL);
	alcMakeContextCurrent(pContext);

/*	streamvolume = 1;
	streamloaded = 0;
	streambuffers = 3;
	alGenBuffers(3, streambuffer);
	alGenSources(1, &streamsource);*/


	// sector ambient
	alGenSources(1, &sectorsource[0].id);
	alGenSources(1, &sectorsource[1].id);
	sectorsource[0].bufferid = -1;
	sectorsource[1].bufferid = -1;
	memset(&sectorfade, 0, sizeof(SoundFade) * 2);
	alSourcef(sectorsource[0].id, AL_ROLLOFF_FACTOR, 0);
	alSourcef(sectorsource[1].id, AL_ROLLOFF_FACTOR, 0);
	alSourcei(sectorsource[0].id, AL_LOOPING, true);
	alSourcei(sectorsource[1].id, AL_LOOPING, true);
	alSourcei(sectorsource[0].id, AL_SOURCE_RELATIVE, true);
	alSourcei(sectorsource[1].id, AL_SOURCE_RELATIVE, true);
	Math::Vector v(0, 0, 0);
	alSourcefv(sectorsource[0].id, AL_POSITION, &v.x);
	alSourcefv(sectorsource[1].id, AL_POSITION, &v.x);
	alSourcefv(sectorsource[0].id, AL_VELOCITY, &v.x);
	alSourcefv(sectorsource[1].id, AL_VELOCITY, &v.x);


//	alDistanceModel(AL_NONE);
	alDistanceModel(AL_INVERSE_DISTANCE_CLAMPED);
}

Sith::Win32::Sound::~Sound()
{
/*	if (streamloaded)
		StreamStop();*/

	Unload();

	alcMakeContextCurrent(NULL);
	if (pContext)
	{
		alcDestroyContext(pContext);
		pContext = NULL;
	}
	if (pDevice)
	{
		alcCloseDevice(pDevice);
		pDevice = NULL;
	}

	StaticInterface = 0;
}

void Sith::Win32::Sound::Unload()
{
	if (source.size())
	{
		for (int n = 0; n < source.size(); n++)
		{
			alSourceStop(source[n].id);
			alDeleteSources(1, &source[n].id);
		}
		source.clear();
	}
	if (buffer.size())
	{
		for (int n = 0; n < buffer.size(); n++)
			alDeleteBuffers(1, &buffer[n].id);
		buffer.clear();
	}
	alSourceStop(sectorsource[0].id);
	alSourceStop(sectorsource[1].id);
	sectorsource[0].bufferid = -1;
	sectorsource[1].bufferid = -1;
}

Sith::SITHERROR Sith::Win32::Sound::Process()
{
	frametime = timer.Query();

	if (level != Gameplay::WorldState::Get().CurrentLevel)
	{
		// what about switching to another level instantly?

		if (level.isValid() && !Gameplay::WorldState::Get().CurrentLevel.isValid())
		{
			level = 0;
			// unload
			Unload();
		}
		else
		{
			Unload();
			// load
			level = Gameplay::WorldState::Get().CurrentLevel;
		}
	}

	if (level.isValid())
	{
		// handle sector ambient fading
		for (int i = 0; i < 2; i++)
		{
			if (sectorsource[i].bufferid == -1)
				continue;
			if (sectorfade[i].time < sectorfade[i].totaltime)
			{
				// going from off to on
				if (sectorfade[i].time == 0)
				{
					if (sectorfade[i].gainstart < .0001f && sectorfade[i].gaintarget >= .0001f)
					{
						int state;
						alGetSourcei(sectorsource[i].id, AL_SOURCE_STATE, &state);
						if (state == AL_STOPPED || state == AL_INITIAL) // AL_PAUSED
						{
							alSourcef(sectorsource[i].id, AL_GAIN, 0);
							alSourcePlay(sectorsource[i].id);
						}
					}
				}

				sectorfade[i].time += frametime;
				if (sectorfade[i].time >= sectorfade[i].totaltime)
				{
					sectorfade[i].gain = sectorfade[i].gaintarget;
					sectorfade[i].pitch = sectorfade[i].pitchtarget;

					// going from on to off, stop at the end
					if (sectorfade[i].gain < .0001f && sectorfade[i].gainstart >= .0001f)
					{
						alSourcef(sectorsource[i].id, AL_GAIN, 0);
						alSourceStop(sectorsource[i].id);
					}
				}
				else
				{
					float c = sectorfade[i].time / sectorfade[i].totaltime;
					if (c > 1)
						c = 1;
					else if (c < 0)
						c = 0;
					sectorfade[i].gain = sectorfade[i].gainstart * (1 - c) + sectorfade[i].gaintarget * c;
					sectorfade[i].pitch = sectorfade[i].pitchstart * (1 - c) + sectorfade[i].pitchtarget * c;
				}

				sectorsource[i].gain = sectorfade[i].gain;
				sectorsource[i].pitch = sectorfade[i].pitch;
				alSourcef(sectorsource[i].id, AL_GAIN, sectorfade[i].gain);
				alSourcef(sectorsource[i].id, AL_PITCH, sectorfade[i].pitch);
			}
		}
	}

	return SE_OK;
}

void Sith::Win32::Sound::AddSound3D(std::string name, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop)
{
//	int b = LoadSound(name);
//	AddSource(b, pitch, gain, position, velocity, loop);
}

void Sith::Win32::Sound::SetSectorSound(std::string name, float pitch, float gain)
{
	if (!name.length())
	{
		sectorfade[0].gainstart = sectorsource[0].gain;
		sectorfade[0].gain = sectorsource[0].gain;
		sectorfade[0].gaintarget = 0;
		sectorfade[0].pitchstart = sectorsource[0].pitch;
		sectorfade[0].pitch = sectorsource[0].pitch;
		sectorfade[0].pitchtarget = sectorsource[0].pitch;
		sectorfade[0].totaltime = 1.5f;
		sectorfade[0].time = 0;

		sectorfade[1].gainstart = sectorsource[1].gain;
		sectorfade[1].gain = sectorsource[1].gain;
		sectorfade[1].gaintarget = 0;
		sectorfade[1].pitchstart = sectorsource[1].pitch;
		sectorfade[1].pitch = sectorsource[1].pitch;
		sectorfade[1].pitchtarget = sectorsource[1].pitch;
		sectorfade[1].totaltime = 1.5f;
		sectorfade[1].time = 0;

		return;
	}

	// check if nothing has changed
	if (sectorsource[0].bufferid != -1 && name == buffer[sectorsource[0].bufferid].name)
	{
		if (sectorsource[0].pitch != pitch || sectorsource[0].gain != gain)
		{
			sectorfade[0].gainstart = sectorsource[0].gain;
			sectorfade[0].gain = sectorsource[0].gain;
			sectorfade[0].gaintarget = gain;
			sectorfade[0].pitchstart = sectorsource[0].pitch;
			sectorfade[0].pitch = sectorsource[0].pitch;
			sectorfade[0].pitchtarget = pitch;
			sectorfade[0].totaltime = 1.5f;
			sectorfade[0].time = 0;
		}

		return;
	}

	int newbuffer = LoadSound(name);
	if (newbuffer != -1)
	{
		SoundSource temp;
		temp = sectorsource[0];
		sectorsource[0] = sectorsource[1];
		sectorsource[1] = temp;

		sectorsource[0].bufferid = newbuffer;
		alSourceStop(sectorsource[0].id);
		alSourcei(sectorsource[0].id, AL_BUFFER, buffer[sectorsource[0].bufferid].id);

		sectorfade[0].gainstart = 0;
		sectorfade[0].gain = 0;
		sectorfade[0].gaintarget = gain;
		sectorfade[0].pitchstart = pitch;
		sectorfade[0].pitch = pitch;
		sectorfade[0].pitchtarget = pitch;
		sectorfade[0].totaltime = 1.5f;
		sectorfade[0].time = 0;

		sectorfade[1].gainstart = sectorsource[1].gain;
		sectorfade[1].gain = sectorsource[1].gain;
		sectorfade[1].gaintarget = 0;
		sectorfade[1].pitchstart = sectorsource[1].pitch;
		sectorfade[1].pitch = sectorsource[1].pitch;
		sectorfade[1].pitchtarget = sectorsource[1].pitch;
		sectorfade[1].totaltime = 1.5f;
		sectorfade[1].time = 0;
	}
}

int Sith::Win32::Sound::LoadSound(std::string name)
{
	for (int n = 0; n < buffer.size(); n++)
	{
		if (name == buffer[n].name)
			return n;
	}

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("sound\\" + name);
	if (!FileBuffer)
		return -1;

	SoundBuffer b;
	b.name = name;
	alGenBuffers(1, &b.id);
	buffer.push_back(b);

	int format = 0;
	void *data = 0;
	unsigned int size = 0;
	unsigned int freq = 0;
	char loop = 0;
	alutLoadWAVMemory((ALbyte*)FileBuffer, &format, &data, &size, &freq, &loop);
	alBufferData(buffer.back().id, format, data, size, freq);
	alutUnloadWAV(format, data, size, freq);

	delete [] FileBuffer;

	return buffer.size() - 1;
}

/*void Sith::Win32::Sound::StreamLoad(char *filename)
{
	char fullname[256];
	sprintf(fullname, "sounds\\%s", filename);
//	int format = 0;
//	void *data = 0;
//	unsigned int size = 0;
//	unsigned int freq = 0;
//	char loop = 0;

	streamvf = new OggVorbis_File;

	FILE *file = fopen(fullname, "rb");
	ov_open(file, streamvf, 0, 0);

	streamvi = ov_info(streamvf, -1);
	if (streamvi->channels == 1)
		streamformat = AL_FORMAT_MONO16;
	else
		streamformat = AL_FORMAT_STEREO16;
	streamfreq = streamvi->rate;

	unsigned char *dbuffer;
	buffersize = 1024 * 32;
	dbuffer = new unsigned char [buffersize];

	int bitstream = 0;
	int bytesread = 0;
	int size = 0;

	// not checking for eof: bytesread == 0
	for (int n = 0; n < streambuffers; n++)
	{
		size = 0;
		do
		{
			bytesread = ov_read(streamvf, (char *)dbuffer + size, buffersize - size, 0, 2, 1, &bitstream);
			size += bytesread;
		} while (bytesread > 0 && size < buffersize);
		alBufferData(streambuffer[n], streamformat, dbuffer, size, streamfreq);
		alSourceQueueBuffers(streamsource, 1, &streambuffer[n]);
	}

	currentbuffer = 0;
	streamfinished = 0;
	streamloaded = 1;

	streamrestart = 0;

	int state;
	alGetSourcei(streamsource, AL_SOURCE_STATE, &state);
	if (state == AL_STOPPED)
		streamrestart = 1;

	delete [] dbuffer;
}

void Sith::Win32::Sound::StreamPlay()
{
	if (streamfinished)
		return;

	alSourcef(streamsource, AL_GAIN, streamvolume);

	int state;
	alGetSourcei(streamsource, AL_SOURCE_STATE, &state);
	if (streamrestart)
	{
		streamrestart = 0;
		alSourcePlay(streamsource);
	}
	if (state == AL_PAUSED || state == AL_INITIAL)
		alSourcePlay(streamsource);
	StreamUpdate();
}

void Sith::Win32::Sound::StreamUpdate()
{
	int processed;
	alGetSourcei(streamsource, AL_BUFFERS_PROCESSED, &processed);

	if (processed > 0)
	{
		if (processed > streambuffers)
			processed = processed;
		unsigned char *dbuffer;
//		buffersize = 4096;
		int size = 0;
		dbuffer = new unsigned char [buffersize];

		int bitstream = 0;
		int bytesread = 0;

		for (int n = 0; n < processed; n++)
		{
			unsigned int unqueued = -1;
			alSourceUnqueueBuffers(streamsource, 1, &unqueued);
			if (unqueued <= 0)
				continue;

			size = 0;
			do
			{
				bytesread = ov_read(streamvf, (char *)dbuffer + size, buffersize - size, 0, 2, 1, &bitstream);
				size += bytesread;
			} while (bytesread > 0 && size < buffersize);
			if (bytesread > 0)
			{
				alBufferData(unqueued, streamformat, dbuffer, size, streamfreq);
				alSourceQueueBuffers(streamsource, 1, &unqueued);

				int state;
				alGetSourcei(streamsource, AL_SOURCE_STATE, &state);
				if (state == AL_STOPPED)
					alSourcePlay(streamsource);
			}
			else
			{
				int state;
				alGetSourcei(streamsource, AL_SOURCE_STATE, &state);
				if (state == AL_STOPPED)
					streamfinished = 1;
			}
		}

		delete [] dbuffer;
	}
}

void Sith::Win32::Sound::StreamPause()
{
	alSourcePause(streamsource);
}

void Sith::Win32::Sound::StreamStop()
{
	if (streamloaded == 1)
	{
		alSourceStop(streamsource);
		alSourcei(streamsource, AL_BUFFER, 0);
		ov_clear(streamvf);
		streamloaded = 0;
	}
}

void Sith::Win32::Sound::StreamVolume(float volume)
{
	streamvolume = volume;
	alSourcef(streamsource, AL_GAIN, streamvolume);
}*/

int Sith::Win32::Sound::AddSource(unsigned int b, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop)
{
	SoundSource s;
	s.pos = position;
	alGenSources(1, &s.id);

//	alSourcef(source[sources - 1], AL_ROLLOFF_FACTOR, (float)1 / 100);

	source.back().bufferid = b;
	source.back().pos = position;
	source.back().vel = velocity;
	source.back().pitch = pitch;
	source.back().gain = gain;

	alSourcei(source.back().id, AL_BUFFER, buffer[b].id);
	alSourcef(source.back().id, AL_PITCH, pitch);
	alSourcef(source.back().id, AL_GAIN, gain);
	alSourcefv(source.back().id, AL_POSITION, &position.x);
	alSourcefv(source.back().id, AL_VELOCITY, &velocity.x);
	alSourcei(source.back().id, AL_LOOPING, loop);
	return source.size() - 1;
}

void Sith::Win32::Sound::SetSource(int n, unsigned int b, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop)
{
	source[n].bufferid = b;
	source[n].pos = position;
	source[n].vel = velocity;
	source[n].pitch = pitch;
	source[n].gain = gain;

	alSourcei(source[n].id, AL_BUFFER, buffer[b].id);
	alSourcef(source[n].id, AL_PITCH, pitch);
	alSourcef(source[n].id, AL_GAIN, gain);
	alSourcefv(source[n].id, AL_POSITION, &position.x);
	alSourcefv(source[n].id, AL_VELOCITY, &velocity.x);
	alSourcei(source[n].id, AL_LOOPING, loop);
}

void Sith::Win32::Sound::SetListener(Math::Vector position, Math::Vector velocity)
{
	alListenerfv(AL_POSITION, &position.x);
	alListenerfv(AL_VELOCITY, &velocity.x);
//	alListenerfv(AL_ORIENTATION, orientation);
}

void Sith::Win32::Sound::PlaySource(unsigned int s)
{
	alSourcePlay(source[s].id);
}

void Sith::Win32::Sound::StopSource(unsigned int s)
{
	alSourceStop(source[s].id);
}

void Sith::Win32::Sound::PauseSource(unsigned int s)
{
	alSourcePause(source[s].id);
}

void Sith::Win32::Sound::SetDoppler(float factor, float velocity)
{
	alDopplerFactor(factor);
	alDopplerVelocity(velocity);
}