/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _OAL_SOUND_H_INCLUDED_
#define _OAL_SOUND_H_INCLUDED_

#pragma comment (lib, "openal32")
#pragma comment (lib, "alut")
//#pragma comment (lib, "vorbisfile_static")
//#pragma comment (lib, "ogg_static")

#include <al.h>
#include <alc.h>
#include <alut.h>
//#include <vorbis\vorbisfile.h>
//#include <vorbis\codec.h>

namespace Sith
{
	namespace Win32
	{
		struct SoundBuffer
		{
			unsigned int id;
			std::string name;
		};

		struct SoundSource
		{
			unsigned int id;
			unsigned int bufferid;
			Math::Vector pos, vel;
			float pitch, gain;
		};

		struct SoundFade
		{
			float pitch;
			float gain;
			float pitchstart;
			float gainstart;
			float pitchtarget;
			float gaintarget;
			float totaltime;
			float time;
		};

		class Sound : public System::Task
		{
			friend class Sith::System::KernelProc;
		private:
			static Sound* StaticInterface;

			// level pointer
			VFS::WisePointer<Res::Level> level;

			Sound();
			~Sound();

			Sith::SITHERROR Process();

			void Unload();

			ALCdevice *pDevice;
			unsigned char devicename[256];
			ALCcontext *pContext;

			std::vector<SoundBuffer>	buffer;
			std::vector<SoundSource>	source;
			SoundSource					sectorsource[2];
			SoundFade					sectorfade[2];

			Win32::Timer				timer;
			float						frametime;

/*			OggVorbis_File *streamvf;
			vorbis_info *streamvi;
			unsigned int streamformat, streamfreq;
			unsigned int streambuffer[3];
			unsigned int currentbuffer;
			int buffersize;
			unsigned int streamsource;
			int streambuffers;
			float streamvolume;
			int streamfinished;
			int streamloaded;
			int streamrestart;*/

			int LoadSound(std::string name);

/*			void StreamLoad(char *filename);
			void StreamPlay();
			void StreamUpdate();
			void StreamPause();
			void StreamStop();
			void StreamVolume(float volume);*/

			int AddSource(unsigned int b, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop);

			void SetListener(Math::Vector position, Math::Vector velocity);
			void SetSource(int n, unsigned int b, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop);

			void PlaySource(unsigned int s);
			void StopSource(unsigned int s);
			void PauseSource(unsigned int s);

			// doppler factor multiplies the doppler effect, velocity is speed of sound * value passed
			void SetDoppler(float factor, float velocity);

		public:

			void AddSound3D(std::string name, float pitch, float gain, Math::Vector position, Math::Vector velocity, int loop);
			void SetSectorSound(std::string name, float pitch, float gain);

			static inline Sound& Get()
			{
				return *StaticInterface;
			};

			COGLIBAPI PlaySoundPos(COG::VMType* sound, COG::VMType* pos, COG::VMType* vol, COG::VMType* minrad, COG::VMType* maxrad, COG::VMType* flags)
			{
				return new COG::VMType(0);
			};

			COGLIBAPI PlaySoundLocal(COG::VMType* sound, COG::VMType* vol, COG::VMType* pan, COG::VMType* flags)
			{
				return new COG::VMType(0);
			};

			COGLIBAPI PlaySoundThing(COG::VMType* sound, COG::VMType* thing, COG::VMType* vol, COG::VMType* minrad, COG::VMType* maxrad, COG::VMType* flags)
			{
				return new COG::VMType(0);
			};
		};
	};
};

#endif //_OAL_SOUND_H_INCLUDED_