/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "ogl_hwbm.h"
#include "nv_dds.h"

#include <gl\gl.h>
#include <gl\glext.h>

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::CreateDDS(FILE *file, char *metaname)
{
	format = HBMF_COMPRESSED;
	if (metaname == 0)
		meta.push_back("diffuse");
	else
		meta.push_back(metaname);

	nv_dds::CDDSImage image;
	if (!image.load(file, false))
	{
		ConOut("ERROR: Failed to load a compressed hardware bitmap.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	}

	x_res = image.get_width();
	y_res = image.get_height();

	texid.resize(texid.size() + 1);
	if (texid.back() == 0)
		glGenTextures(1, &texid.back());
	glBindTexture(GL_TEXTURE_2D, texid.back());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, Sith::Win32::Renderer::Get().anisotropy);
	if (!image.get_num_mipmaps())
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, true);
	if (!image.upload_texture2D())
	{
		ConOut("ERROR: Failed to upload a compressed hardware bitmap.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	}

	if (meta.back() == "diffuse")
		diffuse = texid.back();
	else if (meta.back() == "bump")
		bump = texid.back();
	else if (meta.back() == "gloss")
		gloss = texid.back();

	return SE_OK;
}

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Create(Sith::Win32::HBMGeometry *GeoData)
{
	GLint internalformat, uploadformat;

	switch(GeoData->format)
	{
	case HBMF_BGRA8:
		internalformat = GL_RGBA8;
		uploadformat = GL_BGRA_EXT;
		break;
	case HBMF_BGR8:
		internalformat = GL_RGB8;
		uploadformat = GL_BGR_EXT;
		break;
	case HBMF_RGBA8:
		internalformat = GL_RGBA8;
		uploadformat = GL_RGBA;
		break;
	case HBMF_RGB8:
		internalformat = GL_RGB8;
		uploadformat = GL_RGB;
		break;
	case HBMF_LUMINANCE8:
		internalformat = GL_LUMINANCE8;
		uploadformat = GL_LUMINANCE;
		break;
	case HBMF_LUMINANCE8_ALPHA8:
		internalformat = GL_LUMINANCE8_ALPHA8;
		uploadformat = GL_LUMINANCE_ALPHA;
		break;
	case HBMF_ALPHA8:
		internalformat = GL_ALPHA8;
		uploadformat = GL_ALPHA;
		break;
	default:
		ConOut("ERROR: Unknown hardware bitmap format.\n", ConOut::Error);

		GeoData->format = HBMF_BADFORMAT;
		GeoData->Pow2 = false;
		GeoData->xdim = 0;
		GeoData->ydim = 0;

		return SE_GENERAL_ERROR;
	}

	texid.resize(texid.size() + 1);
	if (texid.back() == 0)
		glGenTextures(1, &texid.back());
	glBindTexture(GL_TEXTURE_2D, texid.back());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, Sith::Win32::Renderer::Get().anisotropy);
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, true);
	glTexImage2D(GL_TEXTURE_2D, 0, internalformat, GeoData->xdim, GeoData->ydim, 0, uploadformat, GL_UNSIGNED_BYTE, GeoData->bitdata);

	x_res = GeoData->xdim;
	y_res = GeoData->ydim;
	format = GeoData->format;
	Pow2 = GeoData->Pow2;
	Locked = false;

	meta.push_back("diffuse");
	diffuse = texid.back();

	return SE_OK;
};

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Lock(Sith::Win32::HBMGeometry *GeoDataBuffer)
{
	if (format == HBMF_COMPRESSED)
	{
		ConOut("ERROR: Attempted to lock a compressed hardware bitmap.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	}

	if(Locked)
	{
		ConOut("WARNING: Attempted to lock a hardware bitmap, but it's already locked.\n", ConOut::Warning);
		return SE_GENERAL_WARNING;
	};


	if (texid.size() && texid[0])
	{
		int bytespp;
		switch (format)
		{
		case HBMF_BGRA8:
			bytespp = 4;
			break;
		case HBMF_BGR8:
			bytespp = 3;
			break;
		case HBMF_RGBA8:
			bytespp = 4;
			break;
		case HBMF_RGB8:
			bytespp = 3;
			break;
		case HBMF_LUMINANCE8:
			bytespp = 1;
			break;
		case HBMF_LUMINANCE8_ALPHA8:
			bytespp = 2;
			break;
		case HBMF_ALPHA8:
			bytespp = 1;
			break;
		default:
			ConOut("ERROR: Unknown hardware bitmap format on lock.\n", ConOut::Error);
			return SE_GENERAL_ERROR;
		}


		GeoDataBuffer->bitdata = malloc(x_res * y_res * bytespp);
		if (GeoDataBuffer->bitdata == 0)
		{
			ConOut("ERROR: Failed to allocate texture lock buffer, out of memory?.\n", ConOut::Error);
			return SE_GENERAL_ERROR;
		}
		GeoDataBuffer->format = format;
		GeoDataBuffer->xdim = x_res;
		GeoDataBuffer->ydim = y_res;
		GeoDataBuffer->Pow2 = Pow2;

		lock_pointer = GeoDataBuffer->bitdata;
		Locked = true;

		return SE_OK;
	}
	else
	{
		ConOut("ERROR: Attempted to lock a hardware bitmap for writing. Most likely a Create call failed but it wasn't checked.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Win32::Hardware_Bitmap::Unlock()
{
	if (format == HBMF_COMPRESSED)
	{
		ConOut("ERROR: Attempted to unlock a compressed hardware bitmap.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	}

	if(!Locked)
	{
		ConOut("WARNING: Attempted to unlock a hardware bitmap, but it isn't locked.\n", ConOut::Warning);
		return SE_GENERAL_WARNING;
	};

	if (!texid[0])
	{
		ConOut("ERROR: Attempted to unlock a hardware bitmap but it doesn't exist. Most likely a Create call failed but was never checked.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	};

	GLint uploadformat;
	switch(format)
	{
	case HBMF_BGRA8:
		uploadformat = GL_BGRA_EXT;
		break;
	case HBMF_BGR8:
		uploadformat = GL_BGR_EXT;
		break;
	case HBMF_RGBA8:
		uploadformat = GL_RGBA;
		break;
	case HBMF_RGB8:
		uploadformat = GL_RGB;
		break;
	case HBMF_LUMINANCE8:
		uploadformat = GL_LUMINANCE;
		break;
	case HBMF_LUMINANCE8_ALPHA8:
		uploadformat = GL_LUMINANCE_ALPHA;
		break;
	case HBMF_ALPHA8:
		uploadformat = GL_ALPHA;
		break;
	default:
		ConOut("ERROR: Unknown hardware bitmap format on unlock.\n", ConOut::Error);
		return SE_GENERAL_ERROR;
	}

	glBindTexture(GL_TEXTURE_2D, texid[0]);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, x_res, y_res, uploadformat, GL_UNSIGNED_BYTE, lock_pointer);

	if (lock_pointer)
		free(lock_pointer);

	Locked = false;

	return SE_OK;
};