/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "jkl_vis.h"
#include "uimain.h"


Sith::SITHERROR Sith::Win32::Renderer::Process()
{
	while(PeekMessage(&Msg, hMainWnd, 0U, 0U, PM_REMOVE) > 0)
	{
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
	};


	if (titleupdate > 30)
	{
		char text[256];
		int sectorcount = 0;
		if (level.isValid())
			sectorcount = level->SectorList.size();
		sprintf(text, "Sith 2: %.1f  Reflection Planes: %d Sectors: %d/%d", 1 / frametime, reflectionplanes, sectors_drawn, sectorcount);
		SetWindowText(hMainWnd, text);
		titleupdate = 0;
	}
	titleupdate++;

	if (level.isValid())
	{
		if (Sith::COG::CogLib::GetSetting("video::renderpath").GetInt())
		{
			// update the water sim
			water.Step(frametime);
		}
	}

	// flip here
	SwapBuffers(hDC);
	frametime = timer.Query();

	int ClearFlags = GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT;
	if (ClearEveryFrame)
	{
		glClearColor(0, 0, 0, 0);
		ClearFlags |= GL_COLOR_BUFFER_BIT;
	};

	glClear(ClearFlags);

	// water debugging
/*	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, wsizex, 0, wsizey, 0, 1);
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_TEXTURE_2D);
	glColor4f(1, 1, 1, 1);
	glRasterPos2f(0, 0);
	glDrawPixels(water.heightdim, water.heightdim, GL_LUMINANCE, GL_FLOAT, water.heightmap);
	glRasterPos2f(water.heightdim, 0);
	glDrawPixels(water.heightdim, water.heightdim, GL_RGB, GL_FLOAT, water.normalmap);*/


	float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();
	projection[4] = projection[8] = projection[12] = projection[1] = projection[9] = projection[13] = projection[2] = projection[6] = projection[3] = projection[7] = projection[15] = 0;
	projection[0] = (1 / tan((float)fov * .5f * 0.0174532925f)) / ((float)wsizex / (float)wsizey);
	projection[5] = (1 / tan((float)fov * .5f * 0.0174532925f));
	projection[14] = (float)-2 * nearclip; // near clip
	projection[10] = projection[11] = -1;
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projection);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	float jkm[16] = {
		1, 0, 0, 0,
		0, 0, -1, 0,
		0, 1, 0, 0,
		0, 0, 0, 1
	};

	glRotatef(-Camera.pitch, 1, 0, 0);
	glRotatef(-Camera.yaw, 0, 1, 0);

	// everything after this point is in jk's coordinate system
	glMultMatrixf(jkm);

	glTranslatef(-Camera.x, -Camera.y, -Camera.z);

	// replace this with a parallel copy of modelview kept in renderer
	glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	DrawLevel();

	// POV model
//	Camera.POVModel = VFS::ResourceManager::Get().GetModel("bryv.3do", "dflt.cmp");
	if (Camera.POVModel.isValid())
	{
		glPushMatrix();
		glLoadIdentity();
		glMultMatrixf(jkm);
		glTranslatef(-.0175f, 0, -.0175f);
		glRotatef(90, 1, 0, 0);
		Render3do(Camera.POVModel, Camera.sector);
		glPopMatrix();
	}

	// 2d stuff, disable depth writes and face culling
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(0);


	// tints
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);

	glDisable(GL_TEXTURE_2D);
	// multiply tint
	glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	glEnable(GL_BLEND);

	// cull out no effect and no tint
	if (Camera.MulTint != Math::VectorSet(1, 1, 1) && Camera.MulTint != Math::VectorSet(0, 0, 0))
	{
		glColor4f(Camera.MulTint.x, Camera.MulTint.y, Camera.MulTint.z, 1);
		glBegin(GL_QUADS);
			glVertex3f(-1, -1, 0);
			glVertex3f(1, -1, 0);
			glVertex3f(1, 1, 0);
			glVertex3f(-1, 1, 0);
		glEnd();
	}

	glDisable(GL_BLEND);


	// 2D overlay stuff
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, wsizex, wsizey, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);

	glAlphaFunc(GL_GREATER, .5f);
	glEnable(GL_ALPHA_TEST);

	glEnable(GL_TEXTURE_2D);

	// Set desktop widget to current window size:
	Sith::UI::UIMain::Get().GetRootWindow().SetRect(UI::UIRect(0, 0, wsizex, wsizey));

	DrawUIWidget(Sith::UI::UIMain::Get().RootWidget, Sith::UI::UIMain::Get().RootWidget);
	glDisable(GL_ALPHA_TEST);

	// turn on depth writes
	glDepthMask(1);

	return SE_OK;
};

void Sith::Win32::Renderer::DrawLevel()
{
	sectors_drawn = 0;
	if (level != Gameplay::WorldState::Get().CurrentLevel)
	{
		if (level.isValid() && !Gameplay::WorldState::Get().CurrentLevel.isValid())
		{
			clevel.Clear();
			level = 0;
			// unload
		}
		else
		{
			// load
			level = Gameplay::WorldState::Get().CurrentLevel;
			
			// Set up material cogs:
			for(std::vector<Res::Level::SURFACE>::iterator s = level->SurfaceList.begin(); s != level->SurfaceList.end(); ++s)
			{
				if((s->TextureIndex >= 0) && (s->TextureIndex < level->MatList.size()))
				{
					// Check if this surface's material has a script:
					if(level->MatList[s->TextureIndex]->Scripted)
					{
						// Instance script:
						COG::VMProcess::Get().AddScript(level->MatList[s->TextureIndex]->Script, COG::VMType(COG::VMTC_SURFACE, s->SurfaceNumber, -1));
					};
				};
			};

			CompileLevel();
		}
	}
	if (level.isValid())
	{
		int renderpath = Sith::COG::CogLib::GetSetting("video::renderpath").GetInt();
		if (renderpath == 0)
		{
			DrawLevelBasic();
			return;
		}
		if (renderpath == 1)
		{
			DrawLevelAdvanced();
			return;
		}
	}
}

void Sith::Win32::Renderer::OnResize(int width, int height, int xoffset, int yoffset)
{
	if (height <= 0)
		height = 1;
	if (width <= 0)
		width = 1;

	wsizex = width;
	wsizey = height;

	glViewport(xoffset, yoffset, width, height);
}