/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "jkl_vis.h"

void Sith::Win32::Renderer::AdvancedStartup()
{
	ShaderStartup();

	water.Start(116, 128, 120.0f, 120.0f, 9.8f, 6.0f, 6.0f, .0008f, .25f);
//	water.Start(116, 64, 75.0f, 75.0f, 9.8f, 6.0f, 6.0f, .0008f, .25f);

	// FBO
	framebuffersize = 1024;
	glGenFramebuffersEXT(1, &framebuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);

	glGenRenderbuffersEXT(1, &renderbufferdepth);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, renderbufferdepth);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, framebuffersize, framebuffersize);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, renderbufferdepth);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);


	glGenTextures(1, &refractiontex);
	glBindTexture(GL_TEXTURE_2D, refractiontex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, framebuffersize, framebuffersize, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// perhaps size this based on an initial best guess, or resize at runtime
	reflection_buffer_size = 0;
	reflect_tex.resize(reflection_buffer_size);
	for (int n = 0; n < reflect_tex.size(); n++)
		reflect_tex[n].GenTex(framebuffersize, framebuffersize);

	waterfovmod = 10;
	reflection_tolerance = .02f;
}

void Sith::Win32::Renderer::DrawLevelAdvanced()
{
	float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();

	// determine sector visibility
	std::vector<unsigned char> camera_vis;
	std::vector<int> sector_vis_start;
	sector_vis_start.push_back(Camera.sector);
	// not getting fov + waterfovmod...but it should be ok
	Sith::JKL::FindVis(sector_vis_start, level, &Camera.x, Camera.forward, Camera.right, Camera.up, fov, (float)wsizex / wsizey, camera_vis);


	// update the refraction tex (just an increased FOV version of the screen, can be reused by other effects)
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, refractiontex, 0);

	glViewport(0, 0, framebuffersize, framebuffersize);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	projection[4] = projection[8] = projection[12] = projection[1] = projection[9] = projection[13] = projection[2] = projection[6] = projection[3] = projection[7] = projection[15] = 0;
	projection[0] = (1 / tan((float)(fov + waterfovmod) * .5f * 0.0174532925f)) / ((float)wsizex / (float)wsizey);
	projection[5] = (1 / tan((float)(fov + waterfovmod) * .5f * 0.0174532925f));
	projection[14] = (float)-2 * nearclip; // near clip
	projection[10] = projection[11] = -1;
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixf(projection);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	float jkm[16] = {
		1, 0, 0, 0,
		0, 0, -1, 0,
		0, 1, 0, 0,
		0, 0, 0, 1
	};

	glRotatef(-Camera.pitch, 1, 0, 0);
	glRotatef(-Camera.yaw, 0, 1, 0);

	// everything after this point is in jk's coordinate system
	glMultMatrixf(jkm);

	glTranslatef(-Camera.x, -Camera.y, -Camera.z);

	// draw without water or skies
	DrawJKLAdvanced(false, false, sector_vis_start[0], camera_vis);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glViewport(0, 0, wsizex, wsizey);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();



	// draw the level
	DrawJKLAdvanced(true, true, sector_vis_start[0], camera_vis);
}

void Sith::Win32::Renderer::DrawJKLAdvanced(bool DrawReflect, bool DrawSky, int startsector, std::vector<unsigned char> sectorvis)
{
	int n;

	memset(clevel.p, 0, sizeof(int) * clevel.drawbins);
	int baseindex, texindex;

	std::vector<std::vector<Res::Level::SURFACE>::iterator> transparentlist;
	std::vector<std::vector<Res::Level::SURFACE>::iterator> parallaxskylist;
	std::vector<std::vector<Res::Level::SURFACE>::iterator> ceilingskylist;
	std::vector<std::vector<Res::Level::SURFACE>::iterator> reflectlist;

	std::vector<int> transparentthings;
	std::vector<std::vector<int> > sectorthings;
	sectorthings.resize(level->SectorList.size());
	for (n = 0; n < Gameplay::WorldState::Get().ThingList.size(); n++)
	{
		if (Gameplay::WorldState::Get().ThingList[n].Model3D.isValid() && Gameplay::WorldState::Get().ThingList[n].Instanced && Gameplay::WorldState::Get().ThingList[n].Sector != -1)// && !(WorldState->ThingList[n].ThingFlags & 0x10))
			sectorthings[Gameplay::WorldState::Get().ThingList[n].Sector].push_back(n);
	}

	std::vector<std::vector<Res::Level::SECTOR>::iterator> sectorlist;
	std::vector<unsigned char> sectorvisited;
	if (startsector != -1) // if we have a starting sector
	{
		sectorlist.push_back(&level->SectorList[startsector]);
		// need to keep track of already visited sectors if we are walking through
		sectorvisited.resize(level->SectorList.size(), 0);
		sectorvisited[startsector] = 1;
	}
	else // otherwise put in every visible sector
	{
		for (int n = 0; n < sectorvis.size(); n++)
		{
			if (sectorvis[n]) // visible
			{
				sectorlist.push_back(&level->SectorList[n]);
			}
		}
	}


	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	for (int sn = 0; sn < sectorlist.size(); sn++)
	{
		std::vector<Res::Level::SECTOR>::iterator s = sectorlist[sn];
		sectors_drawn++;

		// draw things
		for (std::vector<int>::iterator t = sectorthings[s->SectorNumber].begin(); t != sectorthings[s->SectorNumber].end(); ++t)
		{
			Gameplay::Thing *thing = &Gameplay::WorldState::Get().ThingList[*t];

			if (thing->Model3D->Transparency)
			{
				transparentthings.push_back(*t);
				continue;
			}

			glPushMatrix();
			glTranslatef(thing->Pos.x, thing->Pos.y, thing->Pos.z);
			glRotatef(thing->Orient.x, 1, 0, 0);
			glRotatef(thing->Orient.y, 0, 0, 1);
			glRotatef(thing->Orient.z, 0, 1, 0);
			Render3do(thing->Model3D, thing->Sector);
			glPopMatrix();
		}

		std::vector<Res::Level::SURFACE>::iterator surface_end = &level->SurfaceList[s->FirstSurfaceIndex + s->SurfaceCount];
		for (std::vector<Res::Level::SURFACE>::iterator f = &level->SurfaceList[s->FirstSurfaceIndex]; f != surface_end; ++f)
		{
			// if adjoined
			if (f->AdjoinedSector)
			{
				// if it is adjoined and the adjoing sector isn't visible, the adjoin can't be
				if (!sectorvis[f->AdjoinedSector->SectorNumber])
					continue;

				// if we have a starting point, add sectors as we find them, if not already visited
				if (startsector != -1)
				{
					if (!sectorvisited[f->AdjoinedSector->SectorNumber])
					{
						sectorvisited[f->AdjoinedSector->SectorNumber] = 1;
						sectorlist.push_back(&level->SectorList[f->AdjoinedSector->SectorNumber]);
					}
				}
			}


			if (!f->GeoMode)
				continue;
			// water
			if (f->SurfaceFlags & 0x200000)
			//if (f->AdjoinIndex != -1 && (f->ParentSector->SectorFlags & 0x2 || f->AdjoinedSector->SectorFlags & 0x2))
			{
				if (DrawReflect)
				{
					reflectlist.push_back(f);
					continue;
				}
				else if (f->AdjoinIndex != -1) // avoid HOM
					continue;
			}
			// reflect
			if (f->SurfaceFlags & 0x400000)
			{
				if (DrawReflect)
				{
					reflectlist.push_back(f);
					continue;
				}
				else if (f->AdjoinIndex != -1) // avoid HOM
					continue;
			}
			// transparent
			if (f->FaceFlags & 0x2)
			{
				transparentlist.push_back(f);
				continue;
			}
			if (f->SurfaceFlags & 0x400) // ceiling sky
			{
				if (DrawSky)
				{
					ceilingskylist.push_back(f);
					continue;
				}
			}
			if (f->SurfaceFlags & 0x200) // parallax sky
			{
				if (DrawSky)
				{
					parallaxskylist.push_back(f);
					continue;
				}
			}

			texindex = f->TextureIndex;
			if (texindex < 0)
				continue;

			baseindex = clevel.surfaceoffset[f->SurfaceNumber];
			for (int v = 2; v < f->VertexCount; v++)
			{
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex;
				clevel.p[texindex]++;
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex + v - 1;
				clevel.p[texindex]++;
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex + v;
				clevel.p[texindex]++;
			}
		}
	}

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, clevel.drawbuffer);
//	glVertexPointer(3, GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].pos);
//	glNormalPointer(GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].normal);
//	glTexCoordPointer(2, GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].tex);
//	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(CompiledVertex), clevel.vertex[0].color);
	glVertexPointer(3, GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, pos));
	glNormalPointer(GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, normal));
	glTexCoordPointer(2, GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, tex));
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, color));
	for (n = 0; n < clevel.drawbins; n++)
	{
		if (clevel.p[n] > 0)
		{
			glBindTexture(GL_TEXTURE_2D, level->MatList[n]->GetHBM(0).texid[0]);
			glDrawElements(GL_TRIANGLES, clevel.p[n], GL_UNSIGNED_INT, clevel.drawindex[n]);
		}
	}


	// ceiling sky
	if (ceilingskylist.size() > 0)
	{
		glDisable(GL_TEXTURE_2D);
		glColorMask(0, 0, 0, 0);
		glEnable(GL_STENCIL_TEST);
		// only supporting one 0x400 sky texture for now
		glStencilFunc(GL_ALWAYS, 2, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
		for (unsigned int n = 0; n <ceilingskylist.size(); n++)
		{
			std::vector<Res::Level::SURFACE>::iterator f = ceilingskylist[n];
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glColorMask(1, 1, 1, 1);
		glDisable(GL_DEPTH_TEST);
		glStencilFunc(GL_EQUAL, 2, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		VFS::WisePointer<Res::Material> mat;
		mat = level->MatList[ceilingskylist[0]->TextureIndex];
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, mat->GetHBM(ceilingskylist[0]->CurrentCel).texid[0]);

		float umod, vmod;
		umod =  level->Header.CeilingSkyOffset[0] / mat->GetHBM(ceilingskylist[0]->CurrentCel).GetGeometry().xdim;
		vmod =  level->Header.CeilingSkyOffset[1] / mat->GetHBM(ceilingskylist[0]->CurrentCel).GetGeometry().ydim;

		float infiniteedge = 2000;
		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glTexCoord2f(-infiniteedge * 0.0625f + umod, infiniteedge * 0.0625f + vmod);
		glVertex3f(-infiniteedge, infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(infiniteedge * 0.0625f + umod, infiniteedge * 0.0625f + vmod);
		glVertex3f(infiniteedge, infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(infiniteedge * 0.0625f + umod, -infiniteedge * 0.0625f + vmod);
		glVertex3f(infiniteedge, -infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(-infiniteedge * 0.0625f + umod, -infiniteedge * 0.0625f + vmod);
		glVertex3f(-infiniteedge, -infiniteedge, level->Header.CeilingSkyZ); // add camera.z?
		glEnd();

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
	}

	// parallax sky
	if (parallaxskylist.size() > 0)
	{
		glDisable(GL_TEXTURE_2D);
		glColorMask(0, 0, 0, 0);
		glEnable(GL_STENCIL_TEST);
		// only supporting one 0x200 sky texture for now
		glStencilFunc(GL_ALWAYS, 1, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
		for (unsigned int n = 0; n < parallaxskylist.size(); n++)
		{
			std::vector<Res::Level::SURFACE>::iterator f = parallaxskylist[n];
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glColorMask(1, 1, 1, 1);
		glDisable(GL_DEPTH_TEST);
		glStencilFunc(GL_EQUAL, 1, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		VFS::WisePointer<Res::Material> mat;
		mat = level->MatList[parallaxskylist[0]->TextureIndex];
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, mat->GetHBM(parallaxskylist[0]->CurrentCel).texid[0]);

		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		float umod, vmod;
		umod = (-Camera.yaw / 360 * level->Header.HorizonPixelsPerRev + level->Header.HorizonSkyOffset[0]) / mat->GetHBM(parallaxskylist[0]->CurrentCel).GetGeometry().xdim;
		vmod = (Camera.pitch / 180 * level->Header.HorizonPixelsPerRev + level->Header.HorizonSkyOffset[1]) / mat->GetHBM(parallaxskylist[0]->CurrentCel).GetGeometry().ydim;

		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glTexCoord2f(umod, vmod);
		glVertex3f(-1, -1, 0);
		glTexCoord2f(1 + umod, vmod);
		glVertex3f(1, -1, 0);
		glTexCoord2f(1 + umod, 1 + vmod);
		glVertex3f(1, 1, 0);
		glTexCoord2f(umod, 1 + vmod);
		glVertex3f(-1, 1, 0);
		glEnd();

		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
	}

	// draw transparent 3dos
	if (transparentthings.size() > 0)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		for (int n = transparentthings.size() - 1; n >= 0; n--)
		{
			glPushMatrix();
			glTranslatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.x, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.y, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.z);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.x, 1, 0, 0);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.y, 0, 0, 1);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.z, 0, 1, 0);
			Render3do(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Model3D, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Sector);
			glPopMatrix();
		}
		glDisable(GL_BLEND);
	}

	// draw transparent surfaces
	if (transparentlist.size() > 0)
	{
		glAlphaFunc(GL_GREATER, .5f * .352941f);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_ALPHA_TEST);
		glEnable(GL_BLEND);
		for (int n = transparentlist.size() - 1; n >= 0; n--)
		{
			std::vector<Res::Level::SURFACE>::iterator f = transparentlist[n];
			//if (f->GeoMode != 4)
			if (!f->GeoMode)
				continue;
			VFS::WisePointer<Res::Material> mat = level->MatList[f->TextureIndex];
			glBindTexture(GL_TEXTURE_2D, mat->GetHBM(f->CurrentCel).texid[0]);
			float uscale, vscale;
			uscale = ((float)1 / mat->GetHBM(f->CurrentCel).GetGeometry().xdim) * mat->xscale;
			vscale = ((float)1 / mat->GetHBM(f->CurrentCel).GetGeometry().ydim) * mat->yscale;
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				float intensity = 1; // using intensity = 1 for now because it looks too dark otherwise in LEC levels
				glColor4f(intensity, intensity, intensity, .352941f); // 90 / 255
				glTexCoord2f((f->TexShiftU + level->TexvertexList[v->TexvertexIndex].U) * uscale, (f->TexShiftV + level->TexvertexList[v->TexvertexIndex].V) * vscale);
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
	}

	if (reflectlist.size() > 0)
	{
		reflectionplanes = 0;
		float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();

		UpdateReflections(reflectlist);		


		glEnable(GL_VERTEX_PROGRAM_ARB);
		glEnable(GL_FRAGMENT_PROGRAM_ARB);
		for (int r = 0; r < reflect_tex.size(); r++)
		{
			for (n = 0; n < reflect_tex[r].surface.size(); n++)
			{
				std::vector<Res::Level::SURFACE>::iterator f = reflect_tex[r].surface[n];

				// water
				if (f->SurfaceFlags & 0x200000)
				{
					// draw the actual water surfaces
					glActiveTextureARB(GL_TEXTURE2_ARB);
					glBindTexture(GL_TEXTURE_2D, refractiontex);
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glBindTexture(GL_TEXTURE_2D, reflect_tex[r].texid);
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glBindTexture(GL_TEXTURE_2D, water.normaltex);

					glBindProgramARB(GL_VERTEX_PROGRAM_ARB, watervp);
					glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, waterfp);

					glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, Camera.x, Camera.y, Camera.z, 1);
					// direction to light (opengl coords)
					glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, 0, sin(90 * 0.0174532925f), cos(45 * 0.0174532925f), 1);

			//				float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();
					// texscale because reflection texture has a wider fov
					float texscale = tan(fov * .5f * 0.0174532925f) / tan((fov + waterfovmod) * .5f * 0.0174532925f);
					glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, (float)1 / wsizex * texscale, (float)1 / wsizey * texscale, .01, 1); // z is inverse scale
					glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, (1 - texscale) * .5f, (1 - texscale) * .5f, 0, 0);
					// specular color
					glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 4, .2f, .15f, .1f, 1);

					float tangent[3], binormal[3];
					float up[3] = {0, 0, 1};
					if (f->Normal.Z > .99f) // pointing up
					{
						tangent[0] = 1;
						tangent[1] = 0;
						tangent[2] = 0;
						binormal[0] = 0;
						binormal[1] = 1;
						binormal[2] = 0;
					}
					else if (f->Normal.Z < -.99f) // pointing down
					{
						tangent[0] = -1;
						tangent[1] = 0;
						tangent[2] = 0;
						binormal[0] = 0;
						binormal[1] = -1;
						binormal[2] = 0;
					}
					else // other
					{
						// order on these may be wrong
						// tangent = Cross(up, normal)
						// binormal = Cross(normal, tangent)
						float invlen;
						tangent[0] = up[1] * f->Normal.Z - up[2] * f->Normal.Y;
						tangent[1] = up[2] * f->Normal.X - up[0] * f->Normal.Z;
						tangent[2] = up[0] * f->Normal.Y - up[1] * f->Normal.X;
						invlen = 1 / sqrt(tangent[0] * tangent[0] + tangent[1] * tangent[1] + tangent[2] * tangent[2]);
						tangent[0] *= invlen;
						tangent[1] *= invlen;
						tangent[2] *= invlen;
						binormal[0] = f->Normal.Y * tangent[2] - f->Normal.Z * tangent[1];
						binormal[1] = f->Normal.Z * tangent[0] - f->Normal.X * tangent[2];
						binormal[2] = f->Normal.X * tangent[1] - f->Normal.Y * tangent[0];
						invlen = 1 / sqrt(binormal[0] * binormal[0] + binormal[1] * binormal[1] + binormal[2] * binormal[2]);
						binormal[0] *= invlen;
						binormal[1] *= invlen;
						binormal[2] *= invlen;
					}
					glBegin(GL_POLYGON);
					for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
					{
						glNormal3f(f->Normal.X, f->Normal.Y, f->Normal.Z);
						glMultiTexCoord3fvARB(GL_TEXTURE1_ARB, tangent); // u
						glMultiTexCoord3fvARB(GL_TEXTURE2_ARB, binormal); // v
						glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
					}
					glEnd();
				}
				// reflection
				// add bumpy reflections
				else if (f->SurfaceFlags & 0x400000)
				{
					VFS::WisePointer<Res::Material> mat = level->MatList[f->TextureIndex];

					glActiveTextureARB(GL_TEXTURE3_ARB);
					glBindTexture(GL_TEXTURE_2D, reflect_tex[r].texid);
					glActiveTextureARB(GL_TEXTURE2_ARB);
					glBindTexture(GL_TEXTURE_2D, mat->GetHBM(f->CurrentCel).gloss);
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glBindTexture(GL_TEXTURE_2D, mat->GetHBM(f->CurrentCel).bump);
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glBindTexture(GL_TEXTURE_2D, mat->GetHBM(f->CurrentCel).diffuse);

					float uscale, vscale;
					uscale = mat->xscale;
					vscale = mat->yscale;

					glBindProgramARB(GL_VERTEX_PROGRAM_ARB, reflectvp);
					glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, reflectfp);

					glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, Camera.x, Camera.y, Camera.z, 1);

					// texscale because reflection texture has a wider fov
					float texscale = tan(fov * .5f * 0.0174532925f) / tan((fov + waterfovmod) * .5f * 0.0174532925f);
					glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, (float)1 / wsizex * texscale, (float)1 / wsizey * texscale, .0025, 1); // z is inverse scale
					glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, (1 - texscale) * .5f, (1 - texscale) * .5f, 0, 0);

					glBegin(GL_POLYGON);
					for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
					{
						glNormal3fv(&f->Normal.X);
						glMultiTexCoord3fvARB(GL_TEXTURE1_ARB, &f->Tangent.X); // u
						glMultiTexCoord3fvARB(GL_TEXTURE2_ARB, &f->Binormal.X); // v
						glTexCoord2f((f->TexShiftU + level->TexvertexList[v->TexvertexIndex].U) * uscale, (f->TexShiftV + level->TexvertexList[v->TexvertexIndex].V) * vscale);
						glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
					}
					glEnd();
				}
			}
		}
		glDisable(GL_VERTEX_PROGRAM_ARB);
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
	}


	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
}

void Sith::Win32::Renderer::UpdateReflections(std::vector<std::vector<Res::Level::SURFACE>::iterator> &reflectionlist)
{
	float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();

	// clear out the reflection surface lists
	for (int r = 0; r < reflect_tex.size(); r++)
	{
		reflect_tex[r].surface.clear();
	}

	// assign surfaces to textures
	for (int n = 0; n < reflectionlist.size(); n++)
	{
		std::vector<Res::Level::SURFACE>::iterator f = reflectionlist[n];

		// these are expensive to update and draw, so if it is backfacing, skip
		if (Camera.x * f->Normal.X + Camera.y * f->Normal.Y + Camera.z * f->Normal.Z
			< f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
			+ f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
			+ f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z)
			continue;

		for (int r = 0; r < reflect_tex.size(); r++)
		{
			if (reflect_tex[r].surface.size() == 0) // no surfaces yet
			{
				reflect_tex[r].normal[0] = f->Normal.X;
				reflect_tex[r].normal[1] = f->Normal.Y;
				reflect_tex[r].normal[2] = f->Normal.Z;
				reflect_tex[r].normal[3] = f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
										+ f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
										+ f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z;
				reflect_tex[r].surface.push_back(f);
				break;
			}
			// check against existing surface
			float d = reflect_tex[r].normal[0] * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
					+ reflect_tex[r].normal[1] * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
					+ reflect_tex[r].normal[2] * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z;
			if (fabs(d - reflect_tex[r].normal[3]) < reflection_tolerance) // same plane distance
			{
				float dot = reflect_tex[r].normal[0] * f->Normal.X + reflect_tex[r].normal[1] * f->Normal.Y + reflect_tex[r].normal[2] * f->Normal.Z;
				if (dot > 1 - reflection_tolerance) // same orientation
				{
					reflect_tex[r].surface.push_back(f);
					break;
				}
			}
		}
		if (r == reflect_tex.size()) // not enough allocated textures
		{
			// so add a new one
			reflect_tex.resize(reflect_tex.size() + 1);
			reflect_tex[r].GenTex(framebuffersize, framebuffersize);

			reflect_tex[r].normal[0] = f->Normal.X;
			reflect_tex[r].normal[1] = f->Normal.Y;
			reflect_tex[r].normal[2] = f->Normal.Z;
			reflect_tex[r].normal[3] = f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
									+ f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
									+ f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z;
			reflect_tex[r].surface.push_back(f);
		}
	}

	// update reflection textures
	for (int r = 0; r < reflect_tex.size(); r++)
	{
		if (reflect_tex[r].surface.size() == 0) // skip empty ones
			continue;

		reflectionplanes++;

		std::vector<Res::Level::SURFACE>::iterator f = reflect_tex[r].surface[0];

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, reflect_tex[r].texid, 0);

		glViewport(0, 0, framebuffersize, framebuffersize);
		//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glClear(GL_DEPTH_BUFFER_BIT);

		projection[4] = projection[8] = projection[12] = projection[1] = projection[9] = projection[13] = projection[2] = projection[6] = projection[3] = projection[7] = projection[15] = 0;
		projection[0] = (1 / tan((float)(fov + waterfovmod) * .5f * 0.0174532925f)) / ((float)wsizex / (float)wsizey);
		projection[5] = (1 / tan((float)(fov + waterfovmod) * .5f * 0.0174532925f));
		projection[14] = (float)-2 * nearclip; // near clip
		projection[10] = projection[11] = -1;
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadMatrixf(projection);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		float jkm[16] = {
			1, 0, 0, 0,
			0, 0, -1, 0,
			0, 1, 0, 0,
			0, 0, 0, 1
		};

		glRotatef(-Camera.pitch, 1, 0, 0);
		glRotatef(-Camera.yaw, 0, 1, 0);

		// everything after this point is in jk's coordinate system
		glMultMatrixf(jkm);


		// find the reflection matrix
		float rdot = f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
					+ f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
					+ f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z;
		float reflect[16];
		reflect[0] = 1 - 2 * f->Normal.X * f->Normal.X;
		reflect[4] = -2 * f->Normal.X * f->Normal.Y;
		reflect[8] = -2 * f->Normal.X * f->Normal.Z;
		reflect[12] = 2 * rdot * f->Normal.X;
		reflect[1] = -2 * f->Normal.X * f->Normal.Y;
		reflect[5] = 1 - 2 * f->Normal.Y * f->Normal.Y;
		reflect[9] = -2 * f->Normal.Y * f->Normal.Z;
		reflect[13] = 2 * rdot * f->Normal.Y;
		reflect[2] = -2 * f->Normal.X * f->Normal.Z;
		reflect[6] = -2 * f->Normal.Y * f->Normal.Z;
		reflect[10] = 1 - 2 * f->Normal.Z * f->Normal.Z;
		reflect[14] = 2 * rdot * f->Normal.Z;
		reflect[3] = 0;
		reflect[7] = 0;
		reflect[11] = 0;
		reflect[15] = 1;


		glTranslatef(-Camera.x, -Camera.y, -Camera.z);

		glMultMatrixf(reflect);

		// set up a clip plane to only get stuff above the reflection plane
		double cp[4];
		cp[0] = f->Normal.X;
		cp[1] = f->Normal.Y;
		cp[2] = f->Normal.Z;
		cp[3] = -(f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X + f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y + f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z); // since the surface is raised
		glClipPlane(GL_CLIP_PLANE0, cp);
		glEnable(GL_CLIP_PLANE0);


		// get reflected camera position
		float reflectcampos[3];
		float invw;
		invw = 1 / (Camera.x * reflect[3] + Camera.y * reflect[7] + Camera.z * reflect[11] + reflect[15]);
		reflectcampos[0] = (Camera.x * reflect[0] + Camera.y * reflect[4] + Camera.z * reflect[8] + reflect[12]) * invw;
		reflectcampos[1] = (Camera.x * reflect[1] + Camera.y * reflect[5] + Camera.z * reflect[9] + reflect[13]) * invw;
		reflectcampos[2] = (Camera.x * reflect[2] + Camera.y * reflect[6] + Camera.z * reflect[10] + reflect[14]) * invw;
		float reflectcamforward[3];
		reflectcamforward[0] = Camera.forward[0] * reflect[0] + Camera.forward[1] * reflect[4] + Camera.forward[2] * reflect[8];
		reflectcamforward[1] = Camera.forward[0] * reflect[1] + Camera.forward[1] * reflect[5] + Camera.forward[2] * reflect[9];
		reflectcamforward[2] = Camera.forward[0] * reflect[2] + Camera.forward[1] * reflect[6] + Camera.forward[2] * reflect[10];
		float reflectcamright[3];
		reflectcamright[0] = Camera.right[0] * reflect[0] + Camera.right[1] * reflect[4] + Camera.right[2] * reflect[8];
		reflectcamright[1] = Camera.right[0] * reflect[1] + Camera.right[1] * reflect[5] + Camera.right[2] * reflect[9];
		reflectcamright[2] = Camera.right[0] * reflect[2] + Camera.right[1] * reflect[6] + Camera.right[2] * reflect[10];
		float reflectcamup[3];
		reflectcamup[0] = Camera.up[0] * reflect[0] + Camera.up[1] * reflect[4] + Camera.up[2] * reflect[8];
		reflectcamup[1] = Camera.up[0] * reflect[1] + Camera.up[1] * reflect[5] + Camera.up[2] * reflect[9];
		reflectcamup[2] = Camera.up[0] * reflect[2] + Camera.up[1] * reflect[6] + Camera.up[2] * reflect[10];
		// get visibility
		std::vector<unsigned char> reflectvis;
		// need to get a list of appropriate sectors
		std::vector<int> reflectvis_sectors;
		// visibility is calculated from the reflected camera position, through each sector touched by a reflecting surface on the same plane
		for (int n = 0; n < reflect_tex[r].surface.size(); n++)
			reflectvis_sectors.push_back(reflect_tex[r].surface[n]->ParentSector->SectorNumber);
		Sith::JKL::FindVis(reflectvis_sectors, level, reflectcampos, reflectcamforward, reflectcamright, reflectcamup, fov, (float)wsizex / wsizey, reflectvis);

		// reflection reverses vertex order
		glCullFace(GL_FRONT);

		DrawJKLAdvanced(false, false, -1, reflectvis);

		glCullFace(GL_BACK);

		glDisable(GL_CLIP_PLANE0);

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		glViewport(0, 0, wsizex, wsizey);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
}