/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "ogl_render_compiled.h"

void Sith::Win32::Renderer::CompileLevel()
{
	int n;

	clevel.surfaceoffset = new int [level->SurfaceList.size()];

	int size = 0;
	for (n = 0; n < level->SurfaceList.size(); n++)
	{
		clevel.surfaceoffset[n] = size;
		size += level->SurfaceList[n].VertexCount;
	}
	clevel.vertex = new CompiledVertex [size];
	clevel.vertices = size;

	int p = 0;
	Res::Level::SURFACE *f;
	float uscale, vscale, extralight, c;
	for (n = 0; n < level->SurfaceList.size(); n++)
	{
		f = &level->SurfaceList[n];
		if (f->TextureIndex >= 0)
		{
			uscale = (1.0f / level->MatList[f->TextureIndex]->GetHBM(0).GetGeometry().xdim) / level->MatList[f->TextureIndex]->xscale;
			vscale = (1.0f / level->MatList[f->TextureIndex]->GetHBM(0).GetGeometry().ydim) / level->MatList[f->TextureIndex]->yscale;
		}
		extralight = f->ExtraLight + f->ParentSector->ExtraLight;
		for (int v = 0; v < level->SurfaceList[n].VertexCount; v++)
		{
			clevel.vertex[p].pos[0] = level->GeovertexList[f->VertexList[v].GeovertexIndex].X;
			clevel.vertex[p].pos[1] = level->GeovertexList[f->VertexList[v].GeovertexIndex].Y;
			clevel.vertex[p].pos[2] = level->GeovertexList[f->VertexList[v].GeovertexIndex].Z;
			clevel.vertex[p].tangent[0] = f->Tangent.X;
			clevel.vertex[p].tangent[1] = f->Tangent.Y;
			clevel.vertex[p].tangent[2] = f->Tangent.Z;
			clevel.vertex[p].binormal[0] = f->Binormal.X;
			clevel.vertex[p].binormal[1] = f->Binormal.Y;
			clevel.vertex[p].binormal[2] = f->Binormal.Z;
			clevel.vertex[p].normal[0] = f->Normal.X;
			clevel.vertex[p].normal[1] = f->Normal.Y;
			clevel.vertex[p].normal[2] = f->Normal.Z;
			if (f->TextureIndex >= 0)
			{
				clevel.vertex[p].tex[0] = level->TexvertexList[f->VertexList[v].TexvertexIndex].U * uscale;
				clevel.vertex[p].tex[1] = level->TexvertexList[f->VertexList[v].TexvertexIndex].V * vscale;
			}
			else
			{
				clevel.vertex[p].tex[0] = 0;
				clevel.vertex[p].tex[1] = 0;
			}
			/*if (f->FaceFlags & 0x2) // transparent
			{
				clevel.vertex[p].color[0] = 255;
				clevel.vertex[p].color[1] = 255;
				clevel.vertex[p].color[2] = 255;
				clevel.vertex[p].color[3] = 90;
			}
			else*/
			{
				c = (extralight + f->VertexList[v].LightR) * 255.0f;
				if (c > 255.0f)
					c = 255.0f;
				clevel.vertex[p].color[0] = c;
				c = (extralight + f->VertexList[v].LightG) * 255.0f;
				if (c > 255.0f)
					c = 255.0f;
				clevel.vertex[p].color[1] = c;
				c = (extralight + f->VertexList[v].LightB) * 255.0f;
				if (c > 255.0f)
					c = 255.0f;
				clevel.vertex[p].color[2] = c;
				clevel.vertex[p].color[3] = 255;
			}
			p++;
		}
	}


	glGenBuffersARB(1, &clevel.drawbuffer);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, clevel.drawbuffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(CompiledVertex) * clevel.vertices, clevel.vertex, GL_STATIC_DRAW_ARB);




	// can make some (or all, and waste a little) of these memory allocations once per level
//	int *drawsize = 0;
//	int drawbins = level->MatList.size();
	clevel.drawbins = level->MatList.size();
	clevel.drawsize = new int [clevel.drawbins];
	memset(clevel.drawsize, 0, sizeof(int) * clevel.drawbins);
//// this can be done once per level, and easily updated if a surface has its texture changed
	for (n = 0; n < level->SurfaceList.size(); n++)
	{
		if (level->SurfaceList[n].TextureIndex < 0)
			continue;
		clevel.drawsize[level->SurfaceList[n].TextureIndex] += (level->SurfaceList[n].VertexCount - 2) * 3;
	}

//	unsigned int **drawindex;
	clevel.drawindex = new unsigned int* [clevel.drawbins];
	for (n = 0; n < clevel.drawbins; n++)
		clevel.drawindex[n] = new unsigned int [clevel.drawsize[n]];

//	int *p;
	clevel.p = new int [clevel.drawbins];
}

void Sith::Win32::CompiledLevel::Clear()
{
	int n;

	if (vertex)
		delete [] vertex;
	vertex = 0;
	vertices = 0;
	if (surfaceoffset)
		delete [] surfaceoffset;

	if (p)
		delete [] p;
	p = 0;
	if (drawsize)
		delete [] drawsize;
	drawsize = 0;
	if (drawindex)
	{
		for (n = 0; n < drawbins; n++)
			delete [] drawindex[n];
		delete [] drawindex;
	}
	drawindex = 0;

	if (drawbuffer)
		glDeleteBuffersARB(1, &drawbuffer);
	drawbuffer = 0;
}