/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"

LRESULT CALLBACK Sith::Win32::Renderer::RenderWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	Sith::Win32::Renderer* renderer = (Sith::Win32::Renderer*)GetWindowLong(hWnd, GWL_USERDATA);

	switch(uMsg)
	{
	case WM_NCCREATE:
		// Get the initial creation pointer to the window object
		renderer = (Sith::Win32::Renderer*)((CREATESTRUCT *)lParam)->lpCreateParams;

		// Set its USERDATA DWORD to point to the window object
		SetWindowLong(hWnd, GWL_USERDATA, (long)renderer);
		SetWindowText(hWnd, ((CREATESTRUCT *)lParam)->lpszName);

		return 1;
	case WM_CLOSE:
		renderer->Shutdown();
		break;

	case WM_DESTROY:
		if (!renderer->resetting)
			Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);
		break;

	case WM_SIZE:
		renderer->OnResize(LOWORD(lParam), HIWORD(lParam), 0, 0);
		break;

	case WM_KEYDOWN:
		break;

	case WM_SETFOCUS:
		ShowCursor(false);
		break;

	case WM_KILLFOCUS:
		ShowCursor(true);
		break;

	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	};

	return 0;
};

Sith::SITHERROR Sith::Win32::Renderer::CreateOGLWindow()
{
	memset(&wc, 0, sizeof(WNDCLASSEX));
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = RenderWndProc;
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszClassName = "SITH 2 OPENGL";
	wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

	if(!RegisterClassEx(&wc))
	{
		//MessageBox(NULL, "Failed to register renderer window class", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to register renderer window class.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};


	wsizex = Sith::COG::CogLib::GetSetting("video::x_res").GetInt();
	wsizey = Sith::COG::CogLib::GetSetting("video::y_res").GetInt();
	wbitdepth = Sith::COG::CogLib::GetSetting("video::bitdepth").GetInt();
	wdepthbits = Sith::COG::CogLib::GetSetting("video::depthbits").GetInt();
	wstencilbits = Sith::COG::CogLib::GetSetting("video::stencilbits").GetInt();
	fullscreen = Sith::COG::CogLib::GetSetting("video::fullscreen").GetBool();

	DWORD style = WS_OVERLAPPEDWINDOW, exstyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;

	if (fullscreen)
	{
		ConOut("Starting up video in full screen mode...\n", ConOut::Pedantic);

		memset(&dmOriginalScreenSettings, 0, sizeof(DEVMODE));
		dmOriginalScreenSettings.dmSize = sizeof(DEVMODE);
		EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dmOriginalScreenSettings);

		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings, 0, sizeof(DEVMODE));
		dmScreenSettings.dmSize = sizeof(DEVMODE);
		dmScreenSettings.dmPelsWidth = wsizex;
		dmScreenSettings.dmPelsHeight = wsizey;
		dmScreenSettings.dmBitsPerPel = wbitdepth;
		dmScreenSettings.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN);

		exstyle = WS_EX_APPWINDOW;
		style = WS_POPUP;
	}
	else
	{
		ConOut("Starting up video in windowed mode...\n", ConOut::Pedantic);
	};

	RECT window;
	window.bottom = wsizey;
	window.top = 0;
	window.left = 0;
	window.right = wsizex;

	AdjustWindowRectEx(&window, style, false, exstyle);
	hMainWnd = CreateWindowEx(exstyle, "SITH 2 OPENGL", "Sith 2", style | WS_CLIPSIBLINGS | WS_CLIPCHILDREN, 0, 0, window.right - window.left, window.bottom - window.top, NULL, NULL, GetModuleHandle(NULL), this);

	if(hMainWnd == NULL)
	{
		//MessageBox(NULL, "Failed to create renderer window\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to create renderer window.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};

	hDC = GetDC(hMainWnd);
	if (hDC == NULL)
	{
		//MessageBox(NULL, "Failed to get device context\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to get device context.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}


	unsigned int pixelformat;
	PIXELFORMATDESCRIPTOR pfd;

	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = wbitdepth;
	pfd.cDepthBits = wdepthbits;
	pfd.cStencilBits = wstencilbits;
	pfd.iLayerType = PFD_MAIN_PLANE;

	pixelformat = ChoosePixelFormat(hDC, &pfd);
	if (pixelformat == 0)
	{
		//MessageBox(NULL, "Failed to obtain pixel format\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to obtain pixel format.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}

	if (!SetPixelFormat(hDC, pixelformat, &pfd))
	{
		//MessageBox(NULL, "Failed to set pixel format\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to set pixel format.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}
	hRC = wglCreateContext(hDC);
	if (hRC == 0)
	{
		//MessageBox(NULL, "Failed to obtain render context\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to obtain render context.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}
	if (!wglMakeCurrent(hDC, hRC))
	{
		//MessageBox(NULL, "Failed to set current context\n", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to set current context.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}
	ShowWindow(hMainWnd, SW_SHOW);

	return SE_OK;
}