/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "ogl_render.h"

int Sith::Win32::Renderer::LoadVertexProgram(const char *filename)
{
	char program[10000];
	char fullname[512];
	unsigned int id;
	unsigned char *error;
	int instructions, temps, params, attribs, alu, tex, indirect;
	int errorposition;
	FILE *file;
	FILE *shaderlog = fopen("shaderlog.txt", "a+");

	memset(program, 0, 10000);
	sprintf(fullname, "shaders\\%s.txt", filename);
	file = fopen(fullname, "rt");
	if (!file)
		return -1;
	fread(program, filelength(file->_file), 1, file);
	fclose(file);
	glGenProgramsARB(1, &id);
	glBindProgramARB(GL_VERTEX_PROGRAM_ARB, id);
	glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(program), program);

	error = (unsigned char *)glGetString(GL_PROGRAM_ERROR_STRING_ARB);

	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB, &instructions);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEMPORARIES_ARB, &temps);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_PARAMETERS_ARB, &params);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_ATTRIBS_ARB, &attribs);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB, &alu);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB, &tex);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB, &indirect);

	errorposition = 0;
	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errorposition);

	fprintf(shaderlog, "Shader %s\n\n", filename);
	fprintf(shaderlog, "Errors:\n%s\nLocation: %d\n\n", error, errorposition);
	fprintf(shaderlog, "Native Resources:\nInstructions: %d\nTemps: %d\nParams: %d\nAttribs: %d\nALU Instructions: %d\nTEX Instructions: %d\nTex Indirections: %d\n\n",
		instructions, temps, params, attribs, alu, tex, indirect);

	fclose(shaderlog);

	return id;
}

int Sith::Win32::Renderer::LoadFragmentProgram(const char *filename)
{
	char program[10000];
	char fullname[512];
	unsigned int id;
	unsigned char *error;
	int instructions, temps, params, attribs, alu, tex, indirect;
	int errorposition;
	FILE *file;
	FILE *shaderlog = fopen("shaderlog.txt", "a+");

	memset(program, 0, 10000);
	sprintf(fullname, "shaders\\%s.txt", filename);
	file = fopen(fullname, "rt");
	if (!file)
		return -1;
	fread(program, filelength(file->_file), 1, file);
	fclose(file);
	glGenProgramsARB(1, &id);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, id);
	glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(program), program);

	error = (unsigned char *)glGetString(GL_PROGRAM_ERROR_STRING_ARB);

	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB, &instructions);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEMPORARIES_ARB, &temps);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_PARAMETERS_ARB, &params);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_ATTRIBS_ARB, &attribs);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB, &alu);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB, &tex);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB, &indirect);

	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errorposition);

	fprintf(shaderlog, "Shader %s\n\n", filename);
	fprintf(shaderlog, "Errors:\n%s\nLocation: %d\n\n", error, errorposition);
	fprintf(shaderlog, "Native Resources:\nInstructions: %d\nTemps: %d\nParams: %d\nAttribs: %d\nALU Instructions: %d\nTEX Instructions: %d\nTex Indirections: %d\n\n",
		instructions, temps, params, attribs, alu, tex, indirect);

	fclose(shaderlog);

	return id;
}

void Sith::Win32::Renderer::ShaderStartup()
{
	FILE *shaderlog = fopen("shaderlog.txt", "wt");

	int maxinstructions, maxtemps, maxparams, maxattribs, maxalu, maxtex, maxindirect;

	if (!glGetProgramivARB)
		return;

	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB, &maxinstructions);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB, &maxtemps);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB, &maxparams);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB, &maxattribs);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB, &maxalu);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB, &maxtex);
	glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB, &maxindirect);

	fprintf(shaderlog, "Fragment Program Native Maximums:\nInstructions: %d\nTemps: %d\nParams: %d\nAttribs: %d\nALU Instructions: %d\nTEX Instructions: %d\nTex Indirections: %d\n\n",
		maxinstructions, maxtemps, maxparams, maxattribs, maxalu, maxtex, maxindirect);

	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB, &maxinstructions);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB, &maxtemps);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB, &maxparams);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB, &maxattribs);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB, &maxalu);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB, &maxtex);
	glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB, &maxindirect);

	fprintf(shaderlog, "Vertex Program Native Maximums:\nInstructions: %d\nTemps: %d\nParams: %d\nAttribs: %d\nALU Instructions: %d\nTEX Instructions: %d\nTex Indirections: %d\n\n",
		maxinstructions, maxtemps, maxparams, maxattribs, maxalu, maxtex, maxindirect);

	fclose(shaderlog);


	watervp = LoadVertexProgram("watervp1");
	waterfp = LoadFragmentProgram("waterfp1");

	reflectvp = LoadVertexProgram("reflectvp1");
	reflectfp = LoadFragmentProgram("reflectfp1");
}