/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"

#include "fft.h"
#include "random.h"

void Sith::Win32::WaterSim::Step(float timestep)
{
	elapsedtime += timestep;
//	float t = fmod(elapsedtime, 6.2831853f);
	float t = elapsedtime;

	float **rh;
	rh = new float * [heightdim];
	for (int n = 0; n < heightdim; n++)
		rh[n] = new float [heightdim * 2];
	for (int x = 0; x < heightdim; x++)
	{
		for (int y = 0; y < heightdim / 2; y++)
		{
			float wkt = skg[x * heightdim + y] * t;
			float e1 = cos(wkt);
			float e2 = sin(wkt);
			// conjugate is (e1, -e2)
			float rr, ri;
			// h~0(k) * e^(-i*wkt) + h~*0(-k) * e^(i*wkt);
			rr = h0r[x * heightdim + y] * e1 + h0r[x * heightdim + (heightdim - y - 1)] * e1;
			ri = h0i[x * heightdim + y] * e2 + -h0i[x * heightdim + (heightdim - y - 1)] * -e2;
			rh[x][y * 2 + 0] = rr;
			rh[x][y * 2 + 1] = ri;
			rh[x][(heightdim - y - 1) * 2 + 0] = rr;
			rh[x][(heightdim - y - 1) * 2 + 1] = -ri;
		}
	}

	Sith::Math::FFT::cdft2d(heightdim, heightdim * 2, -1, rh, workarea1, workarea2, workarea3);

	for (x = 0; x < heightdim; x++)
	{
		for (int y = 0; y < heightdim; y++)
		{
			if ((x + y) % 2 == 0)
				heightmap[x * heightdim + y] = rh[x][y * 2 + 0];
			else
				heightmap[x * heightdim + y] = -rh[x][y * 2 + 0];
		}
	}

	for (n = 0; n < heightdim; n++)
		delete [] rh[n];
	delete [] rh;


	float invlen;
	for (x = 0; x < heightdim; x++)
	{
		for (int y = 0; y < heightdim; y++)
		{
			// swapping x and z for BGR upload
			normalmap[(x * heightdim + y) * 3 + 0] = 
				(heightmap[x * heightdim + y] - heightmap[x * heightdim + (y + 1) % heightdim]) * (heightdim / heightexty) * normalmapscale;
			normalmap[(x * heightdim + y) * 3 + 1] = 1;
			normalmap[(x * heightdim + y) * 3 + 2] = 
				(heightmap[x * heightdim + y] - heightmap[(x + 1) % heightdim * heightdim + y]) * (heightdim / heightextx) * normalmapscale;

///////////////////////////////////////////////////////////////////////////////////////////
			// could skip renormalizing if it's going to happen on gpu anyway
/////////////////////////////////////////////////////////////
			invlen = 1 / sqrtf(normalmap[(x * heightdim + y) * 3 + 0] * normalmap[(x * heightdim + y) * 3 + 0] + 1 + normalmap[(x * heightdim + y) * 3 + 2] * normalmap[(x * heightdim + y) * 3 + 2]);
			normalmap[(x * heightdim + y) * 3 + 0] *= invlen;
			normalmap[(x * heightdim + y) * 3 + 1] *= invlen;
			normalmap[(x * heightdim + y) * 3 + 2] *= invlen;
		}
	}

	for (int n = 0; n < heightdim * heightdim * 3; n++)
	{
		normaltexdata[n] = (normalmap[n] + 1) * .5f * 255.0f;
	}

	glBindTexture(GL_TEXTURE_2D, normaltex);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, heightdim, heightdim, GL_BGR, GL_UNSIGNED_BYTE, normaltexdata);
}

void Sith::Win32::WaterSim::Shutdown()
{
	if (heightmap)
		delete [] heightmap;
	heightmap = 0;
	if (normalmap)
		delete [] normalmap;
	normalmap = 0;
	if (normaltexdata)
		delete [] normaltexdata;
	normaltexdata = 0;
	if (h0r)
		delete [] h0r;
	h0r = 0;
	if (h0i)
		delete [] h0i;
	h0i = 0;
	if (kv)
		delete [] kv;
	kv = 0;
	if (skg)
		delete [] skg;
	skg = 0;
	if (workarea1)
		delete [] workarea1;
	workarea1 = 0;
	if (workarea2)
		delete [] workarea2;
	workarea2 = 0;
	if (workarea3)
		delete [] workarea3;
	workarea3 = 0;

	if (normaltex)
		glDeleteTextures(1, &normaltex);
	normaltex = 0;
}

void Sith::Win32::WaterSim::Start(int seed, int dim, float sizex, float sizey, float grav, float windx, float windy, float wavesize, float normalscale)
{
	elapsedtime = 0;

	heightdim = dim; // 128

	heightextx = sizex; // 100.0f
	heightexty = sizey; // 100.0f

	waveamp = wavesize; // .0008f

	normalmapscale = normalscale; // .25f;

	gravity = grav; // 9.8f
	wind[0] = windx; // 5.0f
	wind[1] = windy; // 10.0f

	Sith::Math::Random::sgenrand(seed);

	heightmap = new float [heightdim * heightdim];
	h0r = new float [heightdim * heightdim];
	h0i = new float [heightdim * heightdim];
	kv = new float [heightdim * heightdim * 2];
	skg = new float [heightdim * heightdim];
	normalmap = new float [heightdim * heightdim * 3];

	normaltexdata = new unsigned char [heightdim * heightdim * 3];

	// to repeat over a period of TWOPI seconds use integer frequencies
	for (int x = 0; x < heightdim; x++)
	{
		for (int y = 0; y < heightdim; y++)
		{
			kv[(x * heightdim + y) * 2 + 0] = ((float)x - heightdim / 2) * 6.2831853f / heightextx;
			kv[(x * heightdim + y) * 2 + 1] = ((float)y - heightdim / 2) * 6.2831853f / heightexty;
			// frequency
			skg[x * heightdim + y] = sqrt(sqrt(kv[(x * heightdim + y) * 2 + 0] * kv[(x * heightdim + y) * 2 + 0] + kv[(x * heightdim + y) * 2 + 1] * kv[(x * heightdim + y) * 2 + 1]) * gravity);
//			skg[x * heightdim + y] = (int)(2 * Sith::Math::Random::genrand() + 1);
			float ph = sqrt(Phillips(waveamp, &kv[(x * heightdim + y) * 2], wind));
			h0r[x * heightdim + y] = 0.70710678f * Sith::Math::Random::gaussian() * ph;
			h0i[x * heightdim + y] = 0.70710678f * Sith::Math::Random::gaussian() * ph;
		}
	}

	workarea1 = new float [heightdim * 8];
	workarea2 = new int [2 + (int)sqrt(heightdim + .5)];
	workarea2[0] = 0;
	workarea3 = new float [heightdim / 2];


	glGenTextures(1, &normaltex);
	glBindTexture(GL_TEXTURE_2D, normaltex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, true);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, heightdim, heightdim, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
}

float Sith::Win32::WaterSim::Phillips(float a, float k[2], float wind[2])
{
	float k2 = k[0] * k[0] + k[1] * k[1];
	if (k2 == 0)
		return 0;
	float v2 = wind[0] * wind[0] + wind[1] * wind[1];
	float l = v2 / gravity;
	return a * exp(-1 / (k2 * l * l)) / (k2 * k2) * 
		(k[0] * wind[0] + k[1] * wind[1]) * (k[0] * wind[0] + k[1] * wind[1]) / (k2 * v2);
}