/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _OPCODES_H_INCLUDED_
#define _OPCODES_H_INCLUDED_

namespace Sith
{
	namespace COG
	{
		typedef unsigned char Opcode;

		namespace Opcodes
		{
			// GENERAL OPCODES
			const Opcode NOP = 0x00; // NOP - No Operation

			// CALL OPCODES
			const Opcode SYSCALL = 0x10;		// SYSCALL [INTERRUPT] - Virtual machine hardware function / interrupt
			const Opcode LIBCALL = 0x11; 		// LIBCALL [LIBRARY_OFFSET] - COGLIB function call
			const Opcode MSGCALL = 0x12;		// MSGCALL [DESTINATION_ADDRESS] - Local message call
			const Opcode FCALL = 0x13;			// FCALL [DESTINATION_ADDRESS] - Local function call
			const Opcode RETURN = 0x14;		// RETURN - Returns to the calling function

			const Opcode LOAD = 0x20;		// LOAD [HEAP_OFFSET] - Copies the heap value into the stack
			const Opcode STOR = 0x21;		// STOR [HEAP_OFFSET] - Copies the top of the stack into the heap address

			const Opcode ILOD = 0x22;		// ILOD [HEAP_OFFSET] - Copies the heap value onto the stack, indexed by the topmost value
			const Opcode ISTR = 0x23;		// ISTR [HEAP_OFFSET] - Copies the topmost value onto the heap, indexed by the second-topmost value

			const Opcode SLOD = 0x24;		// SLOD - Copies a stack value onto the top of the stack, indexed by the topmost stack value
			const Opcode SSTR = 0x25;		// SSTR - Saves the topmost value of the stack onto the stack indexed by the second-topmost stack value

			const Opcode PUSH = 0x26;		// PUSH [MANAGED_VALUE] - Pushes the managed value onto the stack
			const Opcode POP = 0x27;		// POP - Deletes the top of the stack

			const Opcode IPSH = 0x28;		// IPSH [QUANTITY] - Pushes 'quantity' number of copies of the topmost value onto the stack
			const Opcode IPOP = 0x29;		// IPOP [QUANTITY] - Indexed pop, pops off 'quantity'

			const Opcode FLSH = 0x2A;		// FLSH - Flushes the stack

			const Opcode RPSH = 0x2B;		// RPSH - Pushes the value on the return reg onto the stack
			const Opcode RPOP = 0x2C;		// RPOP - Puts the topmost value on the return register

			const Opcode ADD = 0x30;		// ADD - Adds the topmost value to the second-topmost value
			const Opcode SUB = 0x31;		// SUB - Subtracts the topmost value from the second-topmost value
			const Opcode MUL = 0x32;		// MUL - Multiplies the second-topmost value by the topmost value
			const Opcode DIV = 0x33;		// DIV - Divides the second-topmost value by the topmost value
			const Opcode MOD = 0x34;		// MOD - Returns the remainder of dividing the second-topmost value by the topmost value

			const Opcode AND = 0x35;		// AND - Bitwise-ands the topmost values
			const Opcode OR = 0x36;			// OR - Bitwise-ors the topmost values
			const Opcode XOR = 0x37;		// XOR -- Bitwise-xors the topmost values
			const Opcode NOT = 0x38;		// NOT -- Bitwise-nots the topmost value

			const Opcode INC = 0x39;		// INC -- Increments the topmost value
			const Opcode DEC = 0x3A;		// DEC -- Decrements the topmost value
			const Opcode SHL = 0x3B;		// SHL [INTEGER] - Multiplies the topmost value by 2^n
			const Opcode SHR = 0x3C;		// SHR [INTEGER] - Divides the topmost value by 2^n

			const Opcode CZ = 0x40;			// CZ - Compares the topmost value to 0
			const Opcode CNZ = 0x41;		// CNZ - Compares the topmost value to non-zero

			const Opcode CE = 0x42;			// CE - Tests if the topmost values are equal
			const Opcode CNE = 0x43;		// CNE - Tests if the topmost values are non-equal

			const Opcode CGT = 0x44;		// CGT - Tests if the second-topmost value is greater-than the topmost value
			const Opcode CLT = 0x45;		// CLT - Tests if the second-topmost value is less-than the topmost value
			const Opcode CGTE = 0x46;		// CGTE - Tests if the second-topmost value is greater-than-or-equal-to the topmost value
			const Opcode CLTE = 0x47;		// CLTE - Tests if the second-topmost value is less-than-or-equal-to the topmost value

			const Opcode CT = 0x48;			// CT - Tests if the topmost value is true
			const Opcode CNT = 0x49;		// CNT - Tests if the topmost value is not true

			const Opcode CAND = 0x4A;		// CAND - Tests topmost AND second-topmost
			const Opcode COR = 0x4B;		// COR - Tests topmost OR second-topmost

			const Opcode JMP = 0x50;		// JMP [INSTRUCTION_ADDRESS] - Unconditionally jumps to the target instruction

			const Opcode JZ = 0x51;			// JZ [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the topmost value is zero
			const Opcode JNZ = 0x52;		// JNZ [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the topmost value is nonzero

			const Opcode JE = 0x53;			// JE [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the top 2 values of the stack are equal
			const Opcode JNE = 0x54;		// JNE [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the top 2 values of the stack are not equal

			const Opcode JGT = 0x55;		// JGT [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the second-topmost value is greater-than the topmost value
			const Opcode JLT = 0x56;		// JLT [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the second-topmost value is less-than the topmost value
			const Opcode JGTE = 0x57;		// JGTE [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the second-topmost value is greater-than-or-equal-to the topmost value
			const Opcode JLTE = 0x58;		// JLTE [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the second-topmost value is less-than-or-equal-to the topmost value

			const Opcode JT = 0x59;			// JT [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the topmost value is true
			const Opcode JNT = 0x5A;		// JNT [INSTRUCTION_ADDRESS] - Jumps to the target instruction if the topmost value is not true

			const Opcode JAND = 0x5B;		// JAND [INSTRUCTION_ADDRESS] - Jumps if topmost AND second-topmost
			const Opcode JOR = 0x5C;		// JOR [INSTRUCTION_ADDRESS] - Jumps if topmost OR second-topmost

			const Opcode ZEQ = 0xFF;		// ZEQ - Sets the topmost value of the stack to 0
		};
	};
};
#endif // _OPCODES_H_INCLUDED_
