/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _PARFILE_H_INCLUDED_
#define _PARFILE_H_INCLUDED_

namespace Sith {
	namespace Res {
		struct ParticleCoord {
			float x, y, z, i;
		};
		class Particle : public VFS::Resource {
		public:
			Particle (std::string FileName);
			SITHERROR LoadFromFile (const char* FileBuffer);

		private:
			float						Version;
			float						Size;
			std::string					MaterialName;
			VFS::WisePointer<Material>	Material;
			unsigned int				LightingMode;
			float						Radius;
			Sith::Math::Vector	Offset;
			unsigned int				Vertices;

			std::vector<ParticleCoord>	Points;

			// Parsing data
			unsigned int CurrentOffset;
			unsigned int LineNumber;
			size_t BufferLen;
			const char* Buffer;
			bool EndOfStream;

			// Parsing functions Jon`C said I could go nuts with
			__forceinline void SkipComment()
			{
				if(Buffer[CurrentOffset] == '#')
				{
					while(CurrentOffset < BufferLen)
					{
						if(Buffer[CurrentOffset] == '\n')
						{
							++LineNumber;
							++CurrentOffset;
							return;
						}
						else if(Buffer[CurrentOffset] == '\0')
						{
							return;
							EndOfStream = true;
						};
						++CurrentOffset;
					};
				};
				return;
			};

			__forceinline void SkipWhitespace()
			{
				while(CurrentOffset < BufferLen)
				{
					SkipComment();
					if(Buffer[CurrentOffset] == '\n') ++LineNumber;
					if(isspace((int)Buffer[CurrentOffset]))
					{
						++CurrentOffset;
						continue;
					}
					else break;
				};
				return;
			};

			// HAHA i wrote one myself
			__forceinline void SkipToNewLine() {
				while (CurrentOffset < BufferLen) {
					if (Buffer[CurrentOffset] == '\n') {
						++CurrentOffset;
						++LineNumber;
						return;
					}
					++CurrentOffset;
				}
				return;
			}

			inline std::string GetWordToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(isalpha((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetFilenameToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(isalnum((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '_')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '~')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '.')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else if(Buffer[CurrentOffset] == '-')
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				std::transform(RetVal.begin(), RetVal.end(), RetVal.begin(), tolower);
				return RetVal;
			};

			inline std::string GetPuncToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return "";
				std::string RetVal;
				while(CurrentOffset < BufferLen)
				{
					if(ispunct((int)Buffer[CurrentOffset]))
						RetVal.push_back(Buffer[CurrentOffset++]);
					else break;
				};
				return RetVal;
			};

			inline int GetIntegerToken()
			{
				SkipWhitespace();
				if(Buffer[CurrentOffset] == '\0') return 0;
				std::string Retval;
				// Hex integer
				if(((Buffer[CurrentOffset] == '0') && ((Buffer[CurrentOffset + 1] == 'x') || (Buffer[CurrentOffset + 1] == 'X'))) || ((Buffer[CurrentOffset] == 'x') || Buffer[CurrentOffset] == 'X'))
				{
					CurrentOffset += 2;
					while(CurrentOffset < BufferLen)
					{
						if(isdigit((int)Buffer[CurrentOffset]))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else if((Buffer[CurrentOffset] >= 'A') && (Buffer[CurrentOffset] <= 'F'))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else if((Buffer[CurrentOffset] >= 'a') && (Buffer[CurrentOffset] <= 'f'))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else break;
					};
					size_t Len = Retval.length();
					int *digit = new int[Len];
					int RetVal = 0;
					unsigned int Count = 0;
					unsigned int i = 0, j = 0;
					for(i = 0; i < Retval.length(); ++i)
					{
						if(Retval[i] > 0x29 && Retval[i] < 0x40)
							digit[i] = Retval[i] & 0x0F;
						else if(Retval[i] >= 'a' && Retval[i] <= 'f')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else if(Retval[i] >= 'A' && Retval[i] <= 'F')
							digit[i] = (Retval[i] & 0x0F) + 9;
						else break;

						++Count;
					};
					j = Count - 1;
					for(i = 0; i < Count; ++i)
					{
						RetVal = RetVal | (digit[i] << (j << 2));
						--j;
					};
					return RetVal;
				}
				else // Normal integer
				{
					if(Buffer[CurrentOffset] == '-')
					{
						Retval.push_back(Buffer[CurrentOffset++]);
					};

					while(CurrentOffset < BufferLen)
					{
						if(isdigit((int)Buffer[CurrentOffset]))
						{
							Retval.push_back(Buffer[CurrentOffset++]);
							continue;
						}
						else
						{
							break;
						};
					};

					return atoi(Retval.c_str());
				};
			};

			float GetFloatToken()
			{
				if(Buffer[CurrentOffset] == '\0') return 0.0f;

				SkipWhitespace();

				std::string Retval;

				if(Buffer[CurrentOffset] == '-')
				{
					Retval.push_back(Buffer[CurrentOffset++]);
				};

				if(Buffer[CurrentOffset] == '.')
				{
					// inject 0 before decimal place..
					Retval.push_back('0');
				};

				while(CurrentOffset < BufferLen)
				{
					if(isdigit((int)Buffer[CurrentOffset]))
					{
						Retval.push_back(Buffer[CurrentOffset++]);
						continue;
					}
					else if(Buffer[CurrentOffset] == '.')
					{
						Retval.push_back(Buffer[CurrentOffset++]);
						continue;
					}
					else
					{
						break;
					};
				};

				return (float)atof(Retval.c_str());
			};

			inline void SkipToNextSection()
			{
				for( ; CurrentOffset < BufferLen; ++CurrentOffset)
				{
					if((Buffer[CurrentOffset] == 's') || (Buffer[CurrentOffset] == 'S'))
					{
						++CurrentOffset;
						if((Buffer[CurrentOffset] == 'e') || (Buffer[CurrentOffset] == 'E'))
						{
							++CurrentOffset;
							if((Buffer[CurrentOffset] == 'c') || (Buffer[CurrentOffset] == 'C'))
							{
								++CurrentOffset;
								if((Buffer[CurrentOffset] == 't') || (Buffer[CurrentOffset] == 'T'))
								{
									++CurrentOffset;
									if((Buffer[CurrentOffset] == 'i') || (Buffer[CurrentOffset] == 'I'))
									{
										++CurrentOffset;
										if((Buffer[CurrentOffset] == 'o') || (Buffer[CurrentOffset] == 'O'))
										{
											++CurrentOffset;
											if((Buffer[CurrentOffset] == 'n') || (Buffer[CurrentOffset] == 'N'))
											{
												CurrentOffset -= 6;
												return;
											};
										};
									};
								};
							};
						};
					};

					if(Buffer[CurrentOffset] == '\n') ++LineNumber;

					if(Buffer[CurrentOffset] == '#')
					{
						SkipWhitespace();
						// dial it back one, yo. This one is self-driving :(
						--CurrentOffset;
					};
				};
				return;
			};
		};
	};
};


#endif