/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "physics.h"

Sith::Math::Vector Sith::Physics::OldctSphereWorld(int ThingRef, int SectorRef, Math::Vector NewPosition)
{
	Sith::VFS::WisePointer<Res::Level> curLevel = Gameplay::WorldState::Get().CurrentLevel;
	Gameplay::Thing* curThing = &Gameplay::WorldState::Get().ThingList[ThingRef];
	Res::Level::SECTOR* curSec = &curLevel->SectorList[SectorRef];
 
	float StandingHeight = 0.0f;
	float Radius = curThing->Size;
	if(curThing->Model3D.isValid())
	{
		StandingHeight = curThing->Model3D->InsertOffset.z;
//		Radius = curThing->ThingTemplate.Model3D->Radius;
	};
 
	// Step Two:
	// - Test for collision against world. kekekeke.
 
	// POSSIBLE STATES:
	// 1.) Thing has passed through or is touching a surface cleanly
	// 2.) Thing is touching the edge of a surface
	// 3.) Thing has partially passed into an adjoin
	// 4.) Thing has passed completely through adjoin and is now in a new sector

	// Loop one for solid surfaces:
	for(std::vector<Res::Level::SURFACE*>::iterator it = curLevel->SectorList[SectorRef].SurfaceList.begin(); it != curLevel->SectorList[SectorRef].SurfaceList.end(); ++it)
	{
		if((*it)->AdjoinIndex >= 0)
		{
			if(Gameplay::WorldState::Get().CurrentLevel->AdjoinList[(*it)->AdjoinIndex].Flags & Flags::Adjoin::Passable) continue;
		};
 
		Math::Vector Normal = Math::Vector(&(*it)->Normal.X);

		if(Math::VectorDot(Normal, curThing->Velocity) < 0)
		{
			float Dist = Math::VectorDot(Normal, Math::VectorSub(NewPosition, Math::Vector(&curLevel->GeovertexList[(*it)->VertexList.back().GeovertexIndex].X)));

			if((Dist < StandingHeight) && (StandingHeight > Radius) && ((*it)->SurfaceFlags & 0x01) && (curThing->PhysFlags & Flags::Physics::AttachesToFloors))
			{
				if(isPointWithinSurface((*it), Normal, Math::VectorSub(NewPosition, Math::VectorScale(Normal, Dist))))
				{
					// Arrest motion and the velocity in the direction of gravity

					//NewPosition = Math::VectorSub(NewPosition, Math::VectorScale(Normal, (Dist - StandingHeight)));
					curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(Normal, Math::VectorDot(Normal, curThing->Velocity)));

					NewPosition = Math::VectorSub(NewPosition, Math::VectorSet(0.0f, 0.0f, (Dist - StandingHeight)));
					//curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(Math::VectorSet(0.0f, 0.0f, -1.0f), Math::VectorDot(Math::VectorSet(0.0f, 0.0f, -1.0f), curThing->Velocity)));

					curThing->AttachIndex = (*it)->SurfaceNumber;
					curThing->AttachState = Gameplay::Thing::SURFACE;
					COG::VMProcess::Get().SendMessage("touched", COG::VMType(COG::VMTC_SURFACE, (*it)->SurfaceNumber, -1), 0, COG::VMType(COG::VMTC_THING, ThingRef, -1));
				};
			}
			else if(Dist < Radius)
			{
				if(isPointWithinSurface((*it), Normal, Math::VectorSub(NewPosition, Math::VectorScale(Normal, Dist))))
				{
					// Arrest motion and the velocity in the direction of the normal - slidy physics
					NewPosition = Math::VectorSub(NewPosition, Math::VectorScale(Normal, (Dist - Radius)));
					curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(Normal, Math::VectorDot(Normal, curThing->Velocity)));
					COG::VMProcess::Get().SendMessage("touched", COG::VMType(COG::VMTC_SURFACE, (*it)->SurfaceNumber, -1), 0, COG::VMType(COG::VMTC_THING, ThingRef, -1));
				};
			};
		};
	};

	// Loop two for adjoins:
	for(std::vector<Res::Level::SURFACE*>::iterator it = curLevel->SectorList[SectorRef].SurfaceList.begin(); it != curLevel->SectorList[SectorRef].SurfaceList.end(); ++it)
	{
		if((*it)->AdjoinIndex < 0) continue;
		if(!(Gameplay::WorldState::Get().CurrentLevel->AdjoinList[(*it)->AdjoinIndex].Flags & Flags::Adjoin::Passable)) continue;
 
		Math::Vector Normal = Math::Vector(&(*it)->Normal.X);

		if(Math::VectorDot(Normal, curThing->Velocity) < 0)
		{
			float Dist = Math::VectorDot(Normal, Math::VectorSub(NewPosition, Math::Vector(&curLevel->GeovertexList[(*it)->VertexList.back().GeovertexIndex].X)));
			if(Dist < 0)
			{
				// Sometimes this can be invoked when the thing is in a completely different sector. :(
				if(isPointWithinSurface((*it), Normal, Math::VectorSub(NewPosition, Math::VectorScale(Normal, Dist))))
				{
					curThing->Sector = (*it)->AdjoinedSector->SectorNumber;
					// Send crossed and entered messages:
					COG::VMProcess::Get().SendMessage("crossed", COG::VMType(COG::VMTC_SURFACE, (*it)->SurfaceNumber, -1), 0, COG::VMType(COG::VMTC_THING, ThingRef, -1));
					COG::VMProcess::Get().SendMessage("entered", COG::VMType(COG::VMTC_SECTOR, (*it)->AdjoinedSector->SectorNumber, -1), 0, COG::VMType(COG::VMTC_THING, ThingRef, -1));
					return OldctSphereWorld(ThingRef, curThing->Sector, NewPosition);
				};
			}
			else if((Dist < Radius) || (Dist < StandingHeight))
			{
				// Thing is partially within another sector...
				// Do NOT recurse back on itself!!!!
				if(curThing->Sector != (*it)->AdjoinedSector->SectorNumber)
				{
					if(isPointWithinSurface((*it), Normal, Math::VectorSub(NewPosition, Math::VectorScale(Normal, Dist))))
					{
						NewPosition = OldctSphereWorld(ThingRef, (*it)->AdjoinedSector->SectorNumber, NewPosition);
					};
				};
			};
		};
	};
 
	return NewPosition;
};

Sith::SITHERROR Sith::Physics::OldUpdatePhysicsThing(float FrameTime, int ThingRef)
{
	static const float piover180 = 0.0174532925f;
	Gameplay::Thing* curThing = &Gameplay::WorldState::Get().ThingList[ThingRef];

//	curThing->Force = Math::VectorSet(0.0f, 0.0f, 0.0f);

	Math::Vector Force(0.0f, 0.0f, 0.0f);

	// Process flag-based updates:
	if((curThing->PhysFlags & Flags::Physics::AcceptsThrust) && ((curThing->Type == Gameplay::Template::TT_PLAYER) || (curThing->Type == Gameplay::Template::TT_ACTOR)))
	{
		// Counter-force time!
		if(curThing->AttachState != Gameplay::Thing::FREE)
		{
			curThing->PhysFlags |= Flags::Physics::AffectedByForce;

			// Ordinary ballistic physics

			float xzscale = cosf(curThing->HeadPitch * piover180);

			Force.x -= curThing->Thrust.y * sinf(curThing->Orient.y * piover180) * curThing->Mass * curThing->MaxThrust * xzscale;
			Force.y += curThing->Thrust.y * cosf(curThing->Orient.y * piover180) * curThing->Mass * curThing->MaxThrust * xzscale;
			Force.z += curThing->Thrust.y * sin(curThing->HeadPitch * piover180) * curThing->Mass * curThing->MaxThrust;

			Force.x -= curThing->Thrust.x * sinf((curThing->Orient.y + 90) * piover180) * curThing->Mass * curThing->MaxThrust;
			Force.y += curThing->Thrust.x * cosf((curThing->Orient.y + 90) * piover180) * curThing->Mass * curThing->MaxThrust;

			Force.z += curThing->Thrust.z * curThing->Mass;

			if(Math::VectorLenSq(curThing->Thrust) == 0.0f)
			{
				curThing->Velocity = Math::VectorScale(curThing->Velocity, 0.2f);
			}
			else if(Math::VectorLenSq(curThing->Velocity) > 2.0f)
			{
				curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(Math::VectorNorm(curThing->Velocity), 0.6f));
			}
		};
	};

	if(curThing->PhysFlags & Flags::Physics::AffectedByGravity)
	{
		curThing->PhysFlags |= Flags::Physics::AffectedByForce;

		// apply gravitational force
		if(ThingRef != Gameplay::WorldState::Get().LocalPlayerThing)
		{
            Force.z -= curThing->Mass * Gameplay::WorldState::Get().CurrentLevel->Header.Gravity;
		};
	};

	if(curThing->PhysFlags & Flags::Physics::AffectedByForce)
	{
		// F = ma; F / m = a.
		Math::Vector Acceleration = Force / curThing->Mass;

		// Update velocity:
		curThing->Velocity += Math::VectorScale(Acceleration, FrameTime);
	};

	// Clear attach flags; ctSphereWorld will reset it if it's still valid
	curThing->AttachState = Gameplay::Thing::SURFACE;

	Math::Vector NewPosition = Math::VectorAdd(curThing->Pos, Math::VectorScale(curThing->Velocity, FrameTime));

	NewPosition = OldctSphereWorld(ThingRef, curThing->Sector, NewPosition);

	curThing->Pos = NewPosition;

	return SE_OK;
};

Sith::SITHERROR Sith::Physics::BuildSurfacePCS(std::vector<Sith::Res::Level::SURFACE*> &PCS, std::list<int> &VisSec, Sith::Res::Level::SECTOR *curSector, const Sith::Math::Vector &Vel,
											   const Sith::Math::Vector &Origin, float Rad)
{
	for(std::vector<Sith::Res::Level::SURFACE*>::iterator it = curSector->SurfaceList.begin(); it != curSector->SurfaceList.end(); ++it)
	{
		Math::Vector Normal = Math::Vector(&(*it)->Normal.X);

		// Early escape:
		if(Math::VectorDot(Vel, Normal) >= 0.0f) continue;

		// Compute distance from thing's origin to plane.
		float Dist = Math::VectorDot(Normal, Math::VectorSub(Origin, Math::Vector(&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[(*it)->VertexList.back().GeovertexIndex].X)));

		if(Dist > Rad) continue;

		PCS.push_back(*it);

		if((*it)->AdjoinIndex >= 0)
		{
			for(std::list<int>::iterator jt = VisSec.begin(); jt != VisSec.end(); ++jt)
			{
				if((*jt) < (*it)->AdjoinedSector->SectorNumber) continue;
				else if((*jt) == (*it)->AdjoinedSector->SectorNumber) break;
				else if((*jt) > (*it)->AdjoinedSector->SectorNumber)
				{
					VisSec.insert(jt, (*it)->AdjoinedSector->SectorNumber);
					BuildSurfacePCS(PCS, VisSec, (*it)->AdjoinedSector, Vel, Origin, Rad);
					break;
				};
			};
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Physics::FindNearestPoint(Sith::Res::Level::SURFACE &Surface, const Math::Vector &Vel, const Math::Vector &Origin, float Rad, Math::Vector &Point)
{
	// If this is an adjoin the point of intersection cannot be an edge and it must be at the center of the thing.
	Math::Vector PlaneNormal(&Surface.Normal.X);
	Math::Vector PlaneOrigin(&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[Surface.VertexList.begin()->GeovertexIndex].X);

	VFS::WisePointer<Res::Level> curLevel = Gameplay::WorldState::Get().CurrentLevel;

	Math::Vector RayOrigin;

	if((Surface.AdjoinIndex >= 0) && (Gameplay::WorldState::Get().CurrentLevel->AdjoinList[Surface.AdjoinIndex].Flags & Flags::Adjoin::Passable))
	{
		// Origin must be the center of the object.
		RayOrigin = Origin;
	}
	else
	{
		// Origin must be a point on the outer edge of the object.
		RayOrigin = Math::VectorSub(Origin, Math::VectorScale(PlaneNormal, Rad));
	};
	
	// Find point of intersection.
	float Dist = RayPlane(PlaneOrigin, PlaneNormal, RayOrigin, Vel);

	Math::Vector InterPoint = Math::VectorAdd(Origin, Math::VectorScale(Vel, Dist));

	if(isPointWithinSurface(&Surface, PlaneNormal, InterPoint))
	{
		Point = InterPoint;
		return SE_OK;
	};

	Point = ClosestPointOnLine(&curLevel->GeovertexList[Surface.VertexList.begin()->GeovertexIndex].X, &curLevel->GeovertexList[Surface.VertexList.rbegin()->GeovertexIndex].X, InterPoint);
	float Closest2 = Math::VectorDistSq(Origin, Point);

	Math::Vector Temp;
	float cTemp;

	for(int i = 1; i < Surface.VertexCount; ++i)
	{
		Temp = ClosestPointOnLine(&curLevel->GeovertexList[Surface.VertexList[i - 1].GeovertexIndex].X, &curLevel->GeovertexList[Surface.VertexList[i].GeovertexIndex].X, InterPoint);
		cTemp = Math::VectorDistSq(Origin, Temp);
		if(cTemp < Closest2)
		{
			Point = Temp;
			Closest2 = cTemp;
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Physics::ctSphereWorld(Gameplay::Thing *curThing, float Frametime, Sith::Physics::COLLISION &FinalCollision)
{
	FinalCollision.Dist = Math::VectorLen(Math::VectorScale(curThing->Velocity, Frametime));
	FinalCollision.Normal = Math::Vector(0.0f, 0.0f, 0.0f);
	FinalCollision.CType = Physics::CT_SPACE;
	FinalCollision.CData = -1;

	FinalCollision.Time = Frametime;

	Math::Vector NegNrmlVel = Math::VectorSub(Math::Vector(0.0f, 0.0f, 0.0f), Math::VectorNorm(curThing->Velocity));

	std::vector<Res::Level::SURFACE*> PCS;
	std::list<int> VisSec;

	Physics::BuildSurfacePCS(PCS, VisSec, &Gameplay::WorldState::Get().CurrentLevel->SectorList[curThing->Sector], curThing->Velocity, curThing->Pos, FinalCollision.Dist + (curThing->Size * 2));

	if(PCS.size() == 0) return SE_OK;

	Math::Vector TempPt;
	float TempDist;

	std::vector<Res::Level::SURFACE*>::iterator it = PCS.begin();

	do
	{
		Physics::FindNearestPoint(*(*it), curThing->Velocity, curThing->Pos, curThing->Size, TempPt);

		TempDist = Physics::RaySphere(TempPt, NegNrmlVel, curThing->Pos, curThing->Size);

		if((FinalCollision.Dist > TempDist) && (TempDist != -1.0f))
		{
			FinalCollision.Dist = TempDist;
			FinalCollision.Normal = Math::Vector(&(*it)->Normal.X);

			if(((*it)->AdjoinIndex >= 0) && (Gameplay::WorldState::Get().CurrentLevel->AdjoinList[(*it)->AdjoinIndex].Flags & Flags::Adjoin::Passable))
			{
				FinalCollision.CType = Physics::CT_ADJOIN;
			}
			else if((*it)->SurfaceFlags & 0x01)
			{
				FinalCollision.CType = Physics::CT_FLOOR;
			}
			else
			{
				FinalCollision.CType = Physics::CT_WALL;
			};

			FinalCollision.CData = (*it)->SurfaceNumber;
			
			FinalCollision.Time = FinalCollision.Dist / Math::VectorLen(curThing->Velocity);
		};

		++it;
	} while(it != PCS.end());

	return SE_OK;
};

int Sith::Physics::isPointWithinSurface(const Sith::Res::Level::SURFACE *curSurface, const Sith::Math::Vector &Normal, const Sith::Math::Vector &Point)
{
	Math::Vector v1, v2;
	float angle = 0;
	for(int i = 0; i < curSurface->VertexCount; ++i)
	{
		v1 = *(Math::Vector *)&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[curSurface->VertexList[i].GeovertexIndex].X - Point;
		v2 = *(Math::Vector *)&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[curSurface->VertexList[(i + 1) % curSurface->VertexCount].GeovertexIndex].X - Point;
		angle += acos(Math::VectorDot(v1, v2) / (v1.Length() * v2.Length()));
	}
	if (angle >= .95f * 6.2831853f)
		return 1;
	return 0;
};

Sith::SITHERROR Sith::Physics::UpdatePhysicsThing(float FrameTime, int ThingRef)
{
	static const float piover180 = 0.0174532925f;
	Gameplay::Thing* curThing = &Gameplay::WorldState::Get().ThingList[ThingRef];

	Math::Vector Force = Math::VectorSet(0.0f, 0.0f, 0.0f);

	// Process flag-based updates:
	if((curThing->PhysFlags & Flags::Physics::AcceptsThrust) && ((curThing->Type == Gameplay::Template::TT_PLAYER) || (curThing->Type == Gameplay::Template::TT_ACTOR)))
	{
		// Counter-force time!
		if(curThing->AttachState != Gameplay::Thing::FREE)
		{
			curThing->PhysFlags |= Flags::Physics::AffectedByForce;

			// Ordinary ballistic physics
			Force.x -= curThing->Thrust.y * sinf(curThing->Orient.y * piover180) * curThing->Mass * curThing->MaxThrust;
			Force.y += curThing->Thrust.y * cosf(curThing->Orient.y * piover180) * curThing->Mass * curThing->MaxThrust;
			Force.z += curThing->Thrust.y * sin(curThing->Orient.x * piover180) * curThing->Mass * curThing->MaxThrust;

			Force.x -= curThing->Thrust.x * sinf((curThing->Orient.y + 90) * piover180) * curThing->Mass * curThing->MaxThrust;
			Force.y += curThing->Thrust.x * cosf((curThing->Orient.y + 90) * piover180) * curThing->Mass * curThing->MaxThrust;

			Force.z += curThing->Thrust.z * curThing->Mass;

			if(Math::VectorLenSq(curThing->Thrust) == 0.0f)
			{
				curThing->Velocity = Math::VectorScale(curThing->Velocity, 0.2f);
			}
			else if(Math::VectorLenSq(curThing->Velocity) > 2.0f)
			{
				curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(Math::VectorNorm(curThing->Velocity), 0.6f));
			}
		};
	};

	if(curThing->PhysFlags & Flags::Physics::AffectedByGravity)
	{
		curThing->PhysFlags |= Flags::Physics::AffectedByForce;

		// apply gravitational force
		Force.z -= curThing->Mass * Gameplay::WorldState::Get().CurrentLevel->Header.Gravity;
	};

	if(curThing->PhysFlags & Flags::Physics::AffectedByForce)
	{
		// F = ma; F / m = a.
		Math::Vector Acceleration = Force / curThing->Mass;

		// Update velocity:
		curThing->Velocity += Math::VectorScale(Acceleration, FrameTime);
	};

	// Clear attach flags; ctSphereWorld will reset it if it's still valid
	curThing->AttachState = Gameplay::Thing::FREE;

	float TimeRemaining = FrameTime;

	Physics::COLLISION FinalCollision;

	while(TimeRemaining > 0.0f)
	{
		if(Math::VectorLenSq(curThing->Velocity) == 0.0f) break;

		ctSphereWorld(curThing, TimeRemaining, FinalCollision);

		TimeRemaining -= FinalCollision.Time;

		switch(FinalCollision.CType)
		{
		case CT_ADJOIN:
			curThing->Sector = Gameplay::WorldState::Get().CurrentLevel->SurfaceList[FinalCollision.CData].AdjoinedSector->SectorNumber;
			curThing->Pos = Math::VectorAdd(curThing->Pos, Math::VectorScale(curThing->Velocity, FinalCollision.Time));
			break;

		case CT_FLOOR:
			curThing->AttachState = Gameplay::Thing::SURFACE;
			curThing->AttachIndex = FinalCollision.CData;
		case CT_WALL:
			curThing->Velocity = Math::VectorSub(curThing->Velocity, Math::VectorScale(FinalCollision.Normal, Math::VectorDot(FinalCollision.Normal, curThing->Velocity)));
		default:
			curThing->Pos = Math::VectorAdd(curThing->Pos, Math::VectorScale(curThing->Velocity, FinalCollision.Time));
			break;
		};
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Physics::UpdatePathThing(float FrameTime, int ThingRef)
{
	Gameplay::Thing* PathThing = &Gameplay::WorldState::Get().ThingList[ThingRef];

	if (PathThing->CurFrame == PathThing->DestFrame)
	{
		return SE_OK;
	}

	// d = vt
	Math::Vector ThingPos = PathThing->Pos;
	Math::Vector ThingOrient = PathThing->Orient;
	Math::Vector CurFramePos = Math::VectorSet(PathThing->Frames[PathThing->CurFrame].X, PathThing->Frames[PathThing->CurFrame].Y, PathThing->Frames[PathThing->CurFrame].Z);
	Math::Vector CurFrameOrient = Math::VectorSet(PathThing->Frames[PathThing->CurFrame].Pitch, PathThing->Frames[PathThing->CurFrame].Yaw, PathThing->Frames[PathThing->CurFrame].Roll);
	Math::Vector DestFramePos = Math::VectorSet(PathThing->Frames[PathThing->DestFrame].X, PathThing->Frames[PathThing->DestFrame].Y, PathThing->Frames[PathThing->DestFrame].Z);
	Math::Vector DestFrameOrient = Math::VectorSet(PathThing->Frames[PathThing->DestFrame].Pitch, PathThing->Frames[PathThing->DestFrame].Yaw, PathThing->Frames[PathThing->DestFrame].Roll);

	Math::Vector DeltaPos, DeltaOrient;
	float Portion = Math::VectorDist (CurFramePos, ThingPos) / Math::VectorDist(DestFramePos, CurFramePos);	// fraction of the total distance to the next frame completed thus far
	float TotalTime = Math::VectorDist(CurFramePos, DestFramePos) / PathThing->FrameSpeed;		// time it will take to get from curframe to destframe

	if (Math::VectorDist(ThingPos, DestFramePos) > PathThing->FrameSpeed * FrameTime)
	{		// if there's more distance to cover than can be done in this timeslice
		DeltaPos = Math::VectorScale(Math::VectorSub(DestFramePos, CurFramePos), PathThing->FrameSpeed * FrameTime);
		DeltaOrient = Math::VectorScale(Math::VectorSub(DestFrameOrient, CurFrameOrient), FrameTime / TotalTime);

		PathThing->Pos = Math::VectorAdd(PathThing->Pos, DeltaPos);
		PathThing->Orient = Math::VectorAdd(PathThing->Orient, DeltaOrient);
	}
	else
	{
		// just move it to the frame
		PathThing->Pos = DestFramePos;
		PathThing->Orient = DestFrameOrient;

		PathThing->CurFrame = PathThing->DestFrame;

		// send arrived
		COG::VMProcess::Get().SendMessage("arrived", COG::VMType(COG::VMTC_THING, ThingRef, -1), 0, COG::VMType());
	}
	
	return SE_OK;
}


Sith::SITHERROR Sith::Physics::UpdateThings()
{
	static Win32::Timer theTimer;
	float FrameTime = (float)theTimer.Query();
	int i = 0;
	// Loop through things:
	for(std::vector<Gameplay::Thing>::iterator it = Gameplay::WorldState::Get().ThingList.begin(); it != Gameplay::WorldState::Get().ThingList.end(); ++it, ++i)
	{
		if(it->Move == Gameplay::Template::MT_NONE)
		{
			continue;
		}
		else if(it->Move == Gameplay::Template::MT_PATH)
		{
			UpdatePathThing (FrameTime, i);
			continue;
		}
		else if(it->Move == Gameplay::Template::MT_PHYSICS)
		{
			OldUpdatePhysicsThing(FrameTime, i);
			continue;
		};
	};

	return SE_OK;
};