/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _PHYSICS_H_INCLUDED_
#define _PHYSICS_H_INCLUDED_

namespace Sith
{
	namespace Physics
	{
		enum COLLIDETYPE
		{
			CT_WALL,
			CT_FLOOR,
			CT_ADJOIN,
			CT_SPACE
		};

		struct COLLISION
		{
			float Dist;
			Math::Vector Normal;
			float Time;

			COLLIDETYPE CType;
			int CData;
		};

		inline float RayPlane(Math::Vector pOrigin, Math::Vector pNormal, Math::Vector rOrigin, Math::Vector rVector)
		{
			float d = -(Math::VectorDot(pNormal, pOrigin));
			float numer = Math::VectorDot(pNormal, rOrigin) + d;
			float denom = Math::VectorDot(pNormal, rVector);
			return -(numer / denom);
		};

		inline float RaySphere(Math::Vector rOrigin, Math::Vector rVector, Math::Vector sOrigin, float sRadius)
		{
			Math::Vector Q = sOrigin - rOrigin;
			float c = Math::VectorLen(Q);
			float v = Math::VectorDot(Q, rVector);
			float d = (sRadius * sRadius) - ((c * c) - (v * v));

			if(d < 0.0f) return -1.0f;

			return v - sqrtf(d);
		};

		inline Math::Vector ClosestPointOnLine(Math::Vector A, Math::Vector B, Math::Vector P)
		{
			Math::Vector c = Math::VectorSub(P, A);
			Math::Vector V = Math::VectorNorm(Math::VectorSub(B, A));
			float d = Math::VectorDist(A, B);
			float t = Math::VectorDot(V, c);

			if(t < 0) return A;
			else if(t > d) return B;

			return Math::VectorAdd(A, Math::VectorScale(V, t));
		};

		SITHERROR BuildSurfacePCS(std::vector<Res::Level::SURFACE*>& PCS, std::list<int>& VisSec, Res::Level::SECTOR* curSector, const Math::Vector &Vel, const Math::Vector &Origin, float Rad);
		SITHERROR FindNearestPoint(Res::Level::SURFACE& Surface, const Math::Vector &Vel, const Math::Vector &Origin, float Rad, Math::Vector &Point);

		SITHERROR ctSphereWorld(Gameplay::Thing* curThing, float Frametime, COLLISION &FinalCollision);

		int isPointWithinSurface(const Res::Level::SURFACE *curSurface, const Math::Vector &Normal, const Math::Vector &Point);

		SITHERROR UpdatePhysicsThing(float FrameTime, int ThingRef);
		SITHERROR UpdatePathThing(float FrameTime, int ThingRef);

		SITHERROR UpdateThings();

		SITHERROR OldUpdatePhysicsThing(float FrameTime, int ThingRef);
		Sith::Math::Vector OldctSphereWorld(int ThingRef, int SectorRef, Math::Vector NewPosition);
	};
};

#endif // _PHYSICS_H_INCLUDED_