/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "vfile.h"

Sith::VFS::VFile::VFile(std::string theName, std::string theContainerFile, unsigned int theOffset, unsigned int theLength) : Name(theName), ContainerFileName(theContainerFile),
																															 ContainerFileOffset(theOffset), FileLength(theLength)
{
	return;
};

FILE* Sith::VFS::VFile::Open(bool isAscii)
{
	FILE *FileHandle;
	if(isAscii) FileHandle = fopen(ContainerFileName.c_str(), "rt");
	else FileHandle = fopen(ContainerFileName.c_str(), "rb");

	if(!FileHandle)
	{
		ConOut("ERROR: Failed to open container \'" + ContainerFileName + "\'.\n", ConOut::Error);
		return 0;
	};

	if(fseek(FileHandle, ContainerFileOffset, SEEK_SET) != 0)
	{
		ConOut("ERROR: File offset does not exist, container file \'" + ContainerFileName + "\' appears to be corrupted.\n", ConOut::Error);
		return 0;
	};

	return FileHandle;
};

const char* Sith::VFS::VFile::OpenBuffer()
{
	FILE *FileHandle = Open(false);
	if(!FileHandle)
	{
		ConOut("ERROR: Failed to open \'" + Name + "\'.\n", ConOut::Error);
		return 0;
	};

	char* Buffah = new char[FileLength + 1];
	if(!Buffah)
	{
		ConOut("ERROR: Out of memory, failed to allocate buffer for \'" + Name + "\'.\n", ConOut::Error);
		return 0;
	};

	Buffah[FileLength] = '\0';

	size_t BytesRead = fread(Buffah, sizeof(char), FileLength, FileHandle);
	if(BytesRead != FileLength)
	{
		ConOut("WARNING: File size indicated is incorrect, \'" + ContainerFileName + "\' may be corrupt.\n", ConOut::Warning);
		return 0;
	};

	fclose(FileHandle);

	return Buffah;
};