/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "resman.h"

Sith::VFS::VirtualFileSystem::VirtualFileSystem() : ResourceRoot("RES:\\"), PathRoot("PATH:\\"), CurrentEpisode(0)
{
	return;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::InitializeResource(std::string theResourcePath)
{
	std::ostringstream Stream;

	ResourcePath = theResourcePath;

	// ********* DANGER: WIN32 ONLY!!! **************

	// Initialize resource directory
	intptr_t findptr = 0;
	_finddata_t FindData;

	int NumGOBs = 0;

	// find resource containers
	findptr = _findfirst((ResourcePath + std::string("\\*.gob")).c_str(), &FindData);
	if(findptr != -1)
	{
		do
		{
			if(VFS::LoadFromGOBFile(&ResourceRoot, (ResourcePath + "\\" + FindData.name)) == SE_FILE_CORRUPT)
			{
				Stream << "WARNING: Resource path contained a corrupted GOB, \'" << FindData.name << "\'.\n";
				ConOut(Stream.str(), ConOut::Warning);
				Stream.str("");
			}
			else ++NumGOBs;
		} while(_findnext(findptr, &FindData) == 0);

		_findclose(findptr);
	}
	else
	{
		if(errno = ENOENT)
		{
			Stream << "ERROR: Attempted to set the Resource path to \'" << theResourcePath << "\', which does not exist!\n";
			ConOut(Stream.str(), ConOut::Error);
			return SE_FILE_INVALID_PATH;
		}
		else
		{
            Stream << "WARNING: Loading \'" << theResourcePath << "\' into Resources, but it does not contain any GOB files.\n";
			ConOut(Stream.str(), ConOut::Warning);
			return SE_FILE_WARNING;
		};
	};
	// Done initializing resource directory

	Stream << "\t| Loaded " << NumGOBs << " containers into Resources\n";
	ConOut(Stream.str(), ConOut::Pedantic);

	return SE_OK;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::InitializePath(std::string theModPath)
{
	std::ostringstream Stream;
	ModPath = theModPath;

	// ********* DANGER: WIN32 ONLY!!! **************

	// Initialize resource directory
	intptr_t findptr = 0;
	_finddata_t FindData;

	int NumGOBs = 0;

	// find resource containers
	findptr = _findfirst((ModPath + std::string("\\*.gob")).c_str(), &FindData);
	if(findptr != -1)
	{
		do
		{
			VFS::LoadFromGOBFile(&PathRoot, (ModPath + "\\" + FindData.name));
			++NumGOBs;
		} while(_findnext(findptr, &FindData) == 0);

		_findclose(findptr);
	}
	else
	{
		if(errno = ENOENT)
		{
			Stream << "ERROR: Attempted to set the modification path to \'" << theModPath << "\', which does not exist!\n";
			ConOut(Stream.str(), ConOut::Error);
			return SE_FILE_INVALID_PATH;
		}
		else
		{
            Stream << "WARNING: Loading \'" << theModPath << "\' into Modifications, but it does not contain any GOB files.\n";
			ConOut(Stream.str(), ConOut::Warning);
			return SE_FILE_WARNING;
		};
	};
	// Done initializing resource directory

	Stream << "\t| Loaded " << NumGOBs << " containers into Modifications\n";
	ConOut(Stream.str(), ConOut::Pedantic);

	return SE_OK;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::InitializeEpisode(std::string theEpisodePath)
{
	std::ostringstream Stream;
	EpisodeList.clear();
	EpisodePath = theEpisodePath;

	// ********* DANGER: WIN32 ONLY!!! **************

	// Initialize episode directory
	intptr_t findptr = 0;
	_finddata_t FindData;

	int NumGOBs = 0;

	// find resource containers
	findptr = _findfirst((EpisodePath + std::string("\\*.gob")).c_str(), &FindData);
	if(findptr != -1)
	{
		do
		{
            EpisodeList.push_back(Episode());

			VFS::LoadFromGOBFile(&EpisodeList.back().EpisodeContainer, (EpisodePath + "\\" + FindData.name));

			int i = 0;
			while(FindData.name[i] != '\0')
			{
				if(FindData.name[i] == '.') break;
				else EpisodeList.back().ShortName.push_back(FindData.name[i]);

				++i;
			};

			std::transform(EpisodeList.back().ShortName.begin(), EpisodeList.back().ShortName.end(), EpisodeList.back().ShortName.begin(), tolower);

			const char* epibuffer = EpisodeList.back().EpisodeContainer.GetFile("episode.jk")->OpenBuffer();
			if(epibuffer)
			{
                if(Failed(EpisodeList.back().Load(epibuffer)))
				{
                    delete[] epibuffer;
					Stream.str("");
					Stream << "ERROR: Episode data in container \'" << FindData.name << "\' is corrupt; not possible to initialize episode.\n";
					ConOut(Stream.str(), ConOut::Error);
					Stream.str("");
					EpisodeList.pop_back();
					continue;
				};
				delete[] epibuffer;
			}
			else
			{
				Stream.str("");
				Stream << "ERROR: Episode data in container \'" << FindData.name << "\' does not exist; container is not a valid episode.\n";
				ConOut(Stream.str(), ConOut::Error);
				Stream.str("");
				EpisodeList.pop_back();
				continue;
			};

			++NumGOBs;
		} while(_findnext(findptr, &FindData) == 0);

		_findclose(findptr);
	}
	else
	{
		if(errno = ENOENT)
		{
			Stream << "ERROR: Attempted to set the episode path to \'" << theEpisodePath << "\', which does not exist!\n";
			ConOut(Stream.str(), ConOut::Error);
			return SE_FILE_INVALID_PATH;
		}
		else
		{
            Stream << "WARNING: Loading \'" << theEpisodePath << "\' into Episodes, but it does not contain any GOB files.\n";
			ConOut(Stream.str(), ConOut::Warning);
			return SE_FILE_WARNING;
		};
	};
	// Done initializing resource directory

	Stream << "\t| Loaded " << NumGOBs << " containers into Episodes\n";
	ConOut(Stream.str(), ConOut::Pedantic);

	return SE_OK;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::Shutdown()
{
	ResourceRoot = VDirectory("RES:\\");
	PathRoot = VDirectory("PATH:\\");

	return SE_OK;
};

FILE* Sith::VFS::VirtualFileSystem::Open(std::string FilePath, bool isAscii)
{
	FILE *FileHandle = 0;
	VFile *GottenFile = 0;

	char OpenProperties[3] = "rb";
	if(isAscii) OpenProperties[1] = 't';

	// Check in episode directory
	FileHandle = fopen((EpisodePath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle) return FileHandle;

	// Check in active episode
	if(CurrentEpisode)
	{
		GottenFile = CurrentEpisode->EpisodeContainer.GetFile(FilePath);
		if(GottenFile) return GottenFile->Open(isAscii);
	};

	// Check in path directory
	FileHandle = fopen((ModPath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle) return FileHandle;

	// Check in path vdirectory
	GottenFile = PathRoot.GetFile(FilePath);
	if(GottenFile) return GottenFile->Open(isAscii);

	// Check in resource directory
	FileHandle = fopen((ResourcePath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle) return FileHandle;

	// Check in resource vdirectory
	GottenFile = ResourceRoot.GetFile(FilePath);
	if(GottenFile) return GottenFile->Open(isAscii);

	return 0;
};

const char* Sith::VFS::VirtualFileSystem::OpenBuffer(std::string FilePath)
{
	FILE *FileHandle = 0;
	VFile *GottenFile = 0;
	long int FSLength = 0;
	const char* Buffa = 0;

	char OpenProperties[3] = "rb";

	// Check in episode directory
	FileHandle = fopen((EpisodePath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle)
	{
		Buffa = OpenFSBuffer(FileHandle);
		if(Buffa) return Buffa;
	};

	// Check in active episode
	if(CurrentEpisode)
	{
		GottenFile = CurrentEpisode->EpisodeContainer.GetFile(FilePath);
		if(GottenFile)
		{
			Buffa = GottenFile->OpenBuffer();
			if(Buffa) return Buffa;
		};
	};

	// Check in path directory
	FileHandle = fopen((ModPath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle)
	{
		Buffa = OpenFSBuffer(FileHandle);
		if(Buffa) return Buffa;
	};

	// Check in path vdirectory
	GottenFile = PathRoot.GetFile(FilePath);
	if(GottenFile)
	{
		Buffa = GottenFile->OpenBuffer();
		if(Buffa) return Buffa;
	};

	// Check in resource directory
	FileHandle = fopen((ResourcePath + "\\" + FilePath).c_str(), OpenProperties);
	if(FileHandle)
	{
		Buffa = OpenFSBuffer(FileHandle);
		if(Buffa) return Buffa;
	};

	// Check in resource vdirectory
	GottenFile = ResourceRoot.GetFile(FilePath);
	if(GottenFile)
	{
		Buffa = GottenFile->OpenBuffer();
		if(Buffa) return Buffa;
	};

	return 0;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::MountEpisode(std::string ShortName)
{
	ConOut("Mounting episode \'" + ShortName + "\'...\n", ConOut::Pedantic);
	std::transform(ShortName.begin(), ShortName.end(), ShortName.begin(), tolower);
	for(std::list<Episode>::iterator it = EpisodeList.begin(); it != EpisodeList.end(); ++it)
	{
		if(it->ShortName == ShortName)
		{
			CurrentEpisode = &(*it);
			Sith::COG::CogLib::LoadCOGStrings("misc\\cogstrings.uni");
			ConOut("Mounted\n\n", ConOut::Pedantic);
			return SE_OK;
		};
	};
    
	ConOut("WARNING: Attempted to mount episode \'" + ShortName + "\', but it does not exist.\n", ConOut::Warning);

	return SE_FILE_NOT_FOUND;
};

Sith::SITHERROR Sith::VFS::VirtualFileSystem::UnmountEpisode()
{
	ConOut("Episode \'" + CurrentEpisode->ShortName + "\' has been unmounted.\n", ConOut::Pedantic);
	CurrentEpisode = 0;
	return SE_OK;
};