/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "vmtype.h"
#include "opcodes.h"
#include "instruction.h"
#include "vmheap.h"
#include "stackmachine.h"

#include "di_input.h"

std::vector<Sith::COG::VMHeap::Namespace> Sith::COG::VMHeap::Global;
int Sith::COG::VMHeap::GlobalCount = 0;

Sith::SITHERROR Sith::COG::VMHeap::SerializeGlobal(std::string TargetFileName)
{
	FILE* DestFile = 0;
	DestFile = fopen(TargetFileName.c_str(), "w");
	if(!DestFile)
	{
		ConOut("ERROR: Could not serialize global heap to file \'" + TargetFileName + "\'.\n", ConOut::Error);
		return SE_FILE_ERROR;
	};

	fprintf(DestFile, "# This file has been automatically generated by Sith 2\n\n");

	std::ostringstream Stream;

	for(std::vector<Namespace>::iterator it = Global.begin(); it != Global.end(); ++it)
	{
		for(std::vector<HeapElement>::iterator jt = it->Heap.begin(); jt != it->Heap.end(); ++jt)
		{
			Stream.str("");
			Stream << it->Name << "::" << jt->Name << " = ";

			switch(jt->Data.GetType())
			{
			case VMTC_STRING:
				{
					std::string InStr = jt->Data.GetString();
					std::string StrLit;

					StrLit.push_back('\"');
					for(unsigned int i = 0; i < InStr.length(); ++i)
					{
						if(InStr[i] == '\n') StrLit.push_back('\n');
						else if(InStr[i] == '\t') StrLit.push_back('\t');
						else if(InStr[i] == '\"') StrLit.push_back('\"');
						else if(InStr[i] == '\'') StrLit.push_back('\'');
						else StrLit.push_back(InStr[i]);
					};
					StrLit.push_back('\"');

					Stream << StrLit;
				};
				break;
			case VMTC_INT:
				Stream << jt->Data.GetInt();
				break;
			case VMTC_BOOL:
				{
					if(jt->Data.GetBool() == true) Stream << "true";
					else Stream << "false";
				};
				break;
			case VMTC_FLOAT:
			case VMTC_DOUBLE:
			case VMTC_FLEX:
				Stream << jt->Data.GetFloat();
				break;
			default:
				// Default - type not defined! OMG! LOL!
				// Skip this one. In fact, erase the whole thing and go to the next one.
				continue;
			};

			Stream << ";\n";

			fwrite(Stream.str().c_str(), Stream.str().length(), sizeof(char), DestFile);
		};
	};

	Stream.str("");
	Stream << std::endl;
	fwrite(Stream.str().c_str(), Stream.str().length(), sizeof(char), DestFile);

	// Write input settings
	for(std::multimap<std::string, std::pair<std::string, float> >::iterator it = Win32::Input::Get().Bindings.begin();
		it != Win32::Input::Get().Bindings.end(); ++it)
	{
		Stream.str("");

		Stream << "bind(";
		Stream << "\"" << it->second.first << "\"";
		Stream << ", ";
		Stream << "\"" << it->first << "\"";

		if(it->second.second != 1)
		{
			Stream << ", ";
			Stream << it->second.second;
		};

		Stream << ");\n";

		fwrite(Stream.str().c_str(), Stream.str().length(), sizeof(char), DestFile);
	};

	fclose(DestFile);

	return SE_OK;
};