/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _VMTYPE_H_INCLUDED_
#define _VMTYPE_H_INCLUDED_
 
/********** WARNING, TO PRESERVE YOUR SANITY! **************/
/*	I'm not particularly fond of this bit of code, but it
served its purpose during development. I plan on rewriting
basically everything in here. Essentially it's just a bunch
of operators with internal switch statements - definitely
not something I could see myself maintaining for any serious
length of time. Keep that in mind. If you write an improvement
feel free to submit it to me. */

namespace Sith
{
	namespace COG
	{
		enum VMTypeCode
		{
			VMTC_VOID,

			VMTC_INSTRUCTION_ADDRESS,
			VMTC_HEAP_ADDRESS,

			VMTC_SYSTEM_CALL,
			VMTC_LIBRARY_CALL,

			VMTC_AI,
			VMTC_BOOL,
			VMTC_COG,
			VMTC_DOUBLE,
			VMTC_FLEX,
			VMTC_FLOAT,
			VMTC_INT,
			VMTC_KEYFRAME,
			VMTC_MATERIAL,
			VMTC_MODEL,
			VMTC_SECTOR,
			VMTC_SOUND,
			VMTC_STRING,
			VMTC_SURFACE,
			VMTC_TEMPLATE,
			VMTC_THING,

			VMTC_VECTOR // Added July 20, 2005 [JC]
		};

		// "Light" version of the VMType class
		// - For use by FLEX/BISON
		struct VMTypeLite
		{
			VMTypeCode TypeCode;
			unsigned char Data[sizeof(float) * 3];
		};

		class VMType
		{
		private:
			VMTypeCode TypeCode;
			unsigned char Data[sizeof(float) * 3];		// Data should be large enough to contain any of the types, including a VECTOR

		public:
			VMType()
			{
				TypeCode = VMTC_VOID;
				memset(Data, 0, sizeof(Data));

				return;
			};

			VMType(VMTypeLite vTypeLite)
			{
				TypeCode = vTypeLite.TypeCode;
				memcpy(Data, vTypeLite.Data, sizeof(Data));

				return;
			};

			VMType(VMTypeCode vTypeCode)
			{
				TypeCode = vTypeCode;
				memset(Data, 0, sizeof(Data));

				return;
			};

			VMType(VMTypeCode vTypeCode, int Value, int Subaddress)
			{
				TypeCode = vTypeCode;
				memset(Data, 0, sizeof(Data));

				*(int*)Data = Value;
				*(int*)(Data + sizeof(int)) = Subaddress;
			};

			VMType(bool Value)
			{
				TypeCode = VMTC_BOOL;
				memset(Data, 0, sizeof(Data));

				*(bool*)Data = Value;
			};

			VMType(int Value)
			{
				TypeCode = VMTC_INT;
				memset(Data, 0, sizeof(Data));

				*(int*)Data = Value;
			};

			VMType(double Value)
			{
				TypeCode = VMTC_DOUBLE;
				memset(Data, 0, sizeof(Data));

				*(double*)Data = Value;
			};

			VMType(float Value)
			{
				TypeCode = VMTC_FLOAT;
				memset(Data, 0, sizeof(Data));

				*(float*)Data = Value;
			};

			VMType(Math::Vector Value)
			{
				TypeCode = VMTC_VECTOR;
				memset(Data, 0, sizeof(Data));

				*(Math::Vector*)Data = Value;
			};

			VMType(std::string Value)
			{
				TypeCode = VMTC_STRING;
				*(char**)Data = strdup(Value.c_str());

				return;
			};

			VMType(COG::VMTypeCode tc, std::string Value)
			{
				if((tc == VMTC_STRING) || (tc == VMTC_TEMPLATE))
				{
					TypeCode = tc;
					*(char**)Data = strdup(Value.c_str());
				}
				else
				{
					TypeCode = VMTC_VOID;
				};

				return;
			};

			VMType(const VMType &Value) // copy ctor
			{
				if((Value.TypeCode == VMTC_STRING) || (Value.TypeCode == VMTC_TEMPLATE))
				{
					TypeCode = VMTC_STRING;
					*(char**)Data = strdup(*(char**)Value.Data);
				}
				else
				{
					TypeCode = Value.TypeCode;
					memcpy(Data, Value.Data, sizeof(Data));
				};

				return;
			};

			~VMType()
			{
				if((TypeCode == VMTC_STRING) || (TypeCode == VMTC_TEMPLATE))
				{
					delete[] *(char**)Data;
				};
				return;
			};

			inline VMTypeLite GetLite()
			{
				VMTypeLite ReturnValue;

				ReturnValue.TypeCode = TypeCode;
				memcpy(ReturnValue.Data, Data, sizeof(Data));

				return ReturnValue;
			};

			inline VMTypeCode GetType()
			{
				return TypeCode;
			};

			inline bool GetBool()
			{
				if(TypeCode == VMTC_VOID)
				{
					return false;
				};

				if(*(bool*)Data)
				{
					return true;
				}
				else
				{
					return false;
				};	
			};

			inline int GetInt()
			{
				switch(TypeCode)
				{
				case VMTC_BOOL:
					if(*(bool*)Data)
					{
						return 1;
					}
					else
					{
						return 0;
					};
					break;

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return *(int*)Data;

				case VMTC_DOUBLE:
					return (int)*(double*)Data;

				case VMTC_FLOAT:
					return (int)*(float*)Data;

				case VMTC_VOID:				
				default:
					return 0;
				};
			};

			inline int GetAddr()
			{
				switch(TypeCode)
				{
				case VMTC_INSTRUCTION_ADDRESS:
				case VMTC_HEAP_ADDRESS:
				case VMTC_SYSTEM_CALL:
				case VMTC_LIBRARY_CALL:
					return *(int*)Data;

				default:
					return -1;
				};
			};

			inline int GetAddrBoost()
			{
				switch(TypeCode)
				{
				case VMTC_HEAP_ADDRESS:
					return *(int*)(Data + sizeof(int));

				default:
					return -1;
				};
			};

			inline std::string GetString()
			{
				if((TypeCode == VMTC_STRING) || (TypeCode == VMTC_TEMPLATE))
				{
					return std::string(*(char**)Data);
				}
				else
				{
					return "";
				};
			};

			inline std::string GetTpl()
			{
				return GetString();
			};

			inline double GetDouble()
			{
				switch(TypeCode)
				{
				case VMTC_BOOL:
					if(*(bool*)Data)
					{
						return 1.0;
					}
					else
					{
						return 0.0;
					};
					break;

				case VMTC_INSTRUCTION_ADDRESS:
				case VMTC_HEAP_ADDRESS:
				case VMTC_SYSTEM_CALL:
				case VMTC_LIBRARY_CALL:
					return 0.0;

				case VMTC_INT:
					return (double)*(int*)Data;

				case VMTC_DOUBLE:
					return *(double*)Data;

				case VMTC_FLOAT:
					return (double)*(float*)Data;

				case VMTC_VOID:
				default:
					return 0.0;
				};
			};

			inline float GetFloat()
			{
				switch(TypeCode)
				{
				case VMTC_BOOL:
					if(*(bool*)Data)
					{
						return 1.0f;
					}
					else
					{
						return 0.0f;
					};
					break;

				case VMTC_INSTRUCTION_ADDRESS:
				case VMTC_HEAP_ADDRESS:
				case VMTC_SYSTEM_CALL:
				case VMTC_LIBRARY_CALL:
					return 0.0f;

				case VMTC_INT:
					return (float)*(int*)Data;

				case VMTC_DOUBLE:
					return (float)*(double*)Data;

				case VMTC_FLOAT:
					return *(float*)Data;

				case VMTC_VOID:
				default:
					return 0.0;
				};
			};

			inline Math::Vector GetVector()
			{
				switch(TypeCode)
				{
				case VMTC_VECTOR:
					return *(Math::Vector*)Data;

				default:
					return Math::Vector(0, 0, 0);
				};
			};

			inline VMType operator=(VMType Val)
			{
				// If the typecode is VOID, do **NOT** assign a new value
				if(Val.TypeCode == VMTC_STRING)
				{
					TypeCode = VMTC_STRING;
					*(char**)Data = strdup(*(char**)Val.Data);
					return VMType(std::string(*(char**)Data));
				}
				else if(Val.TypeCode != VMTC_VOID)
				{
					TypeCode = Val.TypeCode;
					memcpy(Data, Val.Data, sizeof(Data));
				};

				return *this;
			};

			inline VMType operator=(bool Val)
			{
				TypeCode = VMTC_BOOL;
				*(bool*)Data = Val;

				return VMType(*this);
			};	

			inline VMType operator=(int Val)
			{
				TypeCode = VMTC_INT;
				*(int*)Data = Val;

				return VMType(*this);
			};

			inline VMType operator=(double Val)
			{
				TypeCode = VMTC_DOUBLE;
				*(double*)Data = Val;

				return VMType(*this);
			};

			inline VMType operator=(float Val)
			{
				TypeCode = VMTC_FLOAT;
				*(float*)Data = Val;

				return VMType(*this);
			};

			inline bool operator==(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() == Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() == Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() == Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() == Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator!=(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() != Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() != Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() != Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() != Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator>(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() > Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() > Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() > Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() > Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator<(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() < Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() < Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() < Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() < Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator>=(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() >= Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() >= Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() >= Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() >= Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator<=(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() <= Val.GetBool();

				case VMTC_INT:
				case VMTC_SECTOR:
				case VMTC_SURFACE:
				case VMTC_THING:
					return GetInt() <= Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() <= Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() <= Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator&&(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() && Val.GetBool();

				case VMTC_INT:
					return GetInt() && Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() && Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() && Val.GetDouble();

				default:
					return false;
				};
			};

			inline bool operator||(VMType Val)
			{
				switch(TypeCode)
				{
				case VMTC_VOID:
					return false;

				case VMTC_BOOL:
					return GetBool() || Val.GetBool();

				case VMTC_INT:
					return GetInt() || Val.GetInt();

				case VMTC_FLOAT:
					return GetFloat() || Val.GetFloat();

				case VMTC_DOUBLE:
					return GetDouble() || Val.GetDouble();

				default:
					return false;
				};
			};

			inline VMType operator+(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() + RValue.GetBool());
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() + RValue.GetInt());
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() + RValue.GetDouble());
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() + RValue.GetFloat());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator-(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() - RValue.GetBool());
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() - RValue.GetInt());
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() - RValue.GetDouble());
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() - RValue.GetFloat());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator*(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() * RValue.GetBool());
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() * RValue.GetInt());
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() * RValue.GetDouble());
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() * RValue.GetFloat());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator/(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(false);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() / RValue.GetInt());
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() / RValue.GetDouble());
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() / RValue.GetFloat());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator%(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(false);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() % RValue.GetInt());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator&(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(bool(GetBool() & RValue.GetBool()));
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() & RValue.GetInt());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator|(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(bool(GetBool() | RValue.GetBool()));
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() | RValue.GetInt());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator^(VMType RValue)
			{
				if(TypeCode == VMTC_VOID || RValue.TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(bool(GetBool() ^ RValue.GetBool()));
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() ^ RValue.GetInt());
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator++()
			{
				if(TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() + 1);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() + 1);
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() + 1.0);
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() + 1.0f);
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator--()
			{
				if(TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() - 1);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() - 1);
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() - 1.0);
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() - 1.0f);
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator<<(int RValue)
			{
				if(TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() << RValue);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() << RValue);
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() * (2 << (RValue - 1)));
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() * (2 << (RValue - 1)));
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};

			inline VMType operator>>(int RValue)
			{
				if(TypeCode == VMTC_VOID)
				{
					return VMType(VMTC_VOID);
				}
				else if(TypeCode == VMTC_BOOL)
				{
					return VMType(GetBool() << RValue);
				}
				else if(TypeCode == VMTC_INT)
				{
					return VMType(GetInt() << RValue);
				}
				else if(TypeCode == VMTC_DOUBLE)
				{
					return VMType(GetDouble() / (2 << (RValue - 1)));
				}
				else if(TypeCode == VMTC_FLOAT)
				{
					return VMType(GetFloat() / (2 << (RValue - 1)));
				}
				else
				{
					return VMType(VMTC_VOID);
				};
			};


		};
	};
};
#endif // _VMTYPE_H_INCLUDED_
