/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "win_chooser.hpp"

#include "vfs.h"


HWND Sith::Win32::WinChooser::episodeListH;
HWND Sith::Win32::WinChooser::mapListH;


LRESULT CALLBACK Sith::Win32::WinChooser::ConWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_CREATE:
		{
			episodeListH = CreateWindowEx(WS_EX_CLIENTEDGE, "LISTBOX", "", 
				WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_NOTIFY | LBS_SORT, 
			    0, 0, 0, 0, hWnd, NULL, GetModuleHandle(NULL), NULL);

			mapListH = CreateWindowEx(WS_EX_CLIENTEDGE, "LISTBOX", "", 
				WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_NOTIFY | LBS_SORT, 
			    0, 0, 0, 0, hWnd, NULL, GetModuleHandle(NULL), NULL);

			Redraw(hWnd, uMsg, wParam, lParam);
		}
		break;

		//case WM_SIZE:
	case WM_WINDOWPOSCHANGED:
		{
			LPWINDOWPOS wp = (LPWINDOWPOS)lParam;
			if (wp->flags & SWP_NOSIZE)  // if set
			{
				break;
			}
			Redraw(hWnd, uMsg, wParam, lParam);
		}
		break;

	case WM_ACTIVATE:
		//SetFocus(GetDlgItem(hWnd, WinChooser::hInDlgItem));
		break;

	case WM_CLOSE:
		DestroyWindow(hWnd);
		break;

	case WM_DESTROY:
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);
		break;

	case WM_COMMAND:
		{
			if ( (HWND) lParam == episodeListH && HIWORD(wParam) == LBN_SELCHANGE )
			{
				char episodeName[256];
				int index = SendMessage(episodeListH, LB_GETCURSEL, (WPARAM) 0, (LPARAM) 0);
				SendMessage(episodeListH, LB_GETTEXT, (WPARAM) index, (LPARAM) episodeName);
				Sith::VFS::VirtualFileSystem::Get().MountEpisode((std::string) episodeName);

				SendMessage(mapListH, LB_RESETCONTENT, (WPARAM) 0, (LPARAM) 0);
				Sith::VFS::Episode *  currentEpisode;
				currentEpisode = Sith::VFS::VirtualFileSystem::Get().GetCurrentEpisode();
				for(std::list<Sith::VFS::Episode::EpisodeElement>::iterator it = currentEpisode->LevelList.begin(); it != currentEpisode->LevelList.end(); ++it)
				{
					//Sith::COG::CogLib::GetUniString(it->ShortName)
					if(it->Type == VFS::Episode::LEVEL)
					{
                        SendMessage(mapListH, LB_ADDSTRING, (WPARAM) -1, (LPARAM) it->ShortName.c_str());
					};
				};
			}
			else
			if ( (HWND) lParam == mapListH && HIWORD(wParam) == LBN_SELCHANGE )
			{
				char mapName[256];
				int index = SendMessage(mapListH, LB_GETCURSEL, (WPARAM) 0, (LPARAM) 0);
				SendMessage(mapListH, LB_GETTEXT, (WPARAM) index, (LPARAM) mapName);
				//TODO: run map
				//Sith::VFS::VirtualFileSystem::Get().MountEpisode((std::string) mapName);
				//std::string  mapRunName;
				//mapRunName = "map(\"";
				//mapRunName += mapName;
				//mapRunName += "\");";
				//Sith::System::ConIn(mapRunName);
				std::stringstream Stream; 
				Stream << "map(\"" << mapName << "\");";
				System::ConIn(Stream.str());
			}
		}
		break;


	default:
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}

	return 0;
}

Sith::Win32::WinChooser::Redraw(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    RECT rcClient;

    GetClientRect(hWnd, &rcClient);

	int  margin = 5;
    SetWindowPos(episodeListH, NULL, 
		margin, margin, 
		(rcClient.right / 2) - 2 * margin, rcClient.bottom - 2 * margin, 
		SWP_NOZORDER);
    SetWindowPos(mapListH, NULL, 
		(rcClient.right / 2) + margin, margin, 
		(rcClient.right / 2) - 2 * margin, rcClient.bottom - 2 * margin, 
		SWP_NOZORDER);

    RedrawWindow(hWnd, NULL, NULL,
            RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
}

Sith::Win32::WinChooser::WinChooser() : Task("Win32 Level Chooser")
{
       //ConOut::RegisterOutput(Stream, ConOut::Standard | ConOut::Network | ConOut::Echo | ConOut::Warning |
       //                                                                                ConOut::Error | ConOut::Critical | ConOut::Pedantic);

       wc.cbSize = sizeof(WNDCLASSEX);
       wc.style = 0;
       wc.lpfnWndProc = ConWndProc;
       wc.cbClsExtra = 0;
       wc.cbWndExtra = 0;
       wc.hInstance = GetModuleHandle(NULL);
       wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
       wc.hCursor = LoadCursor(NULL, IDC_ARROW);
       wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
       wc.lpszMenuName = NULL;
       wc.lpszClassName = "Sith2 Chooser";
       wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

       if(!RegisterClassEx(&wc))
       {
               MessageBox(GetDesktopWindow(), "Failed to register chooser window class", "Error", MB_OK | MB_ICONERROR);
               ConOut("Critical error: Failed to register chooser window class\n", ConOut::Critical);
			   Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

               return;
       };

       hMainWnd = CreateWindowEx(WS_EX_CLIENTEDGE, "Sith2 Chooser", "Chooser", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
               800, 600, NULL, NULL, GetModuleHandle(NULL), NULL);

       if(hMainWnd == NULL)
       {
               MessageBox(GetDesktopWindow(), "Failed to create chooser window", "Error", MB_OK | MB_ICONERROR);
               ConOut("Critical error: Failed to create chooser window\n", ConOut::Critical);
			   Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

               return;
       };

       ShowWindow(hMainWnd, SW_SHOWDEFAULT);
       UpdateWindow(hMainWnd);

       return;
}

Sith::Win32::WinChooser::~WinChooser()
{
       return;
}

Sith::SITHERROR Sith::Win32::WinChooser::Startup()
{
	std::list<Sith::VFS::Episode>  EpisodeList;
	EpisodeList = Sith::VFS::VirtualFileSystem::Get().GetEpisodeList();

	for(std::list<Sith::VFS::Episode>::iterator it = EpisodeList.begin(); it != EpisodeList.end(); ++it)
	{
		SendMessage(episodeListH, LB_ADDSTRING, (WPARAM) -1, (LPARAM) it->ShortName.c_str());
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Win32::WinChooser::Process()
{
       // Win32 Message Pump:
       while(PeekMessage(&Msg, hMainWnd, 0U, 0U, PM_REMOVE) > 0)
       {
               if((Msg.message == WM_KEYDOWN)
                       && (
                       (Msg.wParam == VK_RETURN)
                       || (Msg.wParam == VK_UP)
                       || (Msg.wParam == VK_DOWN) ) )
               {
                       Msg.hwnd = hMainWnd;
               };

               TranslateMessage(&Msg);
               DispatchMessage(&Msg);
       };

       return Sith::SE_OK;
}