/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// SplitterBar.cpp



#include "stdafx.h"
#include "win_splitterbar.hpp"
#include <windowsx.h>



const long  Sith::Win32::SplitterBar::SplitterBarThickness = 3;
const long  Sith::Win32::SplitterBar::SplitterBarMiddle    = 1;  // (zero-based)

HWND       Sith::Win32::SplitterBar::parentWdwH;
Sith::Win32::SplitterBarType  Sith::Win32::SplitterBar::type;
HWND             Sith::Win32::SplitterBar::wdwHandle;
long  Sith::Win32::SplitterBar::left;
long  Sith::Win32::SplitterBar::top;
long  Sith::Win32::SplitterBar::width;
long  Sith::Win32::SplitterBar::height;
long  Sith::Win32::SplitterBar::offset;
long  Sith::Win32::SplitterBar::dragStart;
long  Sith::Win32::SplitterBar::drawOffset;
long  Sith::Win32::SplitterBar::drawX;
long  Sith::Win32::SplitterBar::drawY;
LONG  Sith::Win32::SplitterBar::oldStyle;
LONG  Sith::Win32::SplitterBar::style;



Sith::Win32::SplitterBar::SplitterBar
(
	HWND  newParentWindow, 
	SplitterBarType  newType, 
	long  newOffset
)
{
	WNDCLASSEX  splitterBarClass;
	std::string      windowTitle;
	HCURSOR     cursor;
	RECT        mainClientRect;

	type       = newType;
	//instance   = newInstance;
	parentWdwH = newParentWindow;

	if (type == Horizontal)
	{
		className = "HSplitterBarClass";
		windowTitle = "Horizontal Splitter Bar";
		cursor = LoadCursor(GetModuleHandle(NULL), "CUR_HSPLIT");
	}
	else
	{
		className = "VSplitterBarClass";
		windowTitle = "Vertical Splitter Bar";
		cursor = LoadCursor(GetModuleHandle(NULL), "CUR_VSPLIT");
	}

    splitterBarClass.cbSize        = sizeof(WNDCLASSEX);
    splitterBarClass.style         = 0;
    splitterBarClass.lpfnWndProc   = ProcessWindow;
    splitterBarClass.cbClsExtra    = 0;
    splitterBarClass.cbWndExtra    = 0;
    splitterBarClass.hInstance     = GetModuleHandle(NULL);
    splitterBarClass.hIcon         = NULL;
    splitterBarClass.hCursor       = cursor; 
    splitterBarClass.hbrBackground = (HBRUSH) (COLOR_3DFACE + 1);
    splitterBarClass.lpszMenuName  = NULL;
    splitterBarClass.lpszClassName = className.c_str();
    splitterBarClass.hIconSm       = NULL;

    RegisterClassEx(&splitterBarClass);

	GetClientRect(parentWdwH, &mainClientRect);

	offset = newOffset;
	if (type == Horizontal)
	{
		width  = mainClientRect.right;
		height = SplitterBarThickness;
		if (offset < 0)
		{
			offset = 0;
		}
		else if (offset + SplitterBarThickness 
			> mainClientRect.bottom)
		{
			offset 
				= mainClientRect.bottom - SplitterBarThickness;
		}

		wdwHandle = CreateWindowEx(
			0,
			className.c_str(), windowTitle.c_str(),
			WS_CHILD,
			0, offset, 
			width, height, 
			parentWdwH, NULL, GetModuleHandle(NULL), NULL);
	}
	else
	{
		width  = SplitterBarThickness;
		height = mainClientRect.bottom;
		if (offset < 0)
		{
			offset = 0;
		}
		else if (offset + SplitterBarThickness 
			> mainClientRect.right)
		{
			offset 
				= mainClientRect.right - SplitterBarThickness;
		}

		wdwHandle = CreateWindowEx(
			0,
			className.c_str(), windowTitle.c_str(),
			WS_CHILD,
			offset, 0, width, height,
			parentWdwH, NULL, GetModuleHandle(NULL), NULL);
	}
}



Sith::Win32::SplitterBar::~SplitterBar
(
)
{
	if (wdwHandle != NULL)
	{
		DestroyWindow(wdwHandle);
		wdwHandle = NULL;
		UnregisterClass(className.c_str(), GetModuleHandle(NULL));
	}
}



// Reposition bar based on given remaining client rect of parent window.
// Returns rect of bar window.
long  
Sith::Win32::SplitterBar::Reposition
(
	RECT &  wdwRect
)
{

	if (type == Horizontal)
	{
		RECT  mainClientRect;
		GetClientRect(parentWdwH, &mainClientRect);

		LONG  top;
		LONG  topMax;
		top = mainClientRect.bottom - offset;
		if (top < wdwRect.top)
		{
			top = wdwRect.top;
		}
		topMax = wdwRect.top + wdwRect.bottom - SplitterBarThickness;
		if (topMax < 0)
		{
			topMax = 0;
		}
		if (top > topMax)
		{
			top = topMax;
		}
		offset = mainClientRect.bottom - top;

		wdwRect.top = top;
		wdwRect.bottom = SplitterBarThickness;
	}
	else
	{
		wdwRect.left  = offset;
		wdwRect.right = SplitterBarThickness;
	}

	SetWindowPos(wdwHandle, NULL, 
		wdwRect.left, wdwRect.top, wdwRect.right, wdwRect.bottom,  
		SWP_NOZORDER);

	return 0;
}



long  
Sith::Win32::SplitterBar::ShowWindow
(
	bool  wdwIsVisible
)
{
	if (wdwIsVisible)
	{
		::ShowWindow(wdwHandle, SW_SHOW);
	}
	else
	{
		::ShowWindow(wdwHandle, SW_HIDE);
	}

	return 1;
}



HWND  
Sith::Win32::SplitterBar::GetWindowHandle
(
)
{
	return wdwHandle;
}



long  
Sith::Win32::SplitterBar::GetOffset
(
)
{
	return offset;
}



LRESULT CALLBACK   
Sith::Win32::SplitterBar::ProcessWindow
(
	HWND hwnd, 
	UINT message, 
	WPARAM wParam, 
	LPARAM lParam
)
{
	RECT    mainClientRect;
	POINTS  pt;
	int xParent;
	int xChild;
	int yParent;
	int yChild;
	POINT ptOrg = {0, 0};
	long  newLeft;
	long  newTop;

    switch (message)
    {
    case WM_MOVE:
        left = GET_X_LPARAM(lParam);
		top  = GET_Y_LPARAM(lParam);
//
//		if (type == Horizontal)
//		{
//			GetClientRect(parentWdwH, &mainClientRect);
//			offset = mainClientRect.bottom - top;
////			offset = top;
//		}
//		else
//		{
//			offset = left;
//		}
//
        return 0;

    case WM_SIZE:
        width  = LOWORD(lParam);
		height = HIWORD(lParam);
        return 0;

    case WM_LBUTTONDOWN:
		SetCapture(hwnd);

		//NOTE: Splitter line will show only if main wdw does not
		//      keep children from drawing in its wdw.
		oldStyle = GetWindowLong(parentWdwH, GWL_STYLE);
		if ((oldStyle & WS_CLIPCHILDREN) == WS_CLIPCHILDREN)
		{
			style = oldStyle ^ WS_CLIPCHILDREN;
			SetWindowLong(parentWdwH, GWL_STYLE, style);
		}
		else
		{
			style = oldStyle;
		}

		// Find x offset of splitter bar in main client coordinates.
		ClientToScreen(parentWdwH, &ptOrg);
		xParent = ptOrg.x;
		yParent = ptOrg.y;
		ptOrg.x = 0;
		ptOrg.y = 0;
		ClientToScreen(hwnd, &ptOrg);
		xChild = ptOrg.x;
		yChild = ptOrg.y;

        pt = MAKEPOINTS(lParam);

		// Drag coordinates are based on the origin of the splitter bar.
		if (type == Horizontal)
		{
			dragStart = yChild - yParent;
			drawOffset = SplitterBarMiddle - pt.y;
			drawY = dragStart + SplitterBarMiddle;
		}
		else
		{
			dragStart = xChild - xParent;
			drawOffset = SplitterBarMiddle - pt.x;
			drawX = dragStart + SplitterBarMiddle;
		}

		// Draw a divider using XOR mode
		DrawSplitterLine();
        return 0;

    case WM_MOUSEMOVE:
        if (wParam & MK_LBUTTON)
		{
			pt = MAKEPOINTS(lParam);

			// Erase previous divider and draw new one
			if (type == Horizontal)
			{
				DrawSplitterLine(true, pt.y);
			}
			else
			{
				DrawSplitterLine(true, pt.x);
			}
		}
        return 0;

    case WM_LBUTTONUP:
        pt = MAKEPOINTS(lParam);
		// Calling ReleaseCapture will send WM_CAPTURECHANGED
		ReleaseCapture();

		GetClientRect(parentWdwH, &mainClientRect);

		if (type == Horizontal)
		{
			newTop = dragStart + pt.y + drawOffset - SplitterBarMiddle;
			if (newTop < 0)
			{
				newTop = 0;
			}
			else if (newTop + SplitterBarThickness 
				> mainClientRect.bottom)
			{
				newTop 
					= mainClientRect.bottom - SplitterBarThickness;
			}

			newLeft = 0;
			offset = mainClientRect.bottom - newTop;
		}
		else
		{
			newLeft = dragStart + pt.x + drawOffset - SplitterBarMiddle;
			if (newLeft < 0)
			{
				newLeft = 0;
			}
			else if (newLeft + SplitterBarThickness 
				> mainClientRect.right)
			{
				newLeft 
					= mainClientRect.right - SplitterBarThickness;
			}

			newTop = 0;
			offset = newLeft;
		}

		SetWindowPos(wdwHandle, NULL, 
			newLeft, newTop, 0, 0,  
			SWP_NOSIZE | SWP_NOZORDER);

		SendMessage(parentWdwH, MSG_MOVESPLITTERBAR, 0, (LPARAM) wdwHandle);
        return 0;

    case WM_CAPTURECHANGED:
		// Splitter bar lost capture, so end dragging.
		// Erase splitter line.
		DrawSplitterLine();

		if ((oldStyle & WS_CLIPCHILDREN) == WS_CLIPCHILDREN)
		{
			SetWindowLong(parentWdwH, GWL_STYLE, oldStyle);
		}
        return 0;
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}



void  
Sith::Win32::SplitterBar::DrawSplitterLine
(
	bool  splitterBarIsDragging, 
	SHORT  coord
)
{
    HDC  _hdc;
    int _modeOld;

	_hdc = GetDC(parentWdwH);
    _modeOld = GetROP2(_hdc);
    SetROP2(_hdc, R2_NOTXORPEN);

	if (type == Horizontal)
	{
		MoveToEx(_hdc, left, drawY, 0);
		LineTo(_hdc, left + width - 1, drawY);

		if (splitterBarIsDragging)
		{
			drawY = dragStart + coord + drawOffset;
			MoveToEx(_hdc, left, drawY, 0);
			LineTo(_hdc, left + width - 1, drawY);
		}
	}
	else
	{
		MoveToEx(_hdc, drawX, top, 0);
		LineTo(_hdc, drawX, top + height - 1);

		if (splitterBarIsDragging)
		{
			drawX = dragStart + coord + drawOffset;
			MoveToEx(_hdc, drawX, top, 0);
			LineTo(_hdc, drawX, top + height - 1);
		}
	}

    SetROP2(_hdc, _modeOld);
	ReleaseDC(parentWdwH, _hdc);
}
