/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _WISEPTR_H_INCLUDED_
#define _WISEPTR_H_INCLUDED_

#include <assert.h>

namespace Sith
{
	namespace VFS
	{
		template <typename ResType> class WisePointer
		{
		protected:
			ResType* Data;

		public:
			WisePointer()
			{
				Data = 0;
			};

			WisePointer(ResType* Ptr)
			{
				Data = 0;
				*this = Ptr;
			};

			WisePointer(const WisePointer<ResType> &WPtr)
			{
				Data = 0;
				*this = WPtr;
			};

			~WisePointer()
			{
				if(Data)
				{
					Data->Release();
				};
			};

			inline operator=(ResType *Ptr)
			{
				if(Data)
				{
					Data->Release();
				};

				Data = Ptr;

				if(Data)
				{
					Data->AddRef();
				};

				return 0;
			};

			inline operator=(const WisePointer<ResType> &WPtr)
			{
				if(Data)
				{
					Data->Release();
				};

				Data = WPtr.Data;

				if(Data)
				{
					Data->AddRef();
				};

				return 0;
			};

			inline ResType& operator*() const
			{
				ASSERT(Data != 0 && "Tried to dereference a null Wise Pointer!");
				return *Data;
			};

			inline ResType* operator->() const
			{
				assert(Data != 0 && "Tried to dereference a null Wise Pointer!");
				return Data;
			};

			inline operator ResType*() const
			{
				return Data;
			};

			inline bool isValid() const
			{
				return (Data != NULL);
			};

            inline bool operator!()
			{
				return Data;
			};

			inline bool operator==(const WisePointer<ResType> &Target) const
			{
				return(Data == Target.Data);
			};

			inline bool operator==(const ResType* Target) const
			{
				return (Data == Target);
			};
		};
	};
};

#endif // _WISEPTR_H_INCLUDED_