/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "animcontext.h"
#include "worldstate.h"

Sith::Gameplay::WorldState* Sith::Gameplay::WorldState::StaticInterface = 0;

Sith::Gameplay::WorldState::WorldState() : Task("World State")
{
	// Set static interface
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple world states for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;
	CurrentLevel = 0;

	// COG LIB BINDINGS:
	COG::CogLib::InjectLibraryFunction("preloadhint", 1, &PreloadHint);

	COG::CogLib::InjectLibraryFunction("getlocalplayerthing", 0, &GetLocalPlayerThing);
	COG::CogLib::InjectLibraryFunction("jkgetlocalplayer", 0, &GetLocalPlayerThing);
	COG::CogLib::InjectLibraryFunction("getthingsector", 1, &GetThingSector);
	COG::CogLib::InjectLibraryFunction("getthingpos", 1, &GetThingPos);
	COG::CogLib::InjectLibraryFunction("getcurframe", 1, &GetCurFrame);
	COG::CogLib::InjectLibraryFunction("isthingmoving", 1, &IsThingMoving);
	COG::CogLib::InjectLibraryFunction("ismoving", 1, &IsThingMoving);
	COG::CogLib::InjectLibraryFunction("movetoframe", 3, &MoveToFrame);
	COG::CogLib::InjectLibraryFunction("setsectoradjoins", 2, &SetSectorAdjoins);
	COG::CogLib::InjectLibraryFunction("sectoradjoins", 2, &SetSectorAdjoins);
	COG::CogLib::InjectLibraryFunction("setsectorlight", 3, &SetSectorLight);
	COG::CogLib::InjectLibraryFunction("sectorlight", 3, &SetSectorLight);
	COG::CogLib::InjectLibraryFunction("setadjoinflags", 2, &SetAdjoinFlags);
	COG::CogLib::InjectLibraryFunction("clearadjoinflags", 2, &ClearAdjoinFlags);
	COG::CogLib::InjectLibraryFunction("setsurfaceflags", 2, &SetSurfaceFlags);
	COG::CogLib::InjectLibraryFunction("clearsurfaceflags", 2, &ClearSurfaceFlags);
	COG::CogLib::InjectLibraryFunction("setactorflags", 2, &SetActorFlags);
	COG::CogLib::InjectLibraryFunction("clearactorflags", 2, &ClearActorFlags);
	COG::CogLib::InjectLibraryFunction("setfacegeomode", 2, &SetFaceGeoMode);
	COG::CogLib::InjectLibraryFunction("surfacelightanim", 4, &SurfaceLightAnim);
	COG::CogLib::InjectLibraryFunction("getthinghealth", 1, &GetThingHealth);
	COG::CogLib::InjectLibraryFunction("gethealth", 1, &GetThingHealth);
	COG::CogLib::InjectLibraryFunction("setthinghealth", 2, &SetThingHealth);
	COG::CogLib::InjectLibraryFunction("sethealth", 2, &SetThingHealth);
	COG::CogLib::InjectLibraryFunction("damagething", 4, &DamageThing);
	COG::CogLib::InjectLibraryFunction("getthingparent", 1, &GetThingParent);
	COG::CogLib::InjectLibraryFunction("getthingtype", 1, &GetThingType);
	COG::CogLib::InjectLibraryFunction("getthingflags", 1, &GetThingFlags);
	COG::CogLib::InjectLibraryFunction("getphysicsflags", 1, &GetPhysicsFlags);
	COG::CogLib::InjectLibraryFunction("setcollidetype", 2, &SetCollideType);

	COG::CogLib::InjectLibraryFunction("getsectorcount", 0, &GetSectorCount);
	COG::CogLib::InjectLibraryFunction("getsurfacecount", 0, &GetSurfaceCount);

	return;
};

Sith::Gameplay::WorldState::~WorldState()
{
	CurrentLevel = 0;
	ThingList.clear();
	return;
};

Sith::Gameplay::Template* Sith::Gameplay::WorldState::GetTemplate(std::string TemplateName)
{
	// loop through level
	if(CurrentLevel.isValid())
	{
		return CurrentLevel->FindTemplate(TemplateName);
	};

	if(Static.isValid())
	{
	};

	return 0;
};

Sith::SITHERROR Sith::Gameplay::WorldState::Startup()
{
	// Load static.jkl into templates
	Static = Sith::VFS::ResourceManager::Get().GetLevel("static.jkl");
	return SE_OK;
};

Sith::SITHERROR Sith::Gameplay::WorldState::Process()
{
	return SE_OK;
};

Sith::SITHERROR Sith::Gameplay::WorldState::SetLevel(VFS::WisePointer<Res::Level> NewLevel)
{
	// Stop all COGs
	COG::VMProcess::Get().EndLevel();
	// Stop all animations
	Animation::Get().StopAll();
	// Clear the thing list
	StaticInterface->ThingList.clear();

	StaticInterface->CurrentLevel = NewLevel;

	if(StaticInterface->CurrentLevel.isValid())
	{
		// Clear all animation values:
		for(std::vector<Res::Level::SURFACE>::iterator gt = StaticInterface->CurrentLevel->SurfaceList.begin(); gt != StaticInterface->CurrentLevel->SurfaceList.end(); ++gt)
		{
			gt->CurrentCel = 0;
			gt->TexShiftU = 0;
			gt->TexShiftV = 0;
		};
	};

	return SE_OK;
};

COGLIB Sith::Gameplay::WorldState::PreloadHint(COG::VMType* MapName)
{
	StaticInterface->NextLevel = VFS::ResourceManager::Get().GetLevel(MapName->GetString());
	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::GetThingSector(COG::VMType* ThingNum)
{
	if((ThingNum->GetInt() >= 0) && (ThingNum->GetInt() < StaticInterface->ThingList.size()))
	{
		return new COG::VMType(StaticInterface->ThingList[ThingNum->GetInt()].Sector);
	};

	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::GetThingPos(COG::VMType *Thing)
{
	int ThingNum = Thing->GetInt();

	if (ThingNum < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		return new COG::VMType(Gameplay::WorldState::StaticInterface->ThingList[ThingNum].Pos);
	}
	return new COG::VMType(Math::VectorSet(0.0f, 0.0f, 0.0f));
};

COGLIB Sith::Gameplay::WorldState::GetLocalPlayerThing()
{
	return new COG::VMType(COG::VMTC_THING, StaticInterface->LocalPlayerThing, -1);
};

COGLIB Sith::Gameplay::WorldState::SetSectorAdjoins(COG::VMType* Sector, COG::VMType* AdjoinFlag)
{
	// TODO: Something? Do we need this?
	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::SetSectorLight(COG::VMType* Sector, COG::VMType* Light, COG::VMType* Delay)
{
	// TODO: Something
	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::MoveToFrame(COG::VMType *Thing, COG::VMType *Frame, COG::VMType *Speed)
{
	Gameplay::Thing *ThisThing;
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		ThisThing = &(Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()]);
	}
	else
	{
		return new COG::VMType(COG::VMTC_VOID);
	};

	int FrameNum = Frame->GetInt();

	if (FrameNum < ThisThing->Frames.size())
	{
		/*ThisThing->X = ThisThing->Frames[FrameNum].X;
		ThisThing->Y = ThisThing->Frames[FrameNum].Y;
		ThisThing->Z = ThisThing->Frames[FrameNum].Z;
		ThisThing->Pitch = ThisThing->Frames[FrameNum].Pitch;
		ThisThing->Yaw = ThisThing->Frames[FrameNum].Yaw;
		ThisThing->Roll = ThisThing->Frames[FrameNum].Roll;*/
		ThisThing->DestFrame = FrameNum;
		//ThisThing->FrameSpeed = Math::VectorDist(Math::VectorSet(ThisThing->Frames[FrameNum].X, ThisThing->Frames[FrameNum].Y, ThisThing->Frames[FrameNum].Z), Math::VectorSet(ThisThing->Frames[ThisThing->CurFrame].X, ThisThing->Frames[ThisThing->CurFrame].Y, ThisThing->Frames[ThisThing->CurFrame].Z)) / Speed->GetFloat();
		ThisThing->FrameSpeed = Speed->GetFloat() / 2;
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::IsThingMoving(COG::VMType* Object)
{
	return new COG::VMType(false);
};

COGLIB Sith::Gameplay::WorldState::GetCurFrame(COG::VMType* Object)
{
	if((Object->GetInt() >= 0) && (Object->GetInt() < StaticInterface->ThingList.size()))
	{
		return new COG::VMType(StaticInterface->ThingList[Object->GetInt()].CurFrame);
	};

	return new COG::VMType(0);
};

COGLIB Sith::Gameplay::WorldState::SetAdjoinFlags(COG::VMType* Surface, COG::VMType* Flags)
{
	if (Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		int SurfaceNum = Surface->GetInt();
		if ((SurfaceNum >= 0) && (SurfaceNum < Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList.size()))
		{
			if ((Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex > -1) &&
				(Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex < Gameplay::WorldState::StaticInterface->CurrentLevel->AdjoinList.size()))
			{
				Gameplay::WorldState::StaticInterface->CurrentLevel->AdjoinList[Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex].Flags |= Flags->GetInt();
			};
		};
	};
	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::ClearAdjoinFlags(COG::VMType* Surface, COG::VMType* Flags)
{
	if (Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		int SurfaceNum = Surface->GetInt();
		if (SurfaceNum < Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList.size())
		{
			if ((Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex > -1) && 
				(Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex < Gameplay::WorldState::StaticInterface->CurrentLevel->AdjoinList.size()))
			{
				Gameplay::WorldState::StaticInterface->CurrentLevel->AdjoinList[Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[SurfaceNum].AdjoinIndex].Flags &= ~Flags->GetInt();
			};
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::SetFaceGeoMode(COG::VMType* Surface, COG::VMType* Mode)
{
	if (Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		if (Surface->GetInt() < Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList.size())
		{
			Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[Surface->GetInt()].GeoMode = Mode->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::SetSurfaceFlags(COG::VMType* Surface, COG::VMType* Flags)
{
	if (Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		if (Surface->GetInt() < Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList.size())
		{
			Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[Surface->GetInt()].SurfaceFlags |= Flags->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::ClearSurfaceFlags(COG::VMType* Surface, COG::VMType* Flags)
{
	if (Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		if (Surface->GetInt() < Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList.size())
		{
			Gameplay::WorldState::StaticInterface->CurrentLevel->SurfaceList[Surface->GetInt()].SurfaceFlags &= ~Flags->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::SetActorFlags(COG::VMType* Thing, COG::VMType* Flags)
{
	if(Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		if((Thing->GetInt() >= 0) && (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size()))
		{
			Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].ActorFlags |= Flags->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::ClearActorFlags(COG::VMType* Thing, COG::VMType* Flags)
{
	if(Gameplay::WorldState::StaticInterface->CurrentLevel.isValid())
	{
		if((Thing->GetInt() >= 0) && (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size()))
		{
			Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].ActorFlags &= ~Flags->GetInt();
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::SurfaceLightAnim(COG::VMType* Surface, COG::VMType* StartLight, COG::VMType* EndLight, COG::VMType ChangeTime)
{
	// TODO: implement the actual fading
	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::GetThingHealth(COG::VMType* Thing)
{
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		return new COG::VMType(Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Health);
	};
	return new COG::VMType(-1);
};

COGLIB Sith::Gameplay::WorldState::SetThingHealth(COG::VMType* Thing, COG::VMType* Health)
{
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Health = Health->GetFloat();
	};
	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::WorldState::DamageThing(COG::VMType* Thing, COG::VMType* Damage, COG::VMType* Flags, COG::VMType* Damager)
{
	// TODO: Aside from the class and capture cogs, the 0x10 Inventory Flag will make the local player's damaged message run in an items.dat cog.
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		if(Damage->GetFloat() < 0)
		{
			return new COG::VMType(-1);
		}
		float ClipDamage, UnDamage;
		if (Damage->GetFloat() > Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Health) // more damage than there is health
		{
			ClipDamage = Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Health;
			UnDamage = Damage->GetFloat() - ClipDamage;
		}
		else
		{
			ClipDamage = Damage->GetFloat();
			UnDamage = 0.0f;
		};

		// send damaged message
		// TODO: SendMessageEx so we can send Flags as the second param
		COG::VMProcess::Get().SendMessage("damaged", COG::VMType(COG::VMTC_THING, Thing->GetInt(), -1), 0, COG::VMType(COG::VMTC_THING, Damager->GetInt(), -1), COG::VMType(*Damage), COG::VMType(*Flags));

		Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Health -= ClipDamage;
		return new COG::VMType(UnDamage);
	}
	return new COG::VMType(-1);
};

COGLIB Sith::Gameplay::WorldState::GetThingParent(COG::VMType* Thing)
{
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		return new COG::VMType(Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].Parent);
	};
	return new COG::VMType(-1);
};

COGLIB Sith::Gameplay::WorldState::GetThingType(COG::VMType* Thing)
{
	if (Thing->GetInt() < Gameplay::WorldState::StaticInterface->ThingList.size())
	{
		return new COG::VMType(Gameplay::WorldState::StaticInterface->ThingList[Thing->GetInt()].ThingType);
	};
	return new COG::VMType(-1);
};

COGLIB Sith::Gameplay::WorldState::GetSurfaceCount()
{
	if(StaticInterface->CurrentLevel.isValid())
	{
		return new COG::VMType((int)StaticInterface->CurrentLevel->SurfaceList.size());
	};

	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::GetSectorCount()
{
	if(StaticInterface->CurrentLevel.isValid())
	{
		return new COG::VMType((int)StaticInterface->CurrentLevel->SectorList.size());
	};

	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::GetThingFlags(COG::VMType* ThingNum)
{
	if(StaticInterface->CurrentLevel.isValid())
	{
		if((ThingNum->GetInt() >= 0) && (ThingNum->GetInt() < Get().ThingList.size()))
		{
			return new COG::VMType(Get().ThingList[ThingNum->GetInt()].ThingFlags);
		};
	};

	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::GetPhysicsFlags(COG::VMType* ThingNum)
{
	if(StaticInterface->CurrentLevel.isValid())
	{
		if((ThingNum->GetInt() >= 0) && (ThingNum->GetInt() < Get().ThingList.size()))
		{
			return new COG::VMType(Get().ThingList[ThingNum->GetInt()].ThingFlags);
		};
	};

	return new COG::VMType();
};

COGLIB Sith::Gameplay::WorldState::SetCollideType(COG::VMType* ThingNum, COG::VMType* Collide)
{
	if(StaticInterface->CurrentLevel.isValid())
	{
		if((ThingNum->GetInt() >= 0) && (ThingNum->GetInt() < Get().ThingList.size()))
		{
			switch(Collide->GetInt())
			{
			case 0:
				Get().ThingList[ThingNum->GetInt()].Collide = Gameplay::Template::CT_GHOST;
				break;
			case 1:
			case 2:
				Get().ThingList[ThingNum->GetInt()].Collide = Gameplay::Template::CT_SPHERE;
				break;
			case 3:
				Get().ThingList[ThingNum->GetInt()].Collide = Gameplay::Template::CT_POLY;
			};
		};
	};

	return new COG::VMType();
};