/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _WORLDSTATE_H_INCLUDED_
#define _WORLDSTATE_H_INCLUDED_

#include "worldcamera.h"

namespace Sith
{
	namespace Gameplay
	{
		class WorldState : public System::Task
		{
			friend class System::KernelProc;
		private:
			static WorldState* StaticInterface;

			WorldState();
			~WorldState();

			SITHERROR Startup();
			SITHERROR Process();

			VFS::WisePointer<Res::Level> NextLevel;

		public:
			Sith::VFS::WisePointer<Res::Level> Static;
			Sith::VFS::WisePointer<Res::Level> CurrentLevel;

			Gameplay::Template* GetTemplate(std::string TemplateName);

			std::vector<Gameplay::Thing> ThingList;
			Gameplay::Camera LocalCamera;

			int LocalPlayerThing;

			static WorldState& Get()
			{
				return *StaticInterface;
			};

			SITHERROR SetLevel(Sith::VFS::WisePointer<Res::Level> NewLevel);

			COGLIBAPI PreloadHint(COG::VMType* MapName);
			COGLIBAPI GetThingSector(COG::VMType* ThingNum);
			COGLIBAPI GetThingPos(COG::VMType *Thing);
			COGLIBAPI GetLocalPlayerThing();
			COGLIBAPI SetSectorAdjoins(COG::VMType* Sector, COG::VMType* AdjoinFlag);
			COGLIBAPI SetSectorLight(COG::VMType* Sector, COG::VMType* Light, COG::VMType* Delay);
			COGLIBAPI MoveToFrame(COG::VMType *Thing, COG::VMType *Frame, COG::VMType *Speed);
			COGLIBAPI IsThingMoving(COG::VMType* Object);
			COGLIBAPI GetCurFrame(COG::VMType* Object);
			COGLIBAPI SetAdjoinFlags(COG::VMType* Surface, COG::VMType* Flags);
			COGLIBAPI ClearAdjoinFlags(COG::VMType* Surface, COG::VMType* Flags);
			COGLIBAPI SetFaceGeoMode(COG::VMType* Surface, COG::VMType* Mode);
			COGLIBAPI SetSurfaceFlags(COG::VMType* Surface, COG::VMType* Flags);
			COGLIBAPI ClearSurfaceFlags(COG::VMType* Surface, COG::VMType* Flags);
			COGLIBAPI SetActorFlags(COG::VMType* Thing, COG::VMType* Flags);
			COGLIBAPI ClearActorFlags(COG::VMType* Thing, COG::VMType* Flags);
			COGLIBAPI SurfaceLightAnim(COG::VMType* Surface, COG::VMType* StartLight, COG::VMType* EndLight, COG::VMType ChangeTime);
			COGLIBAPI GetThingHealth(COG::VMType* Thing);
			COGLIBAPI SetThingHealth(COG::VMType* Thing, COG::VMType* Health);
			COGLIBAPI DamageThing(COG::VMType* Thing, COG::VMType* Damage, COG::VMType* Flags, COG::VMType* Damager);
			COGLIBAPI GetThingParent(COG::VMType* Thing);
			COGLIBAPI GetThingType(COG::VMType* Thing);
			COGLIBAPI GetThingFlags(COG::VMType* Thing);
			COGLIBAPI GetPhysicsFlags(COG::VMType* Thing);
			COGLIBAPI SetCollideType(COG::VMType* Thing, COG::VMType* Collide);

			COGLIBAPI GetSurfaceCount();
			COGLIBAPI GetSectorCount();
		};
	};
};
#endif // _WORLDSTATE_H_INCLUDED_